/** \file    common.h
    \brief   provides forward class declarations, several abstract data classes, 
             a couple of global functions, and global constants (header file only).
    \author  EV
    \date    2009-2013
*/
#pragma once
#include <utility>
#include <vector>
#include <string>

/// common namespace for all core SMILE classes, functions and variables
namespace smile{

/* ------- Global constants not changeable in runtime ------- */

/// \name ----------- Parameters for potential and force evaluation -------------  
///@{

/// Dimensions of the real world. DO NOT CHANGE (unless you are God)!
const unsigned int N_DIM=3;

/// max number of basis-function expansion members (radial and angular). 
const size_t MAX_NCOEFS_RADIAL=101;
const size_t MAX_NCOEFS_ANGULAR=51;

/// relative accuracy of density computation
const double EPSREL_DENSITY_INT=1e-4;

/// relative accuracy of potential computation (integration tolerance parameter)
const double EPSREL_POTENTIAL_INT=1e-6;

/// absolute error in potential computation (supercedes the relative error in case of very small coefficients)
const double EPSABS_POTENTIAL_INT=1e-15;

/// relative accuracy of cell mass computation in Schwarzschild modelling
const double EPSREL_MASS_INT=1e-2;

/// inner and outer cutoff radii appearing in calculations of M(r), r(E) and related
const double MIN_RADIUS_CUTOFF=1e-6;
const double MAX_RADIUS_CUTOFF=1e+6;

/// auxiliary softening parameter for Spline potential to avoid singularities in some odd cases
const double SPLINE_MIN_RADIUS=1e-10;

///@}
/// \name ------------- Orbit integration parameters ----------
///@{

/// maximal number of integration steps (in case something went wrong with the timestep, do not dead-lock)
const size_t NUMSTEP_MAX=static_cast<size_t>(1e7);

/// for DOP853 integrator, setting this flag eliminates three calls to force function per timestep at the expense 
/// of getting a slightly lower (still acceptable) accuracy in the interpolated solution (6th instead of 7th order)
#define DOP853_DENSE_OUTPUT_FAST

///@}

/* ------- Declare several classes implemented in common.cpp and various other modules -------- */

// forward declarations
class CDensity;
class CPotential;
class COrbit;
class CMassModel;
class CBasicOrbitRuntimeFnc;
class CBasicOrbitRuntimeFncCreator;
class CBasicInformation;
template<typename NumT> struct CPosPoint;
template<typename NumT> struct CPosVelPoint;

/// \name ------- define several convenience types ------- 
///@{
typedef float sdNumType;                  ///< how to store orbit data for Schwarzschild modelling (float or double)
typedef float smNumType;                  ///< how to handle Schwarzschild modelling data internally (float or double)
typedef std::vector<double> vectord;      ///< vector of double values
typedef std::vector<float> vectorf;       ///< vector of float values
typedef std::pair<double, double> paird;  ///< a pair of doubles
typedef std::vector< paird > vectorpd;    ///< vector of pairs of doubles
typedef std::pair<float, float> pairf;    ///< a pair of floats
typedef std::vector< pairf > vectorpf;    ///< vector of pairs of floats
typedef vectord OdeStateType;             ///< container for the array of variables used in ODE integration
///@}
/// \name ----------- definition of various data structures ------------- 
///@{

/** Data-only class containing 3 coordinates and 3 velocities. 
    \tparam NumT is a numerical type (double or float).
    Copy constructor provided for both instantiations so that each one may be 
    seamlessly converted to another one by (implicit) type cast.
*/
template<typename NumT> struct CPosVelPoint
{
public:
    /// default constructor
    CPosVelPoint() { for(unsigned int i=0; i<N_DIM; i++) { Pos[i]=0; Vel[i]=0; } }
    /// construct from six float/double numbers
    CPosVelPoint( const NumT Pos0, const NumT Pos1, const NumT Pos2, const NumT Vel0, const NumT Vel1, const NumT Vel2)
    { Pos[0]=Pos0; Pos[1]=Pos1; Pos[2]=Pos2; Vel[0]=Vel0; Vel[1]=Vel1; Vel[2]=Vel2; };
    /// copy constructor from CPosVelPoint<double>
    CPosVelPoint( const CPosVelPoint<double> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); Vel[i]=static_cast<NumT>(src.Vel[i]); } };
    /// copy constructor from CPosVelPoint<float>
    CPosVelPoint( const CPosVelPoint<float> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); Vel[i]=static_cast<NumT>(src.Vel[i]); } };
    NumT Pos[N_DIM];   //< Components of position vector. Warning: no range check performed when setting Pos or Vel[n>=N_DIM] !
    NumT Vel[N_DIM];   //< Components of velocity vector
};

/** Data-only class containing 3 coordinates. 
    \tparam NumT may be float or double.
    Copy constructor provided for both instantiations so that each one may be 
    seamlessly converted to another one by (implicit) type cast.
*/
template<typename NumT> struct CPosPoint
{
public:
    /// default constructor
    CPosPoint() { for(unsigned int i=0; i<N_DIM; i++) { Pos[i]=0; } }
    /// construct from three float/double numbers
    CPosPoint( const NumT Pos0, const NumT Pos1, const NumT Pos2)
    { Pos[0]=Pos0; Pos[1]=Pos1; Pos[2]=Pos2; };
    /// copy constructor from CPosPoint<double>
    CPosPoint( const CPosPoint<double> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    /// copy constructor from CPosPoint<float>
    CPosPoint( const CPosPoint<float> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    /// copy constructor from CPosVelPoint<double>
    CPosPoint( const CPosVelPoint<double> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    /// copy constructor from CPosVelPoint<float>
    CPosPoint( const CPosVelPoint<float> &src ) { for(unsigned int i=0; i<N_DIM; i++) 
    { Pos[i]=static_cast<NumT>(src.Pos[i]); } };
    NumT Pos[N_DIM];   //< Components of position vector. Warning: no range check performed when setting Pos or Vel[n>=N_DIM] !
};

/** An array of particles with positions, velocities and masses.
    It is implemented as a separate structure instead of just a vector of pairs, 
    because of the limitations of C++ that don't allow to template a typedef without a class.
    \tparam NumT is a numerical type (double or float)
*/
template<typename NumT> struct CPointMassSet {
    typedef std::pair< CPosVelPoint<NumT>, NumT> ElemType;  ///< templated typedef of a single particle
    typedef std::vector< ElemType > Type;                   ///< templated typedef of an array of particles
    Type data;           ///< particles are stored in this array
    ///  default empty constructor
    CPointMassSet() {};
    /** a seamless conversion constructor from another point mass set with a possibly different template argument.
        \tparam OtherNumT is a numerical type (double or float) of the source PointMassSet */
    template<typename OtherNumT> CPointMassSet(const CPointMassSet<OtherNumT> &src) {
        data.reserve(src.size()); 
        for(size_t i=0; i<src.size(); i++) 
            data.push_back(ElemType(CPosVelPoint<NumT>(src[i].first), static_cast<NumT>(src[i].second)));
    }
    inline ElemType& at(size_t index) { return data.at(index); }             // convenience shorthand
    inline const ElemType& at(size_t index) const { return data.at(index); } // convenience shorthand
    inline ElemType& operator[](size_t index) { return data[index]; }        // convenience shorthand
    inline const ElemType& operator[](size_t index) const { return data[index]; } // another shorthand
    inline size_t size() const { return data.size(); }  // convenience shorthand
    inline void add( const CPosVelPoint<NumT> &first, const NumT second) /// convenience function to add an element
    { data.push_back(ElemType(first, second)); }
};
///@}
/// \name ----------- definition of some basic (abstract) classes to be implemented in various modules ------------- 
///@{

/** defines a routine which may be called after each integration timestep 
    to perform user-specific data collection.
    (such as computing cell occupation times in Schw model or recording Poincare section points).
*/
class CBasicOrbitRuntimeFnc
{
public:
    /// lists all possible runtime functions
enum FNCTYPE {
    FT_TRAJ_ANALYSIS,   ///< COrbitRuntimeTrajectory
    FT_TRAJ_SAMPLE,     ///< COrbitRuntimeTrajSample
    FT_POINCARE,        ///< COrbitRuntimePoincare
    FT_PERICENTER,      ///< COrbitRuntimePericenter
    FT_SCHW,            ///< COrbitRuntimeSchw
    FT_VEL_PERTURBER    ///< COrbitRuntimeVelPerturber
};
    /// result of processing of a timestep 
enum STEPRESULT {
    SR_CONTINUE,        ///< need to continue integration
    SR_TERMINATE,       ///< may terminate integration if necessary
    SR_REINIT           ///< variables have changed, need to re-initialize integrator
};
    /// constructs instance and initializes some internal data
    explicit CBasicOrbitRuntimeFnc(const COrbit* _orbit): orbit(_orbit) {} ;

    /// destroys all necessary internal data
    virtual ~CBasicOrbitRuntimeFnc() {};

    /// derived classes return their type
    virtual FNCTYPE FncType() const=0;

    /** the following method is called after each timestep.
        \param[in] told is the value of time at the beginning of the timestep
        \param[in] tcur is the time at the end of timestep
        \param[in,out] vars are the phase-space variables at the end of timestep; 
        the function may change them and return SR_REINIT to indicate that the ODE integrator 
        needs to be restarted.

        This procedure may call orbit->getPos() and orbit->getVel()
        to get interpolated coordinates on the interval [told, tcur].
        \return true if integration may be finished right now [default], false if needs to continue.
        (the ultimate decision is taken by orbit integration routine and may ignore the demands of runtime functions) 
    */
    virtual STEPRESULT Timestep(const double told, const double tcur, CPosVelPoint<double> *vars)=0;

    /// called when integration is finished, performs cleanup/postprocessing but doesn't destroy collected data
    virtual void Finish() {};

    /// returns the collected data as an instance of an information object
    virtual CBasicInformation* getData() const = 0;
protected:
    const COrbit* orbit;  ///< pointer to the orbit that is being integrated
};

/** defines a basic class that creates some sort of runtime function for any orbit in the orbit library.
    Derived classes may receive additional data in constructor and emit specific types 
    of timestep functions for a given orbit.
    Thus the interface between orbit library integrator and the specific class of runtime functions 
    is mediated by corresponding creator classes, one for each class of orbit runtime fnc.
*/
class CBasicOrbitRuntimeFncCreator
{
public:
    virtual ~CBasicOrbitRuntimeFncCreator() {};

    /// create an instance of the runtime function for a given orbit.
    virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const=0;
};

/** parent class for data container objects used to store various data and exchange it between objects.
    Data stored here is related to some orbit properties (e.g. orbit classification results, 
    or points in the trajectory, or data for Schwarzschild modelling).
    The instances of runtime functions derived from CBasicOrbitRuntimeFnc return specific information 
    objects which contain data collected about the orbit during integration.
    Classes defined in fileio.h handle storing and loading data.
    Classes derived from CBasicSchwData also use the CSchwInformation class to store data for modelling.
*/
class CBasicInformation
{
public:
    /// lists all possible information types
enum INFOTYPE {
    IT_UNKNOWN,         ///< dark matter
    IT_TRAJ_ANALYSIS,   ///< orbit analysis: COrbitInformation
    IT_TRAJ_SAMPLE,     ///< CTrajSampleInformation
    IT_POINCARE,        ///< CPoincareInformation [unused]
    IT_PERICENTER,      ///< CPericenterInformation
    IT_SCHW,            ///< CSchwInformation
    IT_VEL_PERTURBER    ///< CVelPerturberInformation
};
    virtual ~CBasicInformation() {};
    virtual INFOTYPE infoType() const=0;  ///< return the derived class type
    /// present a textual representation of the data contained in this class (may be empty string)
    virtual std::string toString() const = 0;
};

/** basic class of ODE system used in the orbit integrator. 
    It provides a function for computing the right-hand side of the system of differential equations, 
    and a number of auxiliary functions (for initializing the system and providing some operations).
    The COrbit class is descending from this one.
*/
class COdeSystem {
public:
    /// result of processing of a timestep 
enum STEPRESULT {
    SR_CONTINUE,        ///< need to continue integration
    SR_TERMINATE,       ///< may terminate integration if necessary
    SR_REINIT           ///< variables have changed, need to re-initialize integrator
};
    COdeSystem() {};
    virtual ~COdeSystem() {};
    virtual void initState(OdeStateType* state)=0;
    virtual void odeFnc(double t, const OdeStateType& x, OdeStateType* dx)=0;
    virtual bool isStdHamiltonian() { return false; };
    virtual double getRelErrorToleranceFactor(const OdeStateType& /*state*/) const { return 1; };
    virtual STEPRESULT processStep(double timePrev, double timeCurr, OdeStateType* state)=0;
};

/** basic class for numerical integrators of ODE system */
class CBasicOdeIntegrator
{
public:
    /// list of various ODE stepper types
enum STEPPERKIND {
    SK_DEFAULT,
    SK_LEAPFROG_NB,
    SK_DOP853,
    SK_IAS15,
    SK_ODEINT_CK5,
    SK_ODEINT_DP5,
    SK_ODEINT_BS3,
    SK_ODEINT_BS,
    SK_ODEINT_RK4,
    SK_ODEINT_SYMPL4
};

    CBasicOdeIntegrator(COdeSystem* _odeSystem):
      odeSystem(_odeSystem), timePrev(0), timeCurr(0) {};

    virtual ~CBasicOdeIntegrator() {};

    /** perform integration */
    virtual void integrateToTime(double timeEnd)=0;

    /** return the time to which the integration has proceeded so far */
    double getTime() const { return timeCurr; }

    /** return interpolated value of c-th variable at time t, where t must lie within current timestep interval;
        called from various runtime functions that need to obtain x(t) at arbitrary times */
    virtual double getInterpolatedSolution(unsigned int c, double t) const=0;

protected:
    COdeSystem* odeSystem;
    double timePrev, timeCurr;      //< previous and current value of time to which the integration has advanced
};

///@}
/// \name -------- Various convenience definitions --------
///@{

/// convenience definition of vectors of timestep functions and function creators
typedef std::vector<CBasicOrbitRuntimeFnc*> vectorRuntimeFncs;
typedef std::vector<const CBasicOrbitRuntimeFncCreator*> vectorRuntimeFncCreators;
typedef std::vector<const CBasicInformation*> vectorInformation;

/// convenience definitions of templated vectors containing particle sets with/without masses
typedef std::vector< CPosVelPoint<double> > CPosVelDataDouble;
typedef std::vector< CPosVelPoint<float > > CPosVelDataFloat;

/// convenience shorthand for squaring a number
inline double pow_2(double x) { return x*x; }
inline float pow_2(float x) { return x*x; }
inline size_t pow_2(size_t x) { return x*x; }

/// convenience function for sorting arrays of std::pair<float/double, something>
template<typename T1, typename T2> bool comparepair(const std::pair<T1,T2>& val1, const std::pair<T1,T2>& val2)
{  return val1.first < val2.first;  }

///@}

}  // namespace

