#include "configfile.h"
#include <cassert>
#include "utils.h"
#include "potential.h"
#include "orbitint.h"
#include "odeint.h"
#include <fstream>

namespace smile{

/* implementation-dependent functions */

struct CSection {
    std::string name;
    std::vector< std::pair< std::string, std::string > > keyValues;
};

struct CIniData {
    std::string fileName;
    std::vector<CSection> sections;
    std::string header;
    bool modified;
};

bool comparestr(const std::string& one, const std::string& two) 
{
    if(one.size()!=two.size()) return false;
    for(size_t i=0; i<one.size(); i++) {
        char o=one[i];
        if(o>=65&&o<=90) o+=32;
        char t=two[i];
        if(t>=65&&t<=90) t+=32;
        if(o!=t) return false;
    }
    return true;
}

CConfigFile::CConfigFile(const std::string& fileName)
{
    CIniData* data=new CIniData();
    pImpl=data;
    data->modified=false;
    data->fileName=fileName;
    std::ifstream strm(fileName.c_str());
    if(!strm) return;  // file does not exist
    std::string buffer, key;
    int secIndex=-1;
    while(std::getline(strm, buffer)) {
        std::string::size_type indx=buffer.find_first_not_of(" \t\n\r");
        if(indx!=std::string::npos) 
            buffer.erase(0, indx);
        else  // don't deal with empty lines
            continue;
        if(buffer.size()>=3 && buffer[0]=='[' && (indx=buffer.find(']'))!=std::string::npos) {  // section start
            buffer=buffer.substr(1, indx-1);
            indx=buffer.find_first_not_of(" \t\n\r");
            if(indx!=std::string::npos && indx>0) buffer.erase(0, indx);
            indx=buffer.find_last_not_of(" \t\n\r");
            if(indx!=std::string::npos && indx+1<buffer.size()) buffer.erase(indx+1);
            if(buffer.empty()) continue;
            secIndex=-1;
            for(size_t i=0; i<data->sections.size(); i++)
                if(comparestr(data->sections[i].name, buffer))
                    secIndex=i;
            if(secIndex<0) { // section not found before, add it
                CSection sec;
                sec.name=buffer;
                data->sections.push_back(sec);
                secIndex=data->sections.size()-1;
            }
        } else {  // not a section
            if(secIndex<0 || buffer.empty() || buffer[0]=='#' || buffer[0]==';' || buffer[0]=='/') {  // comment
                if(secIndex<0)  // no section started yet, append to header
                    data->header+=buffer+"\n";
                else {  // add ignored line to the current section
                    data->sections[secIndex].keyValues.push_back(std::pair<std::string, std::string>("", buffer));
                }
                continue;
            }
            indx=buffer.find('=');
            if(indx!=std::string::npos && indx>0) {  // key-value pair
                key=buffer.substr(0, indx);
                buffer.erase(0, indx+1);
                indx=key.find_last_not_of(" \t\n\r");
                if(indx!=std::string::npos && indx+1<key.size()) key.erase(indx+1);
                indx=buffer.find_first_not_of(" \t\n\r");
                if(indx!=std::string::npos && indx>0) buffer.erase(0, indx);
                indx=buffer.find_last_not_of(" \t\n\r");
                if(indx!=std::string::npos && indx+1<buffer.size()) buffer.erase(indx+1);
                data->sections[secIndex].keyValues.push_back(std::pair<std::string, std::string>(key, buffer));
            } else 
                data->sections[secIndex].keyValues.push_back(std::pair<std::string, std::string>("", buffer));
        }
    }
    strm.close();
}

CConfigFile::~CConfigFile()
{
    CIniData* data=static_cast<CIniData*>(pImpl);
    if(data->modified && !data->fileName.empty()) { // need to save ini file
        std::ofstream strm(data->fileName.c_str());
        if(!strm) {
            delete data;
            return;
        }
        if(!data->header.empty())
            strm << data->header;
        for(size_t is=0; is<data->sections.size(); is++) {
            strm << ("["+data->sections[is].name+"]\n");
            for(size_t ik=0; ik<data->sections[is].keyValues.size(); ik++) {
                strm << (data->sections[is].keyValues[ik].first.empty() ? data->sections[is].keyValues[ik].second+"\n" : 
                    data->sections[is].keyValues[ik].first+"="+data->sections[is].keyValues[ik].second+"\n");
            }
            strm << "\n";
        }
    }
    delete data;
}

std::string CConfigFile::getValueString(const std::string& key, const std::string& defaultValue) const
{
    // split key into SectionName/KeyName
    std::string::size_type indx=key.find_first_of("/");
    if(indx==std::string::npos || indx<1 || indx+1>=key.size()) 
        return defaultValue;
    std::string secName=key.substr(0, indx);
    std::string keyName=key.substr(indx+1);
    CIniData* data=static_cast<CIniData*>(pImpl);
    for(size_t is=0; is<data->sections.size(); is++) 
        if(comparestr(data->sections[is].name, secName)) {
            for(size_t ik=0; ik<data->sections[is].keyValues.size(); ik++)
                if(comparestr(data->sections[is].keyValues[ik].first, keyName))
                    return data->sections[is].keyValues[ik].second;
        }
    return defaultValue;
}

void CConfigFile::setValue(const std::string& key, const std::string& value)
{
    // split key into SectionName/KeyName
    std::string::size_type indx=key.find_first_of("/");
    if(indx==std::string::npos || indx<1 || indx+1>=key.size()) 
        return;
    std::string secName=key.substr(0, indx);
    std::string keyName=key.substr(indx+1);
    CIniData* data=static_cast<CIniData*>(pImpl);
    for(size_t is=0; is<data->sections.size(); is++) 
        if(comparestr(data->sections[is].name, secName)) {
            for(size_t ik=0; ik<data->sections[is].keyValues.size(); ik++)
                if(comparestr(data->sections[is].keyValues[ik].first, keyName)) {
                    if(data->sections[is].keyValues[ik].second!=value) {
                        data->sections[is].keyValues[ik].second=value;
                        data->modified=true;
                    }
                    return;
                }
            // key not found -- add new
            data->sections[is].keyValues.push_back(std::pair<std::string, std::string>(keyName, value));
            data->modified=true;
            return;
        }
    // section not found -- add new
    CSection sec;
    sec.name=secName;
    sec.keyValues.push_back(std::pair<std::string, std::string>(keyName, value));
    data->sections.push_back(sec);
    data->modified=true;
}

void CConfigFile::listGroups(std::vector<std::string> *list) const
{
    assert(list);
    list->resize(static_cast<CIniData*>(pImpl)->sections.size());
    for(size_t i=0; i<list->size(); i++)
        list->at(i)=static_cast<CIniData*>(pImpl)->sections[i].name;
}

/* derived functions */
int CConfigFile::getValueInt(const std::string& key, int defaultValue) const
{
    return convertTo<int>(getValueString(key, convertToString(defaultValue)));
}

double CConfigFile::getValueDouble(const std::string& key, double defaultValue) const
{
    return convertTo<double>(getValueString(key, convertToString(defaultValue)));
}

bool CConfigFile::getValueBool(const std::string& key, bool defaultValue) const
{
    return convertTo<bool>(getValueString(key, defaultValue?"True":"False"));
}

void CConfigFile::setValue(const std::string& key, int value)
{
    setValue(key, convertToString(value));
}

void CConfigFile::setValue(const std::string& key, unsigned int value)
{
    setValue(key, convertToString(value));
}

void CConfigFile::setValue(const std::string& key, double value)
{
    setValue(key, convertToString(value));
}

void CConfigFile::setValue(const std::string& key, double value, unsigned int width)
{
    setValue(key, convertToString(value, width));
}

void CConfigFile::setValue(const std::string& key, bool value)
{
    setValue(key, convertToString(value));
}

/* ---------- reading/writing various configuration blocks of SMILE ---------- */
void readConfigPotential (const CConfigFile& configReader, CConfigPotential* ptrConfigPotential, unsigned int potGroup)
{
    assert(ptrConfigPotential);
    std::string pgroup="Potential"+(potGroup>0?convertToString(potGroup):"");
    ptrConfigPotential->PotentialType = getPotentialTypeByName(configReader.getValueString(pgroup+"/Type"));
    ptrConfigPotential->DensityType = getDensityTypeByName(configReader.getValueString(pgroup+"/DensityModel", ""));
    ptrConfigPotential->mass = std::max<double>(0, configReader.getValueDouble(pgroup+"/Mass", ptrConfigPotential->mass));
    ptrConfigPotential->q = std::max<double>(0, std::min<double>(1, configReader.getValueDouble(pgroup+"/q_YtoX", ptrConfigPotential->q))); 
    ptrConfigPotential->p = std::max<double>(0, std::min<double>(1, configReader.getValueDouble(pgroup+"/p_ZtoX", ptrConfigPotential->p))); 
    ptrConfigPotential->Mbh = std::max<double>(0, configReader.getValueDouble(pgroup+"/Mbh", ptrConfigPotential->Mbh));
    ptrConfigPotential->scalerad = std::max<double>(0, configReader.getValueDouble(pgroup+"/Rscale", ptrConfigPotential->scalerad));
    ptrConfigPotential->scalerad2 = std::max<double>(0, configReader.getValueDouble(pgroup+"/Rscale2", ptrConfigPotential->scalerad2));
    ptrConfigPotential->Gamma = std::max<double>(0, configReader.getValueDouble(pgroup+"/Gamma", ptrConfigPotential->Gamma));
    ptrConfigPotential->sersicIndex = std::max<double>(0, configReader.getValueDouble(pgroup+"/SersicIndex", ptrConfigPotential->sersicIndex));
    ptrConfigPotential->Ncoefs_radial = std::min<size_t>(MAX_NCOEFS_RADIAL, configReader.getValueInt(pgroup+"/Ncoefs_radial", 
        static_cast<unsigned int>(ptrConfigPotential->Ncoefs_radial)));
    ptrConfigPotential->Ncoefs_vertical = std::min<size_t>(MAX_NCOEFS_RADIAL, configReader.getValueInt(pgroup+"/Ncoefs_vertical", 
        static_cast<unsigned int>(ptrConfigPotential->Ncoefs_vertical)));
    ptrConfigPotential->Ncoefs_angular = std::min<size_t>(MAX_NCOEFS_ANGULAR, configReader.getValueInt(pgroup+"/Ncoefs_angular", 
        static_cast<unsigned int>(ptrConfigPotential->Ncoefs_angular)));
    ptrConfigPotential->Alpha = std::max<double>(0, configReader.getValueDouble(pgroup+"/Alpha", ptrConfigPotential->Alpha));
    ptrConfigPotential->Rmax = std::max<double>(0, configReader.getValueDouble(pgroup+"/Rmax", ptrConfigPotential->Rmax));
    std::string NbodyFile = configReader.getValueString(pgroup+"/NbodyFile");
    //QFileInfo NbodyFileName(QDir(WorkDir), NbodyFile);   ////!!! fixme
    ptrConfigPotential->NbodyFile=NbodyFile;//Name.absoluteFilePath().toStdString();
    ptrConfigPotential->SymmetryType = getSymmetryTypeByName(configReader.getValueString(pgroup+"/Symmetry"));
    ptrConfigPotential->nbpotEps   = configReader.getValueDouble(pgroup+"/treecodeEps", ptrConfigPotential->nbpotEps);
    ptrConfigPotential->nbpotTol   = std::min<double>(1, std::max<double>(0, configReader.getValueDouble(pgroup+"/treecodeTheta", ptrConfigPotential->nbpotTol)));
    ptrConfigPotential->splineSmoothFactor = std::max<double>(0, configReader.getValueDouble(pgroup+"/splineSmoothFactor", ptrConfigPotential->splineSmoothFactor));
    ptrConfigPotential->splineRadiusMin = configReader.getValueDouble(pgroup+"/splineRadiusMin", 0);
    ptrConfigPotential->splineRadiusMax = configReader.getValueDouble(pgroup+"/splineRadiusMax", 0);
    ptrConfigPotential->splineZMin = configReader.getValueDouble(pgroup+"/splineZMin", 0);
    ptrConfigPotential->splineZMax = configReader.getValueDouble(pgroup+"/splineZMax", 0);
}

void writeConfigPotential(CConfigFile& configWriter, const CConfigPotential* ptrConfigPotential, unsigned int potGroup)
{
    assert(ptrConfigPotential);
    std::string pgroup="Potential"+(potGroup>0?convertToString(potGroup):"");
    configWriter.setValue(pgroup+"/Type", PotentialNames[ptrConfigPotential->PotentialType]);
    configWriter.setValue(pgroup+"/DensityModel", DensityNames[ptrConfigPotential->DensityType]);
    configWriter.setValue(pgroup+"/Mass", ptrConfigPotential->mass);
    configWriter.setValue(pgroup+"/q_YtoX", ptrConfigPotential->q);
    configWriter.setValue(pgroup+"/p_ZtoX", ptrConfigPotential->p);
    configWriter.setValue(pgroup+"/Mbh", ptrConfigPotential->Mbh);
    configWriter.setValue(pgroup+"/Rscale", ptrConfigPotential->scalerad);
    configWriter.setValue(pgroup+"/Rscale2", ptrConfigPotential->scalerad2);
    configWriter.setValue(pgroup+"/Gamma", ptrConfigPotential->Gamma);
    configWriter.setValue(pgroup+"/SersicIndex", ptrConfigPotential->sersicIndex);
    configWriter.setValue(pgroup+"/Ncoefs_radial",  static_cast<unsigned int>(ptrConfigPotential->Ncoefs_radial));
    configWriter.setValue(pgroup+"/Ncoefs_angular", static_cast<unsigned int>(ptrConfigPotential->Ncoefs_angular));
    configWriter.setValue(pgroup+"/Ncoefs_vertical",static_cast<unsigned int>(ptrConfigPotential->Ncoefs_vertical));
    configWriter.setValue(pgroup+"/Alpha", ptrConfigPotential->Alpha);
    configWriter.setValue(pgroup+"/Rmax", ptrConfigPotential->Rmax);
    configWriter.setValue(pgroup+"/Symmetry", SymmetryNames[ptrConfigPotential->SymmetryType]);
    configWriter.setValue(pgroup+"/NbodyFile", ptrConfigPotential->NbodyFile);
    configWriter.setValue(pgroup+"/treecodeEps", ptrConfigPotential->nbpotEps);
    configWriter.setValue(pgroup+"/treecodeTheta", ptrConfigPotential->nbpotTol);
    configWriter.setValue(pgroup+"/splineSmoothFactor", ptrConfigPotential->splineSmoothFactor);
    configWriter.setValue(pgroup+"/splineRadiusMin", ptrConfigPotential->splineRadiusMin);
    configWriter.setValue(pgroup+"/splineRadiusMax", ptrConfigPotential->splineRadiusMax);
    configWriter.setValue(pgroup+"/splineZMin", ptrConfigPotential->splineZMin);
    configWriter.setValue(pgroup+"/splineZMax", ptrConfigPotential->splineZMax);
}

void readConfigOrbit (const CConfigFile& configReader, CConfigOrbit* ptrConfigOrbit)
{
    assert(ptrConfigOrbit);
    ptrConfigOrbit->N_dim = std::max<int>(2, std::min<int>(3, configReader.getValueInt("Orbit/N_dim", N_DIM)));
    ptrConfigOrbit->Omega = configReader.getValueDouble("Orbit/Omega", 0);
    ptrConfigOrbit->integratorType=getIntegratorTypeByName(configReader.getValueString("Orbit/integratorType", getIntegratorNameByType(ptrConfigOrbit->integratorType)));
    ptrConfigOrbit->accuracyRel= configReader.getValueDouble("Orbit/accuracyRelative", ptrConfigOrbit->accuracyRel);
    ptrConfigOrbit->accuracyAbs= configReader.getValueDouble("Orbit/accuracyAbsolute", ptrConfigOrbit->accuracyAbs);
    ptrConfigOrbit->accuracyIAS15= configReader.getValueDouble("Orbit/accuracyIAS15", ptrConfigOrbit->accuracyIAS15);
    ptrConfigOrbit->accuracyTreecode= configReader.getValueDouble("Orbit/accuracyTreecode", ptrConfigOrbit->accuracyTreecode);
    ptrConfigOrbit->treecodeSymmetrizeTimestep= configReader.getValueBool("Orbit/treecodeSymmetrizeTimestep", ptrConfigOrbit->treecodeSymmetrizeTimestep);
    ptrConfigOrbit->accuracyFixedTimestep= configReader.getValueDouble("Orbit/accuracyFixedTimestep", ptrConfigOrbit->accuracyFixedTimestep);
    ptrConfigOrbit->relaxRate = configReader.getValueDouble("Orbit/relaxationRate", ptrConfigOrbit->relaxRate);
}

void writeConfigOrbit(CConfigFile& configWriter, const CConfigOrbit* ptrConfigOrbit)
{
    assert(ptrConfigOrbit);
    configWriter.setValue("Orbit/N_dim", ptrConfigOrbit->N_dim);
    configWriter.setValue("Orbit/Omega", ptrConfigOrbit->Omega);
    configWriter.setValue("Orbit/integratorType", getIntegratorNameByType(ptrConfigOrbit->integratorType));
    configWriter.setValue("Orbit/accuracyRelative", ptrConfigOrbit->accuracyRel);
    configWriter.setValue("Orbit/accuracyAbsolute", ptrConfigOrbit->accuracyAbs);
    configWriter.setValue("Orbit/accuracyIAS15", ptrConfigOrbit->accuracyIAS15);
    configWriter.setValue("Orbit/accuracyTreecode", ptrConfigOrbit->accuracyTreecode);
    configWriter.setValue("Orbit/treecodeSymmetrizeTimestep", ptrConfigOrbit->treecodeSymmetrizeTimestep);
    configWriter.setValue("Orbit/accuracyFixedTimestep", ptrConfigOrbit->accuracyFixedTimestep);
    configWriter.setValue("Orbit/relaxationRate", ptrConfigOrbit->relaxRate);
}

}; // namespace