/** \file    configfile.h
    \brief   Reading/writing of INI files
    \author  EV
    \date    2014

    This file defines the class for reading and writing parameters from/to INI text files.
    The file format is identical to MS .ini file: 

    [Section]
    key=value
    ;key=old value -- commented out, lines starting with ; # / are ignored but preserved in the file upon saving

    In addition, several convenience function are provided to read and write entire groups 
    of SMILE parameters related to potential and orbit integration.
*/

#pragma once
#include <string>
#include <vector>

namespace smile{

class CConfigFile {
public:
    CConfigFile(const std::string& fileName);
    ~CConfigFile();
    
    std::string getValueString(const std::string& key, const std::string& defaultValue="") const;
    int getValueInt(const std::string& key, int defaultValue=0) const;
    double getValueDouble(const std::string& key, double defaultValue=0) const;
    bool getValueBool(const std::string& key, bool defaultValue=false) const;
    void setValue(const std::string& key, int value);
    void setValue(const std::string& key, unsigned int value);
    void setValue(const std::string& key, double value);
    void setValue(const std::string& key, double value, unsigned int width);
    void setValue(const std::string& key, const std::string& value);
    void setValue(const std::string& key, bool value);
    void listGroups(std::vector<std::string> *list) const;
private:
    void* pImpl;
};

struct CConfigPotential;  // forward declaration
struct CConfigOrbit;      // forward declaration

void readConfigPotential (const CConfigFile& configReader, CConfigPotential* ptrConfigPotential, unsigned int potGroup=0);
void writeConfigPotential(CConfigFile& configWriter, const CConfigPotential* ptrConfigPotential, unsigned int potGroup=0);
void readConfigOrbit (const CConfigFile& configReader, CConfigOrbit* ptrConfigOrbit);
void writeConfigOrbit(CConfigFile& configWriter, const CConfigOrbit* ptrConfigOrbit);

}; // namespace