#include "createpotential.h"
#include "iopotential.h"
#include "utils.h"

namespace smile {

const CPotential* initPotentialNB(CConfigPotential* config)
{
    if(config->NbodyFile=="") {
        my_message(FUNCNAME, "Empty file name");
        return NULL;  // will be initialized by default potential
    }
    return readPotential(config);
}

const CPotential* createPotential(CConfigPotential* config)
{
    const CPotential* potential=NULL;
    switch(config->PotentialType)
    {
    case CDensity::PT_LOG:
        potential=new CPotentialLog(config->q, config->p, config->scalerad);
        break;
    case CDensity::PT_HARMONIC:
        potential=new CPotentialHarmonic(config->q, config->p);
        break;
    case CDensity::PT_DEHNEN:
        potential=new CPotentialDehnen(config->mass, config->scalerad, 
            config->q, config->p, config->Gamma);
        break;
    case CDensity::PT_FERRERS:
        potential = new CPotentialFerrers(config->mass, config->scalerad, config->q, config->p); 
        break;
    case CDensity::PT_NB:
    case CDensity::PT_SPHERICAL:
        potential=initPotentialNB(config);
        break;
    case CDensity::PT_SPLINE:
    case CDensity::PT_BSE:
    case CDensity::PT_BSECOMPACT:
        // two possible variants: either init from analytic density model or from Nbody/coef/massmodel file
        if( config->DensityType==CDensity::PT_NB || 
            config->DensityType==CDensity::PT_ELLIPSOIDAL || 
            config->DensityType==CDensity::PT_MGE || 
            config->DensityType==CDensity::PT_COEFS )
        {
            potential=initPotentialNB(config);
        } else {   
            // create temporary instance of analytic density profile, 
            // pass it to the constructor to compute expansion coefficients
            const CDensity* densModel=NULL;
            switch(config->DensityType) 
            {
            case CDensity::PT_DEHNEN: 
                densModel = new CPotentialDehnen(config->mass, config->scalerad, config->q, config->p, config->Gamma); 
                break;
            case CDensity::PT_PLUMMER:
                densModel = new CDensityPlummer(config->mass, config->scalerad, config->q, config->p);
                break;
            case CDensity::PT_ISOCHRONE:
                densModel = new CDensityIsochrone(config->mass, config->scalerad, config->q, config->p);
                break;
            case CDensity::PT_PERFECTELLIPSOID:
                densModel = new CDensityPerfectEllipsoid(config->mass, config->scalerad, config->q, config->p); 
                break;
            case CDensity::PT_NFW: {
                double concentration = std::max<double>(config->scalerad2/config->scalerad, 1.0);
                densModel = new CDensityNFW(config->mass, config->scalerad, config->q, config->p, concentration);
                break; }
            case CDensity::PT_SERSIC:
                densModel = new CDensitySersic(config->mass, config->scalerad, config->q, config->p, config->sersicIndex);
                break;
            case CDensity::PT_FERRERS:
                densModel = new CPotentialFerrers(config->mass, config->scalerad, config->q, config->p); 
                break;
            default: densModel=NULL;
            }
            if(densModel!=NULL)  
            {
                if(config->PotentialType==CPotential::PT_BSE)
                    potential = new CPotentialBSE(
                        config->Alpha, config->Ncoefs_radial, config->Ncoefs_angular, densModel);
                else if(config->PotentialType==CPotential::PT_BSECOMPACT)
                    potential = new CPotentialBSECompact(
                        config->Rmax, config->Ncoefs_radial, config->Ncoefs_angular, densModel);
                else if(config->PotentialType==CPotential::PT_SPLINE) {
                    if(config->splineRadiusMin>0 && config->splineRadiusMax>0)
                    {
                        vectord grid;
                        createNonuniformGrid(config->Ncoefs_radial+1, 
                            config->splineRadiusMin, config->splineRadiusMax, true, &grid);
                        potential = new CPotentialSpline(
                            config->Ncoefs_radial, config->Ncoefs_angular, densModel, &grid);
                    } else  // no user-supplied grid extent -- will be assigned by defaule
                        potential = new CPotentialSpline(
                            config->Ncoefs_radial, config->Ncoefs_angular, densModel);
                }
                else 
                    my_error(FUNCNAME, "Unknown type of potential expansion");
                delete densModel;
            }
        }
        break;
    default: ;
    }
    if(!potential)
        return NULL;  // signal of error
    config->PotentialType = potential->PotentialType();
    config->SymmetryType  = potential->symmetry();
    if(config->DensityType==CPotential::PT_UNKNOWN)
        config->DensityType=CPotential::PT_DEHNEN;
#ifdef DEBUGPRINT
    my_message(FUNCNAME, 
        "Gamma="+convertToString(potential->getGamma())+", total mass="+convertToString(potential->totalMass()));
#endif
    return potential;
}

// helper function that creates an equivalent spherical mass model for the given density profile; returns NULL on fail
CMassModel* createMassModel(const CDensity* density, int numNodes, double innerShellMass, double outerShellMass, const CMassModel* poten)
{
    if(density->PotentialType()==CPotential::PT_SPHERICAL) {
        return new CMassModel(*(static_cast<const CPotentialSpherical*>(density)->getModel()));
    }
    double Mbh=density->Mass(0);
    double totalMass=density->totalMass()-Mbh;
    if(innerShellMass==0) innerShellMass=totalMass*0.1/numNodes;
    if(outerShellMass==0) outerShellMass=totalMass*(1-0.1/numNodes);
    if(Mbh>0) innerShellMass=std::min<double>(innerShellMass, Mbh/10);
    double rout = totalMass<0 ? 10 : density->getRadiusByMass(outerShellMass+Mbh);   // totalMass<0 should not really happen! means that mass is infinite
    double rin = std::min<double>(0.1/numNodes, density->getRadiusByMass(innerShellMass+Mbh));  // somewhat arbitrary choice for min/max radii, but probably reasonable
    if(rin<0 && totalMass<0) rin=rout/numNodes;   // trying to create a grid for an infinite-mass model.. silly enough but we could try
    if(rout<0 || rin<0 || rout<rin) return NULL;  // some weird error
    if(Mbh>0) rin=std::min<double>(rin, Mbh/10);
    vectord radii(numNodes), inmass(numNodes);
    createNonuniformGrid(numNodes/*-1*/, rin, rout, true, &radii); 
    //radii.push_back(rout*10);  // !!disabled because it produces non-monotonic mass for finite-extent models such as Ferrers
    for(int i=0; i<numNodes; i++)
        inmass[i]=density->Mass(radii[i]);
    // ensure that mass strictly increases with radius
    for(int i=1; i<numNodes; i++)
        if(inmass[i]<=inmass[i-1])
        {
            radii.erase(radii.begin()+i);
            inmass.erase(inmass.begin()+i);
            numNodes--;
            i--;
        }
    CMassModel* massModel=new CMassModel(radii, inmass, poten);
    if(massModel->getStatus()!=CMassModel::MS_OK && massModel->getStatus()!=CMassModel::MS_DF_NEGATIVE)
    { delete massModel; return NULL; }
    return massModel;
}

}; // namespace