/** \file    createpotential.h
    \brief   Creation of CPotential and CMassModel instances
    \author  EV
    \date    2010-2014

    This file provides two utility function: 
    one for creating an instance of CPotential from provided parameters in CConfigPotential variable, 
    and the other for creating a spherical mass model that approximates the given density model.
*/

#pragma once
#include "potential.h"
#include "massmodel.h"

namespace smile {

/** create an instance of CPotential according to the passed settings. 
    \param[in,out] config specifies the potential parameters, which could be modified, 
         e.g. if the potential coefficients are loaded from a file.
         Massive black hole (config->Mbh) is not included in the potential 
         (the returned potential is always non-composite)
    \return the instance of potential, or NULL in case of failure
*/
const CPotential* createPotential(CConfigPotential* config);

/** helper function that creates an equivalent spherical mass model for the given density profile. 
    \param[in] density is the non-spherical density model to be approximated
    \param[in] poten (optional) another spherical mass model that provides the potential 
         (if it is not given self-consistently by this density profile).
    \param[in] numNodes specifies the number of radial grid points in the M(r) profile
    \param[in] innerShellMass is the fraction of mass enclosed by the innermost grid node in radius
    \param[in] outerShellMass is the fraction of mass enclosed by the outermost grid node in radius
    \return new instance of CMassModel on success, or NULL on fail (e.g. if the density model was infinite) 
*/
CMassModel* createMassModel(const CDensity* density, 
    int numNodes=50, double innerShellMass=0, double outerShellMass=0, const CMassModel* poten=NULL);

}; // namespace