#include "iopotential.h"
#include <fstream>
#include <cassert>
#include "potential.h"
#include "iosnapshot.h"
#include "massmodel.h"
#include "utils.h"

namespace smile{

/* ----- reading potential from a text or snapshot file ---- */

const CPotential* readPotential(CConfigPotential *configPotential)
{
    if( configPotential==NULL ||  
      !(configPotential->PotentialType == CPotential::PT_SPHERICAL ||
        configPotential->PotentialType == CPotential::PT_BSE ||
        configPotential->PotentialType == CPotential::PT_BSECOMPACT ||
        configPotential->PotentialType == CPotential::PT_SPLINE ||
        configPotential->PotentialType == CPotential::PT_UNKNOWN ) ) 
        return NULL;   // only may load these types of potential; unknown means try reading coefs for SHE/BSE potential
    const std::string& fileName=configPotential->NbodyFile;
    std::ifstream strmText(fileName.c_str(), std::ios::in);
    if(!strmText) {
        my_error(FUNCNAME, "Cannot read from file "+fileName);
        return NULL;   // file is not readable
    }
    std::string buffer;
    std::vector<std::string> fields;
    // check header (in case we are loading Coefs file but densityType is set to something else)
    if(std::getline(strmText, buffer) && buffer.size()<256) {  // to avoid parsing a binary file as a text
        splitString(buffer, "# \t", &fields);
        if( fields[0] == "BSEcoefs" || fields[0] == "BSECcoefs" || fields[0] == "SHEcoefs" || 
            fields[0] == "SFcoefs" || fields[0] == "CylSpline" )
            configPotential->DensityType = CPotential::PT_COEFS;
    }
    strmText.seekg(0, std::ios_base::beg);  // back to start
    if(configPotential->PotentialType == CPotential::PT_SPHERICAL &&
        configPotential->DensityType == CPotential::PT_COEFS)
    {   // attempt to read r,M(r) pairs from file
        vectord radii, inmass;
        while(std::getline(strmText, buffer) && !strmText.eof())
        {
            splitString(buffer, "# \t", &fields);
            if(fields.size()>=2 && ((fields[0][0]>='0' && fields[0][0]<='9') || fields[0][0]=='-' || fields[0][0]=='+'))
            {
                double rad=convertTo<double>(fields[0]);
                double mass=convertTo<double>(fields[1]);
                /*if(radii.size()==0 && configPotential->Mbh>0 && rad>0) {  // need to add 0th element
                    radii.push_back(0);
                    inmass.push_back(configPotential->Mbh);
                }*/
                radii.push_back(rad);
                inmass.push_back(mass /*+ configPotential->Mbh*/);
            }
        }
        const CPotentialSpherical* pot=new CPotentialSpherical(radii, inmass);
        CMassModel::MODELSTATUS status=pot->getModel()->getStatus();
        if(status==CMassModel::MS_OK || status==CMassModel::MS_DF_NEGATIVE)  // ok
            return pot;
        else {
            std::string errorstr;
            switch(status)
            {
            case CMassModel::MS_ARRAY_NOT_EQ:  errorstr="Invalid format of input table file (columns of unequal length)"; break;
            case CMassModel::MS_ARRAY_SHORT:   errorstr="Too few points"; break;
            case CMassModel::MS_RADII_NOT_ASC: errorstr="Radii not in ascending order"; break;
            case CMassModel::MS_MASS_NOT_ASC:  errorstr="Mass not increasing with radius"; break;
            default: ;  // do nothing
            }
            my_error(FUNCNAME, "Error loading spherical mass profile from file "+fileName+":\n"+errorstr);
            delete pot;
            return NULL;
        }
    } else if(configPotential->PotentialType != CPotential::PT_NB &&
        configPotential->DensityType == CPotential::PT_COEFS)
    {   // attempt to load coefficients stored in a text file
        bool valid=true;
        if(!std::getline(strmText, buffer)) valid=false;
        splitString(buffer, "# \t", &fields);
        CDensity::POTENTIALTYPE ptype = 
            fields[0] == "BSEcoefs" ? CDensity::PT_BSE :
            fields[0] == "BSECcoefs"? CDensity::PT_BSECOMPACT :
            fields[0] == "SHEcoefs" ? CDensity::PT_SPLINE : CDensity::PT_UNKNOWN;
        if(valid && ptype!=CDensity::PT_UNKNOWN)
        {   // process file
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            unsigned int ncoefsRadial=convertTo<int>(fields[0]);
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            unsigned int ncoefsAngular=convertTo<int>(fields[0]);
            if(!std::getline(strmText, buffer)) valid=false;
            splitString(buffer, "# \t", &fields);
            double param=convertTo<double>(fields[0]);   // meaning of this parameter depends on potential type
            if(/*ncoefsRadial==0 || ncoefsAngular==0 || */  // zero values are possible, means just a single term in expansion
                (ptype == CDensity::PT_BSE && param<0.5) || 
                (ptype == CDensity::PT_BSECOMPACT && param<=0) || 
                (ptype == CDensity::PT_SPLINE && ncoefsRadial<4) ) valid=false;
            std::vector< vectord > coefs;
            std::vector< double > radii;
            if(ncoefsRadial>MAX_NCOEFS_RADIAL) ncoefsRadial=MAX_NCOEFS_RADIAL;
            if(ncoefsAngular>MAX_NCOEFS_ANGULAR) ncoefsAngular=MAX_NCOEFS_ANGULAR;
            while(valid && std::getline(strmText, buffer))  // time, ignored
            {
                std::getline(strmText, buffer);  // comments, ignored
                radii.clear();
                coefs.clear();
                for(unsigned int n=0; valid && n<=ncoefsRadial; n++)
                {
                    std::getline(strmText, buffer);
                    splitString(buffer, "# \t", &fields);
                    radii.push_back(convertTo<double>(fields[0]));
                    if(((ptype == CDensity::PT_BSE || ptype == CDensity::PT_BSECOMPACT) && radii.back()!=n) || 
                        (ptype == CDensity::PT_SPLINE && n>0 && radii.back()<=radii[n-1])) 
                        valid=false;  // for BSE this field is basis function index, for spline the radii should be in increasing order
                    coefs.push_back( vectord() );
                    for(int l=0; l<=static_cast<int>(ncoefsAngular); l++)
                        for(int m=-l; m<=l; m++)
                        {
                            unsigned int fi=1+l*(l+1)+m;
                            coefs.back().push_back( fi<fields.size() ? convertTo<double>(fields[fi]) : 0);
                        }
                }
            }
            if(!valid || (configPotential->PotentialType != CPotential::PT_UNKNOWN && ptype!=configPotential->PotentialType))
            {   // load coefs only if potential type is the same as requested, or if request is not specific
                my_error(FUNCNAME, "Error loading potential "+PotentialNames[configPotential->PotentialType]+
                    " coefs from file "+fileName);
                return NULL;
            }
            CPotential* pot=NULL;
            switch(ptype)
            {
            case CDensity::PT_BSE: 
                configPotential->Alpha=param;
                pot=new CPotentialBSE(/*Alpha*/param, coefs); 
                break;
            case CDensity::PT_BSECOMPACT: 
                configPotential->Rmax=param;
                pot=new CPotentialBSECompact(/*Rmax*/param, coefs); 
                break;
            case CDensity::PT_SPLINE: 
                pot=new CPotentialSpline(radii, coefs); 
                break;
            default:  return NULL;  // shouldn't happen
            }
            configPotential->Ncoefs_radial=ncoefsRadial;    
            configPotential->Ncoefs_angular=ncoefsAngular;
            configPotential->SymmetryType=pot->symmetry();
            return pot;
        } 
        else {   // file did not contain a valid header
            my_error(FUNCNAME, "Error loading potential "+PotentialNames[configPotential->PotentialType]+
                " coefs from file "+fileName);
            return NULL;
        }
    }   // PT_COEFS
    else if((configPotential->PotentialType == CPotential::PT_SPLINE ||
        configPotential->PotentialType == CPotential::PT_BSE ||
        configPotential->PotentialType == CPotential::PT_BSECOMPACT) &&
        (configPotential->DensityType == CPotential::PT_ELLIPSOIDAL ||
         configPotential->DensityType == CPotential::PT_MGE))
    {   // attempt to load ellipsoidal or MGE mass model from a text file
        const CDensity* dens=NULL;
        std::getline(strmText, buffer);
        if(buffer.substr(0, 11)=="Ellipsoidal") {
            vectord radii, inmass, axesradii, axesq, axesp;  // for loading Ellipsoidal mass profile
            while(std::getline(strmText, buffer) && !strmText.eof())
            {
                splitString(buffer, "# \t", &fields);
                if(fields.size()>=2 && ((fields[0][0]>='0' && fields[0][0]<='9') || fields[0][0]=='-' || fields[0][0]=='+'))
                {
                    radii.push_back(convertTo<double>(fields[0]));
                    inmass.push_back(convertTo<double>(fields[1]));
                    if(fields.size()>=4)  // also has anisotropy parameters
                    {
                        axesradii.push_back(radii.back());
                        axesq.push_back(convertTo<double>(fields[2]));
                        axesp.push_back(convertTo<double>(fields[3]));
                    }
                }
            }
            if(radii.size()<=2) {
                my_error(FUNCNAME, "Error loading ellipsoidal mass profile from file "+fileName);
                return NULL;
            }
            dens=new CDensityEllipsoidal(radii, inmass, axesradii, axesq, axesp);
        } else 
        if(buffer.substr(0, 3)=="MGE") {
            CDensityMGE::vectorg data;
            while(std::getline(strmText, buffer) && !strmText.eof())
            {
                splitString(buffer, "# \t", &fields);
                if(fields.size()>=2 && ((fields[0][0]>='0' && fields[0][0]<='9') || fields[0][0]=='-' || fields[0][0]=='+'))
                {
                    data.push_back(CDensityMGE::CGaussian(
                        convertTo<double>(fields[0]), convertTo<double>(fields[1]), 
                        fields.size()>=3 ? convertTo<double>(fields[2]) : 1,
                        fields.size()>=4 ? convertTo<double>(fields[3]) : 1));
                }
            }
            if(data.size()==0) {
                my_error(FUNCNAME, "Error loading MGE mass model from file "+fileName);
                return NULL;
            }
            dens=new CDensityMGE(data);
        } else {
            my_error(FUNCNAME, "Unknown density model in file "+fileName);
            return NULL;
        }
        const CPotential* pot=NULL;
        if(configPotential->PotentialType==CDensity::PT_SPLINE) {
            if(configPotential->splineRadiusMin>0 && configPotential->splineRadiusMax>0) 
            {
                vectord grid;
                createNonuniformGrid(configPotential->Ncoefs_radial+1, configPotential->splineRadiusMin, 
                    configPotential->splineRadiusMax, true, &grid);
                pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens, &grid);
            } else
                pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens);
        }
        else if(configPotential->PotentialType==CDensity::PT_BSE)
            pot = new CPotentialBSE(configPotential->Alpha, configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens);
        else if(configPotential->PotentialType==CDensity::PT_BSECOMPACT)
            pot = new CPotentialBSE(configPotential->Rmax, configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, dens);
        delete dens;
        return pot;
    }   // PT_ELLIPSOIDAL or PT_MGE
    else 
    {   // try to load a point mass set
        CPointMassSet<double> *points=NULL;
        strmText.close();
        CBasicIOSnapshot* snapRead=createIOSnapshotRead(fileName);
        if(snapRead!=NULL)
            snapRead->readSnapshot(points);
        delete snapRead;
        if(points==NULL || points->size()==0)
        {
            my_error(FUNCNAME, "Error loading N-body snapshot for potential "+
                PotentialNames[configPotential->PotentialType]+" from file "+fileName);
            if(points!=NULL) delete points;
            return NULL;
        }
        const CPotential* pot=NULL;
        if(configPotential->PotentialType==CDensity::PT_SPLINE)
        {
            if(configPotential->splineRadiusMin>0 && configPotential->splineRadiusMax>0) 
            {
                vectord radii;
                createNonuniformGrid(configPotential->Ncoefs_radial+1, configPotential->splineRadiusMin, 
                    configPotential->splineRadiusMax, true, &radii);
                pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, 
                    *points, configPotential->SymmetryType, configPotential->splineSmoothFactor, &radii);
            } else
                pot = new CPotentialSpline(configPotential->Ncoefs_radial, configPotential->Ncoefs_angular, 
                    *points, configPotential->SymmetryType, configPotential->splineSmoothFactor);
        }
        else if(configPotential->PotentialType==CDensity::PT_BSE)
            pot = new CPotentialBSE(configPotential->Alpha, configPotential->Ncoefs_radial, 
                configPotential->Ncoefs_angular, *points, configPotential->SymmetryType);
        else if(configPotential->PotentialType==CDensity::PT_BSECOMPACT)
            pot = new CPotentialBSECompact(configPotential->Rmax, configPotential->Ncoefs_radial, 
                configPotential->Ncoefs_angular, *points, configPotential->SymmetryType);
        else if(configPotential->PotentialType==CDensity::PT_SPHERICAL) {
            if(configPotential->splineRadiusMin>0 && configPotential->splineRadiusMax>0) 
            {
                vectord radii;
                createNonuniformGrid(configPotential->Ncoefs_radial+1, configPotential->splineRadiusMin, 
                    configPotential->splineRadiusMax, true, &radii);
                pot = new CPotentialSpherical(*points, configPotential->splineSmoothFactor, &radii);
            } else
                pot = new CPotentialSpherical(*points, configPotential->splineSmoothFactor);
        }
        delete points;
        return pot;
    }
    return NULL;
}

/* ------ writing potential expansion coefficients to a text file -------- */

bool writePotential(const std::string &fileName, const CPotential *potential)
{
    if(! (potential->PotentialType()==CDensity::PT_BSE || 
        potential->PotentialType()==CDensity::PT_BSECOMPACT || 
        potential->PotentialType()==CDensity::PT_SPLINE || 
        potential->PotentialType()==CDensity::PT_SPHERICAL) )
        return false;  // export potential coefs for BSE/Spline/Scale-FreeSH potentials only
    std::ofstream strmText(fileName.c_str(), std::ios::out);
    if(!strmText) 
        return false;  // file not writable
    if(potential->PotentialType()==CDensity::PT_SPHERICAL) {
        vectord radii, mass;
        static_cast<const CPotentialSpherical*>(potential)->getModel()->getMassProfile(&radii, &mass);
        strmText << "r\tM\n";
        for(size_t i=0; i<radii.size(); i++)
            strmText << radii[i] << "\t" << mass[i] << "\n";
        if(!strmText) return false;
        return true;
    }
    vectord indices;
    std::vector< vectord > coefs;
    size_t ncoefsAngular=0;
    switch(potential->PotentialType())
    {
    case CDensity::PT_BSE: {
        const CPotentialBSE* potBSE=static_cast<const CPotentialBSE*>(potential);
        indices.resize(potBSE->getNcoefs_radial()+1);
        for(size_t i=0; i<indices.size(); i++) indices[i]=i*1.0;
        potBSE->getCoefs(&coefs);
        assert(coefs.size()==indices.size());
        ncoefsAngular=potBSE->getNcoefs_angular();
        strmText << "BSEcoefs\t#header\n" << potBSE->getNcoefs_radial() << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            potBSE->getAlpha() <<"\t#alpha\n0\t#time\n";
        strmText << "#index";
        break; }
    case CDensity::PT_BSECOMPACT: {
        const CPotentialBSECompact* potBSE=static_cast<const CPotentialBSECompact*>(potential);
        indices.resize(potBSE->getNcoefs_radial()+1);
        for(size_t i=0; i<indices.size(); i++) indices[i]=i*1.0;
        potBSE->getCoefs(&coefs);
        assert(coefs.size()==indices.size());
        ncoefsAngular=potBSE->getNcoefs_angular();
        strmText << "BSECcoefs\t#header\n" << potBSE->getNcoefs_radial() << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            potBSE->getRmax() <<"\t#Rmax\n0\t#time\n";
        strmText << "#index";
        break; }
    case CDensity::PT_SPLINE: {
        const CPotentialSpline* potSpline=static_cast<const CPotentialSpline*>(potential);
        potSpline->getCoefs(&indices, &coefs);
        assert(coefs.size()==indices.size());
        assert(indices[0]==0);  // leftmost radius is 0
        coefs[0].resize(1);     // retain only l=0 term for r=0, the rest is supposed to be zero
        ncoefsAngular=potSpline->getNcoefs_angular();
        strmText << "SHEcoefs\t#header\n" << potSpline->getNcoefs_radial() << "\t#n_radial\n" << ncoefsAngular << "\t#n_angular\n" << 
            0 <<"\t#unused\n0\t#time\n";
        strmText << "#radius";
        break; }
    default:
        my_error(FUNCNAME, "wrong potential type");
        return false;   // shouldn't occur, we checked potential type in the beginning
    }
    for(int l=0; l<=static_cast<int>(ncoefsAngular); l++)
        for(int m=-l; m<=l; m++)
            strmText << "\tl="<<l<<",m="<<m;  // header line
    strmText << "\n";
    for(size_t n=0; n<indices.size(); n++)
    {
        strmText << indices[n];
        strmText << "\t" << std::setprecision(14) << coefs[n][0] << std::setprecision(7);   // leading coeft should be high-accuracy at least for spline potential
        for(size_t i=1; i<coefs[n].size(); i++)
            strmText << "\t" << coefs[n][i];
        strmText << "\n";
    }
    if(!strmText) return false;
    return true;
}

}  //namespace