/** \file    iopotential.h
    \brief   Input/output of data for constructing potentials
    \author  EV
    \date    2010-2013

    Two functions are provided, for creating a potential from a variety of input data stored 
    in a file (potential expansion coefficients, an N-body snapshot, or a density model), 
    and for writing expansion coefficients to a text file.
*/

#pragma once
#include "common.h"

namespace smile{

struct CConfigPotential;   // forward declaration

/** load a potential from a text or snapshot file.

    The input file may contain one of the following kinds of data:
    - a Nbody snapshot in a text or binary format, handled by classes derived from CBasicIOSnapshot;
    - a potential coefficients file for CPotentialBSE or CPotentialSpline;
    - a density model described by CDensityEllipsoidal or CDensityMGE.
    The data format is determined from the first line of the file, and 
    if it is allowed by the parameters passed in configPotential, then 
    the file is read and the instance of a corresponding potential is created. 
    If the input data was not the potential coefficients and the new potential 
    is of BSE or Spline type, then a new file with potential coefficients is 
    created and written via writePotential(), so that later one may load this 
    coef file instead of the original one, which speeds up initialization.
    \param[in,out] configPotential contains the potential parameters and may be updated 
                   upon reading the file (e.g. the number of expansion coefficients 
                   may change). If the file doesn't contain appropriate kind of potential
                   (i.e. if configPotential->PotentialType is PT_NB but the file contains 
                   BSE coefficients or a description of MGE model), an error is returned.
                   configPotential->NbodyFile contains the file name from which the data is loaded.
    \return a new instance of CPotential on success, or NULL on failure (display an error 
                   message by calling my_error() in that case).
*/
const CPotential* readPotential(CConfigPotential *configPotential);

/** writes potential expansion coefficients to a text file.

    The potential must be one of the following: CPotentialBSE, CPotentialSpline, CPotentialSpherical,
    The coefficients stored in a file may be later loaded by readPotential() function.
    \param[in] fileName is the output file
    \param[in] potential is the pointer to potential
    \return    true on success, false on failure (file not writeable or potential is of inappropriate type).
*/
bool writePotential(const std::string &fileName, const CPotential* potential);
};  //namespace