#include "massmodel.h"
#include <cmath>
#include <algorithm>
#include <gsl/gsl_integration.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_errno.h>
#include "utils.h"

namespace smile{

//----------------------------------------------------------------------------//
inline double pow_2(const double x) { return x*x; }

void CMassModel::getMassProfile(std::vector<double> *radii, std::vector<double> *inmass) const
{
    if(radii==NULL || inmass==NULL) return;
    radii->resize(spl_loginmass->size+1);
    inmass->resize(spl_loginmass->size+1);
    radii->at(0)=0;
    inmass->at(0)=masscenter;
    for(size_t i=0; i<spl_loginmass->size; i++) {
        radii->at(i+1)=exp(spl_loginmass->x[i]);
        inmass->at(i+1)=mass(radii->at(i+1));
    }
}

double CMassModel::scaledE(double E) const
{
    return !totalmasscenter ? log(1/potentialcenter-1/E) : log(-1/E);
}

double CMassModel::unscaledE(double scaledE) const
{
    return !totalmasscenter ? 1/(1/potentialcenter-exp(scaledE)) : -exp(-scaledE);
}

double CMassModel::scaledEder(double E) const
{
    return 1/E/(!totalmasscenter ? E/potentialcenter-1 : -1);
}

double CMassModel::computeSplineInScaledEnergy(double E, const gsl_spline* spl) const
{
    double S = scaledE(E);
    //if(gsl_isinf(S)) return 0;
    return gsl_spline_evalx(spl, S, NULL);
}

void CMassModel::computeSplineDerivInScaledEnergy(double E, const gsl_spline* spl, 
    double* value, double* deriv, double* deriv2) const
{
    double S = scaledE(E);
    /*if(gsl_isinf(S)) {
        if(value!=NULL) *value=0;
        if(deriv!=NULL) *deriv=0;
        if(deriv2!=NULL) *deriv2=0;
        return;
    }*/
    if(value!=NULL) *value=gsl_spline_evalx(spl, S, NULL);
    if(deriv!=NULL || deriv2!=NULL) {
        double dSdE=scaledEder(E);
        double dVdS=gsl_spline_eval_derivx(spl, S, NULL);
        if(deriv!=NULL) *deriv=dVdS * dSdE;
        if(deriv2!=NULL) *deriv2=pow_2(dSdE)*(gsl_spline_eval_deriv2x(spl, S, NULL) -
            dVdS*( !totalmasscenter ? 2*E/potentialcenter-1 : -1) );
    }
}

gsl_spline* CMassModel::initSplineInScaledEnergy(const std::vector<double> &initE, const std::vector<double> &initVal)
{
    gsl_spline* spl=NULL;
    size_t npoints=initE.size();
    if(npoints==initVal.size())
        spl = gsl_spline_alloc(gsl_interp_cspline, npoints);
    else if(npoints+2==initVal.size())
        spl = gsl_spline_alloc(gsl_interp_ccspline, npoints);
    else return NULL;
    std::vector<double> scE(npoints);
    for(size_t i=0; i<npoints; i++)
        scE[i]=scaledE(initE[i]);
    gsl_spline_init(spl, &(scE.front()), &(initVal.front()), npoints);
    return spl;
}

// interpolated mass inside radius
double CMassModel::mass(double r) const
{
    if(r==0) return masscenter;
    if(r>=rout)  // special treatment for extrapolation beyond the last grid point
        return totalmass - coefmrout*pow(r, 3-gammaout);
    double mu = exp(gsl_spline_evalx(spl_loginmass, log(r), NULL));
    return (mu*totalmass + masscenter)/(mu+1);
}

// interpolated radius as a function of enclosed mass
double CMassModel::rad(double m) const
{
    if(m<=masscenter) return 0;
    if(m>=totalmass) return gsl_posinf();
    return exp(gsl_spline_evalx(spl_logrmass, log( (m-masscenter)/(totalmass-m) ), NULL));
}

// interpolated density at radius
double CMassModel::dens(double r) const
{
    if(densconst && r<rin) return rho0*(1-coefdrhodr*pow(r, alpha));
    if(r>rout) return (gammaout-3)/(4*M_PI)*coefmrout*pow(r, -gammaout);
    double inm=mass(r);
    return gsl_spline_eval_derivx(spl_loginmass, log(r), NULL) * 
        (inm-masscenter) * (totalmass-inm) / (totalmass-masscenter) / (4*M_PI*pow(r, 3.0));
}

// interpolated potential at radius
double CMassModel::pot(double r) const
{
    if(r==0) return !totalmasscenter ? potentialcenter : gsl_neginf();
    if(r<=rin)
    {
        double phi1=potentialin+totalmasscenter/rin;
        return potentialcenter + (phi1-potentialcenter)*pow(r/rin, 2-gammain) - totalmasscenter/r;
    }
    return unscaledE(gsl_spline_evalx(spl_logpotential, log(r), NULL));
}

// interpolated distribution function of energy, given by pre-computed spline via 2nd derivative of rho(Phi)
double CMassModel::distrf(double E) const
{
    if(gsl_isinf(scaledE(E))) return 0;  // avoid indeterminate values in other calculations involving f
    return exp(computeSplineInScaledEnergy(E, spl_logdistrf));
}

// interpolated differential energy distribution (obtained as d(phasevol)/dE )
double CMassModel::ded(double E) const
{
    if(E==pot(0)) return 0;
    double value, deriv;
    computeSplineDerivInScaledEnergy(E, spl_logphasevol, &value, &deriv, NULL);
    return exp(value)*deriv;
}

// interpolated phase volume
double CMassModel::phasevol(double E) const
{
    return exp(computeSplineInScaledEnergy(E, spl_logphasevol));
}

// interpolated radial period at energy
double CMassModel::trad(double E) const
{
    return exp(computeSplineInScaledEnergy(E, spl_logtrad));
}

// epicyclic period for a circular orbit with given energy
double CMassModel::tepi(double E) const
{
    double rc=rcirc(E);
    return 2*M_PI/sqrt(mass(rc)/pow(rc,3.0) + 4*M_PI*dens(rc));
}

// interpolated radius of circular orbit at energy
double CMassModel::rcirc(double E) const
{
    return exp(computeSplineInScaledEnergy(E, spl_logrcirc));
}

// interpolated angular momentum of circular orbit at energy
double CMassModel::lcirc(double E) const
{
    double rc=rcirc(E);
    return sqrt(2*(E-pot(rc)))*rc;
}

// compute mass of particles having energy <E
double CMassModel::masse(double E) const
{
    if(spl_logfgint==NULL || E<=pot(0) || E>=0) return 0;
    return exp(computeSplineInScaledEnergy(E, spl_logfgint));
}

/* --- helper functions for integration and root-finding --- */

/// \cond INTERNAL_DOCS
// helper structure keeping params for GSL call-back functions
typedef double (CMassModel::*memberFnc)(double x) const;
struct CHelperParam{
    const CMassModel* model;  // pointer to the model in question
    double n,m,v,w,a,L2,q2,q3;// parameters for general integrands and root-finders
    double p2,p3,r1,r2,r3;    // for constant-density core estimator
    gsl_spline *spl;          // parameter for interpolation routine
    memberFnc memberFunction; // pointer to a member function of model used in general integrands
    bool exact;               // whether to perform exact integration or table interpolation for diffusion coefs
    CHelperParam() { model=NULL; memberFunction=NULL; n=m=v=w=0; a=1;}
};
/// \endcond

// generic integrand:  return a * |func(x)-v|^m * |x-w|^n
double int_generic(double x, void* params)
{
    double mult1=1;
    if(((CHelperParam*)params)->memberFunction!=NULL) {
        memberFnc fnc=((CHelperParam*)params)->memberFunction;
        const CMassModel& model=*(((CHelperParam*)params)->model);
        double result = (model.*fnc)(x) - ((CHelperParam*)params)->v;
        mult1=pow( fabs(result), ((CHelperParam*)params)->m);
    }
    double mult2=1;
    if(((CHelperParam*)params)->n!=0 && x!=((CHelperParam*)params)->w)
        mult2=pow(fabs(x-((CHelperParam*)params)->w), ((CHelperParam*)params)->n);
    return ((CHelperParam*)params)->a*mult1*mult2;
}

// integrand for potential:  x=r, fnc=mass(r), v=mass(0), m=1, n=-2
// integrand for Eddington inversion formula, using second derivative of density as a function of potential:  x=Phi, fnc=d2rhodphi2(Phi), w=E, n=-1/2
// integrand for finding radial period of purely radial orbit: x=r, fnc=pot, v=E, m=-1/2, n=0
// integrand for phase volume: x=r, fnc=pot, v=E, n=2, m=3/2

// generic integrand with distribution function:  return f(E) * |func(E)-v|^m * |x|^n
double int_generic_f(double x, void* params)
{
    const CMassModel& model=*(((CHelperParam*)params)->model);
    double mult1=1;
    if(((CHelperParam*)params)->memberFunction!=NULL) {
        memberFnc fnc=((CHelperParam*)params)->memberFunction;
        double membFncResult = (model.*fnc)(x) - ((CHelperParam*)params)->v;
        mult1=pow(fabs(membFncResult), ((CHelperParam*)params)->m);
    }
    double mult2=1;
    if(((CHelperParam*)params)->n!=0)
        mult2=pow(fabs(x), ((CHelperParam*)params)->n);
    double res= mult1*mult2*model.distrf(x);
    if(gsl_isinf(res)||gsl_isnan(res)) { 
        res=0; // safety measure, but might conceal errors!
    }
    return res;
}
// integrand for f*g: fnc=ded, m=1, n=0
// integrand for projected density
double int_densproj(double z, void* params)
{
    return ((CHelperParam*)params)->model->dens(sqrt(pow_2(z)+pow_2(((CHelperParam*)params)->r1)));
}
// integrand for line-of-sight velocity dispersion
double int_losvd(double z, void* params)
{
    double r=sqrt(pow_2(z)+pow_2(((CHelperParam*)params)->r1));
    return ((CHelperParam*)params)->model->dens(r)*((CHelperParam*)params)->model->mass(r)*z*z/pow(r,3);
}
// generic root-finder:  fnc(x) - v
double find_generic(double x, void* params)
{
    const CMassModel& model=*(((CHelperParam*)params)->model);
    memberFnc fnc=((CHelperParam*)params)->memberFunction;
    return (model.*fnc)(x) - ((CHelperParam*)params)->v;
}
// helper root-finder for a constant-density core
double find_alpha(double a, void* params)
{
    return (1-((CHelperParam*)params)->p2)*(pow(((CHelperParam*)params)->r3,a)-((CHelperParam*)params)->p3*pow(((CHelperParam*)params)->r1,a)) 
         - (1-((CHelperParam*)params)->p3)*(pow(((CHelperParam*)params)->r2,a)-((CHelperParam*)params)->p2*pow(((CHelperParam*)params)->r1,a));
}
// helper root-finder for circular radius at energy
double find_rcirc(double r, void* params)
{
    return ((CHelperParam*)params)->model->pot(r) + 
        ((CHelperParam*)params)->model->mass(r)/2/r - ((CHelperParam*)params)->v;
}
// helper root-finder for r_min,r_max(E,L);  E=params->v and L^2=params->L2
double find_vrad(double r, void* params)
{
    return 2*(((CHelperParam*)params)->v - ((CHelperParam*)params)->model->pot(r)) -
        ((CHelperParam*)params)->L2/(r*r);
}
// helper function for finding outer density slope
double find_gamma_out(double y, void* params)
{
    return( pow(((CHelperParam*)params)->r2, 3-y) - pow(((CHelperParam*)params)->r1, 3-y))
        / ( pow(((CHelperParam*)params)->r3, 3-y) - pow(((CHelperParam*)params)->r1, 3-y))
        - ((CHelperParam*)params)->m;
}

// rmax: r(Phi)
double CMassModel::rmax(double phi) const
{
    CHelperParam params;
    params.model=this;
    params.memberFunction=&CMassModel::pot;
    params.v=phi;
    gsl_function F;
    F.function=&find_generic;
    F.params=&params;
    return findRoot(&F, -totalmasscenter/phi, -totalmass/phi*1.2); // safety margin?
}

// interpolated d(rho)/d(phi) at given phi
double CMassModel::drhodphi(double phi) const
{
    if(phi<potentialin) {
        if(densconst) return coefdrhodphi*pow(phi-potentialcenter, alpha/2-1);
        double rad=0;  // find radius corresponding to phi
        double massrin=mass(rin)-masscenter;
        if(!totalmasscenter) 
            rad=rin*pow( (2-gammain)*(phi-potentialcenter)*rin/massrin, 1./(2-gammain));
        else { // no analytical expression -- need to solve numerically
            CHelperParam params;
            params.model=this;
            params.memberFunction=&CMassModel::pot;
            params.v=phi;
            gsl_function F;
            F.function=&find_generic;
            F.params=&params;
            rad=findRoot(&F, -totalmasscenter/phi, rin);
        }
        double massr=massrin*pow(rad/rin, 3-gammain);
        double dphi1dr1=(totalmasscenter+massr)/(rad*rad);
        double drho1dr1=-gammain*(3-gammain)/(4*M_PI)*massr/pow(rad,4);
        return drho1dr1/dphi1dr1;
    }
    double value, deriv;
    computeSplineDerivInScaledEnergy(phi, spl_logdenspot, &value, &deriv, NULL);
    return exp(value)*deriv;
}

// interpolated d^2(rho)/d(phi)^2 at given phi
double CMassModel::d2rhodphi2(double phi) const
{
    if(phi<potentialin) {
        if(densconst) return coefdrhodphi*(alpha/2-1)*pow(phi-potentialcenter, alpha/2-2);
        double rad=0;  // find radius corresponding to phi
        double massrin=mass(rin)-masscenter;
        if(totalmasscenter==0) 
            rad=rin*pow( (2-gammain)*(phi-potentialcenter)*rin/massrin, 1./(2-gammain));
        else { // no analytical expression -- need to solve numerically
            CHelperParam params;
            params.model=this;
            params.memberFunction=&CMassModel::pot;
            params.v=phi;
            gsl_function F;
            F.function=&find_generic;
            F.params=&params;
            rad=findRoot(&F, -totalmasscenter/phi, rin);
        }
        double massr=massrin*pow(rad/rin, 3-gammain);
        double dphi1dr1=(totalmasscenter+massr)/(rad*rad);
        double dphi2dr2=(-2*totalmasscenter+massr*(1-gammain))/pow(rad,3);
        double drho1dr1=-gammain*(3-gammain)/(4*M_PI)*massr/pow(rad,4);
        double drho2dr2=-(gammain+1)*drho1dr1/rad;
        return (drho2dr2*dphi1dr1-drho1dr1*dphi2dr2)/pow(dphi1dr1,3);
    }
    double value, deriv, deriv2;
    computeSplineDerivInScaledEnergy(phi, spl_logdenspot, &value, &deriv, &deriv2);
    return exp(value)*(pow_2(deriv)+deriv2);
}

// direct evaluation of distribution function via Eddington inversion
double CMassModel::distrf2(double E) const
{
    gsl_function F;
    CHelperParam params;
    F.params=&params;
    F.function=&int_generic;
    params.memberFunction=&CMassModel::d2rhodphi2;
    params.model=this;
    params.w=E;
    params.m=1;
    params.n=-0.5;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    double result, error;
    gsl_integration_qags (&F, 0, E, 0, 1e-10, 1000, ws, &result, &error); 
    gsl_integration_workspace_free(ws);
    return -(sqrt(0.125)/M_PI/M_PI) * result;
}

// compute velocity dispersion at given radius
double CMassModel::veldisp(double rad) const
{
    if(rad==0 && totalmasscenter!=0) 
        return gsl_posinf();
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    CHelperParam params;
    params.model=this;
    params.memberFunction=&CMassModel::distrf;
    params.m=1;
    params.w=pot(rad);
    gsl_function F;
    F.function=&int_generic;
    F.params=&params;
    double result, result0, error;
    params.n=1.5;
    gsl_integration_qag(&F, params.w, 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
    params.n=0.5;
    gsl_integration_qag(&F, params.w, 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result0, &error);
    error=dens(rad);
    gsl_integration_workspace_free(ws);
    return sqrt(result/result0*2/3);
}

// compute projected density at given radius
double CMassModel::densproj(double rad) const
{
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    CHelperParam params;
    params.r1=rad;
    params.model=this;
    gsl_function F;
    F.function=&int_densproj;
    F.params=&params;
    double result, error;
    gsl_integration_qagiu(&F, 0, 0, 1e-7, 1000, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return 2*result;
}

// compute line-of-sight velocity dispersion at a given projected radius
double CMassModel::losvd(double rad) const
{
    gsl_integration_workspace* ws = gsl_integration_workspace_alloc(1000);
    CHelperParam params;
    params.r1=rad;
    params.model=this;
    gsl_function F;
    F.function=&int_losvd;
    F.params=&params;
    double result, error;
    gsl_integration_qagiu(&F, 0, 0, 1e-7, 1000, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return sqrt(2*result/densproj(rad));
}

// generate particle from f(E) at given r
double CMassModel::samplev(double radius) const
{
    if(!gsl_finite(radius)) return 0;  // infinity
    gsl_function F;
    CHelperParam params;
    F.function=&int_generic;
    F.params=&params;
    params.model=this;
    params.memberFunction=&CMassModel::distrf;
    params.w=pot(radius);
    params.n=0.5;
    params.m=1;
    params.a=-sqrt(2.0);
    // determine maximum of v^2*f(Phi+v^2/2) = minimum of -sqrt(2*(E-Phi))*f(E)
    double vesc=sqrt(-2*params.w);
    double root = findMinGuess(&F, params.w*(1-1E-10), params.w*1E-10, 1e-3);
    double fmax = -F.function(root, &params) * 1.1;
    double v, f1;
    do{ // rejection algorithm
        v=vesc*getRandomNumber();
        f1=fmax*getRandomNumber();
    } while(f1>distrf(params.w+v*v/2)*v*v);
    return v;
}

// initialization of M(r)
void CMassModel::initSplineMass(const std::vector<double> &initRadii, const std::vector<double> &initMass)
{
    int Npoints=static_cast<int>(initRadii.size()), Nskip=0;
    if(initRadii.front()<=0)
    {
        masscenter=initMass.front();
        Npoints--;
        Nskip=1; // skip first point if it is at origin
    }
    else
        masscenter=0;
    rout=initRadii.back();
    rin=initRadii[Nskip];
    gsl_function F;
    CHelperParam params;
    params.model=this;
    F.params=&params;

    // estimate outermost slope  (needed for accurate extrapolation beyond last grid point)
    totalmass=0;
    params.r1=initRadii[Npoints+Nskip-1];
    params.r2=initRadii[Npoints+Nskip-2];
    params.r3=initRadii[Npoints+Nskip-3];
    params.m = (initMass[Npoints+Nskip-1] - initMass[Npoints+Nskip-2])/(initMass[Npoints+Nskip-1] - initMass[Npoints+Nskip-3]);
    F.function=&find_gamma_out;
    double Ylow=3.1, Yupp=10;
    gammaout = findRoot(&F, Ylow, Yupp, 5.0);
    totalmass = 
        (initMass[Npoints+Nskip-1]*pow(initRadii[Npoints+Nskip-1], gammaout-3) -
         initMass[Npoints+Nskip-2]*pow(initRadii[Npoints+Nskip-2], gammaout-3)) /
        (pow(initRadii[Npoints+Nskip-1], gammaout-3) - pow(initRadii[Npoints+Nskip-2], gammaout-3));
    if(totalmass<=0)  // shouldn't happen?
        totalmass = 2*initMass.back()-initMass[Npoints+Nskip-2];  // quite arbitrary
    coefmrout=(totalmass-initMass.back())*pow(rout, gammaout-3);
    
    std::vector<double> logradii(Npoints), loginmass(Npoints);
    for(int i=0; i<Npoints; i++)
    {
        logradii[i]=log(initRadii[i+Nskip]);
        loginmass[i]=log((initMass[i+Nskip]-masscenter)/(totalmass-initMass[i+Nskip]));
    }
    // check if density tends to constant at r=0
    gammain=3. - (loginmass[1]-loginmass[0])/(logradii[1]-logradii[0]);
#ifdef DEBUGPRINT
    my_message(FUNCNAME, "Total mass = "+convertToString(totalmass)+
        ", outer density slope = "+convertToString(gammaout)+", inner = "+convertToString(gammain));
#endif
    //potfinite = masscenter==0 && gammain<2;
    if(!masscenter && gammain < 0.1)
    {   // provision for higher accuracy for constant-density core (??!!)
        densconst=true;
        gammain=0;
        params.p2=initMass[Nskip+1]/initMass[Nskip]*pow(initRadii[Nskip]/initRadii[Nskip+1],3.0);
        params.p3=initMass[Nskip+2]/initMass[Nskip]*pow(initRadii[Nskip]/initRadii[Nskip+2],3.0);
        params.r1=initRadii[Nskip];
        params.r2=initRadii[Nskip+1];
        params.r3=initRadii[Nskip+2];
        F.function = &find_alpha;
        alpha = findRoot(&F, 0.1, 5.0, 2.0);   // 2 is default for Plummer and similar types of cored potentials
        coefdrhodr = (1+alpha/3)*(1-params.p2)/(pow(initRadii[2],alpha)-params.p2*pow(initRadii[1],alpha));
        if(coefdrhodr<0) {
#ifdef DEBUGPRINT
            my_message(FUNCNAME, 
                "Inner density profile was rising with radius, forced to be constant-density core");
#endif
            coefdrhodr=0;  // density should be monotonic with radius
        }
        rho0 = initMass[Nskip]*3/(4*M_PI)/pow(initRadii[Nskip],3.0)/(1 - 3./(alpha+3)*coefdrhodr*pow(initRadii[Nskip],alpha));
        if(rho0<=0) {
#ifdef DEBUGPRINT
            my_message(FUNCNAME,
                "Central density was inferred to be negative, forced to be a constant-density core");
#endif
            coefdrhodr=0;
            rho0=initMass[Nskip]/(4*M_PI*pow(initRadii[Nskip],3.0));
        }
#ifdef DEBUGPRINT
        my_message(FUNCNAME, 
            "Constant-density core determined: rho(r) = "+convertToString(rho0)+
            "*(1-"+convertToString(coefdrhodr)+"*r^"+convertToString(alpha)+")");
#endif
        coefdrhodphi = -coefdrhodr*3/(4*M_PI)*alpha* pow(3/(2*M_PI*rho0), alpha/2-1);
        loginmass.push_back( 3*totalmass/(totalmass-initMass[Nskip])   // slope at r->0
            * (1-coefdrhodr*pow(initRadii[Nskip],alpha))/(1-coefdrhodr*pow(initRadii[Nskip],alpha)/(1+alpha/3)) );
    } else {
        densconst=false;
        if(gammain>=1.95) {
            gammain=1.95;  // do not allow slope greater than -2, otherwise potential is divergent
#ifdef DEBUGPRINT
            my_message(FUNCNAME,
                "Inner density slope was greater than 2, weird results are likely to arise");
#endif
        }
        loginmass.push_back((3-gammain)*(totalmass-masscenter)/(totalmass-initMass[Nskip]));  // slope at r->0
    }
    // init spline to compute mass and density within radius
    loginmass.push_back((gammaout-3)*totalmass/initMass.back());   // slope at r->infinity
    if(spl_loginmass!=NULL) gsl_spline_free(spl_loginmass);
    spl_loginmass = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    gsl_spline_init (spl_loginmass, &(logradii.front()), &(loginmass.front()), Npoints);
    // inverse of the above (radius as a function of mass)
    if(spl_logrmass!=NULL) gsl_spline_free(spl_logrmass);
    spl_logrmass = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    logradii.push_back(1./(3-gammain));
    logradii.push_back(1./((gammaout-3)*totalmass/initMass.back()));
    gsl_spline_init (spl_logrmass, &(loginmass.front()), &(logradii.front()), Npoints);
    if(!check_spline_monotonic(spl_loginmass)) {
        errorcode = MS_MASS_NOT_ASC;
#ifdef DEBUGPRINT
        my_message(FUNCNAME,
            "Interpolated mass is not monotonic in radius");
#endif
        return;
    }
}

// initialization of Phi(r) and other auxiliary splines
void CMassModel::initSplinePhi(const CMassModel* model_pot)
{
    size_t Npoints=spl_loginmass->size;
    gsl_function F;
    CHelperParam params;
    params.model=this;
    F.function=&int_generic;
    F.params=&params;
    double result, error;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    std::vector<double> phi(Npoints);
    std::vector<double> logpotential(Npoints);
    std::vector<double> logdens(Npoints);
    std::vector<double> logphasevol(Npoints);
    std::vector<double> logtrad(Npoints);
    std::vector<double> logrcirc(Npoints);

    if(model_pot==NULL) {   // compute potential by integration of M(r)/r^2
        totalmasscenter=masscenter;
        params.memberFunction=&CMassModel::mass;
        params.v=masscenter;
        params.m=1;
        params.n=-2;
        for(size_t i=0; i<Npoints; i++)
        {
            gsl_integration_qag (&F, i==0? 0 : exp(spl_loginmass->x[i-1]), exp(spl_loginmass->x[i]), 
                0, 1e-7, 1000, GSL_INTEG_GAUSS21, w, &result, &error); 
            phi[i]=result + (i>0?phi[i-1]:0);
        }
        //double rout=exp(spl_loginmass->x[Npoints-1]);
        gsl_integration_qagiu (&F, rout, 0, 1e-7, 1000, w, &result, &error); 
        // workaround for GSL bug in extrapolating the above integral
        result=std::max<double>(result, (mass(rout)-masscenter)/rout);
        potentialcenter = -(result+phi.back());
        for(size_t i=0; i<Npoints; i++)
            phi[i] += potentialcenter - (masscenter>0 ? masscenter/exp(spl_loginmass->x[i]) : 0);
    } else {   // use external potential
        for(size_t i=0; i<Npoints; i++)
            phi[i] = model_pot->pot(exp(spl_loginmass->x[i]));
        potentialcenter=model_pot->potentialcenter;
        totalmasscenter=model_pot->totalmasscenter;
    }
    for(size_t i=0; i<Npoints; i++)
        logpotential[i]= scaledE(phi[i]);
    potentialin=phi.front();  // store innermost point for extrapolation to r->0
    
    // init spline: potential(radius)
    if(model_pot==NULL) {
        logpotential.push_back(masscenter == 0 ? (2-gammain) :   // inner slope
            -mass(rin)/( potentialcenter*rin-masscenter + 
            (mass(rin)-masscenter)/(2-gammain) ) );
    } else {
        /*double val=gsl_spline_evalx(model_pot->spl_logpotential, log(rin), NULL);
        double der=gsl_spline_eval_derivx(model_pot->spl_logpotential, log(rin), NULL);
        logpotential.push_back(der*val*(model_pot->totalmasscenter?1:1-val/model_pot->potentialcenter)/rin);*/
        logpotential.push_back(gsl_spline_eval_derivx(model_pot->spl_logpotential, log(rin), NULL));
    }
    logpotential.push_back(1.0);   // outer slope (not quite true but ignore the first-order correction for extrapolation)
    if(spl_logpotential!=NULL) gsl_spline_free(spl_logpotential);
    spl_logpotential = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    gsl_spline_init (spl_logpotential, spl_loginmass->x, &(logpotential.front()), Npoints);  // use the same x-values as M(r)
    if(!check_spline_monotonic(spl_logpotential)) {
        errorcode = MS_MASS_NOT_ASC;   // shouldn't happen if mass is monotonic??
        gsl_integration_workspace_free(w);
        return;
    }
    
    // initspline: density(potential)
    for(size_t i=0; i<Npoints; i++)
        logdens[i]=log(dens(exp(spl_loginmass->x[i])));
    double drhodphi_in;
    //if(model_pot==NULL) {
        drhodphi_in = densconst ?
            coefdrhodphi*pow(phi.front()-potentialcenter, alpha/2-1) :
            -gammain*(3-gammain)/(4*M_PI*pow_2(rin)) * (1-masscenter/mass(rin));
    //} else {   ///!!!! doesn't work properly for external potential!!!?
    logdens.push_back( exp(-logdens.front()) * phi.front()*(!totalmasscenter ? phi.front()/potentialcenter-1 : -1) * drhodphi_in);
    double drhodphi_out= (gammaout-3)*coefmrout/(4*M_PI) *    // <- rho(r_out)*r_out^gamma_out
        -gammaout * pow(-phi.back(), gammaout-1) * pow(mass(rout), -gammaout);
    logdens.push_back( exp(-logdens[Npoints-1]) * phi.back()*(!totalmasscenter ? phi.back()/potentialcenter-1 : -1) * drhodphi_out);
    spl_logdenspot = gsl_spline_alloc (gsl_interp_ccspline, Npoints);
    gsl_spline_init (spl_logdenspot, &(logpotential.front()), &(logdens.front()), Npoints);

    // compute phase volume as a function of energy
    params.memberFunction=&CMassModel::pot;
    params.m=1.5;
    params.n=2;
    for(size_t i=0; i<Npoints; i++)
    {
        params.v=phi.at(i);
        gsl_integration_qag (&F, 0, exp(spl_loginmass->x[i]), 0, 1e-7, 1000, GSL_INTEG_GAUSS21, w, &result, &error); 
        logphasevol[i]=log(result*16*M_PI*M_PI/3*2*M_SQRT2);
    }
    if(totalmasscenter) { logphasevol.push_back(1.5); logphasevol.push_back(gsl_nan()); }
    if(spl_logphasevol!=NULL) gsl_spline_free(spl_logphasevol);
    spl_logphasevol = initSplineInScaledEnergy(phi, logphasevol);

    // compute T_rad(E) - radial period as a function of energy
    params.memberFunction=&CMassModel::pot;
    params.m=-0.5;
    params.n=0;
    for(size_t i=0; i<Npoints; i++)
    {
        params.v=phi.at(i);
        gsl_integration_qags (&F, 0, exp(spl_loginmass->x[i]), 0, 1e-7, 1000, w, &result, &error); 
        logtrad[i]=log(M_SQRT2*result);
    }
    if(spl_logtrad!=NULL) gsl_spline_free(spl_logtrad);
    spl_logtrad = initSplineInScaledEnergy(phi, logtrad);

    // compute r_circ(E) - radius of circular orbit as a function of energy
    F.function=&find_rcirc;
    for(size_t i=0; i<Npoints; i++)
    {
        params.v=phi.at(i);
        double rupp=exp(spl_loginmass->x[i]), rlow=rupp/4;
        while(find_rcirc(rlow, &params)>0) rlow/=2; //probably not necessary..
        result = findRoot(&F, rlow, rupp);
        logrcirc[i]=log(result);
    }
    if(spl_logrcirc!=NULL) gsl_spline_free(spl_logrcirc);
    spl_logrcirc = initSplineInScaledEnergy(phi, logrcirc);

    gsl_integration_workspace_free(w);
}

// initialization of distribution function in energy
void CMassModel::initEddingtonDF()
{
    gsl_function F;
    CHelperParam params;
    params.model=this;
    F.params=&params;
    double result, result1, error;
    gsl_integration_workspace * w = gsl_integration_workspace_alloc(1000);
    size_t Npoints=spl_logpotential->size;
    std::vector<double> phi, logf;
    // compute integrated DF using Eddington inversion formula with second derivative of rho(Phi)
    F.function=&int_generic;
    params.memberFunction=&CMassModel::d2rhodphi2;
    params.m=1;
    params.n=-0.5;
    for(size_t i=0; i<Npoints; i++)
    {
        params.w=unscaledE(spl_logpotential->y[i]);
        gsl_integration_qags (&F, params.w, 0, 0, 1e-7, 1000, w, &result, &error); 
        result *= (sqrt(0.125)/M_PI/M_PI);
        if(result>0 && !gsl_isinf(result) && !gsl_isnan(result)) {
            phi.push_back(params.w);
            logf.push_back(log(result));
        }
        else errorcode=MS_DF_NEGATIVE;
    }
    if(logf.size()<3) {  // pathological case, just fill in with very small (but finite and convergent) values
        phi.resize(4); 
        phi[0]=unscaledE(spl_logpotential->y[0]);
        phi[3]=unscaledE(spl_logpotential->y[Npoints-1]);
        phi[1]=(phi[0]*2+phi[2])/3;
        phi[2]=(phi[0]+phi[2]*2)/3;
        logf.resize(6);
        logf[0]=-100.0;
        logf[1]=-100.0;
        logf[3]=-100.0-(scaledE(phi[3])-scaledE(phi[2]))*3.5;
        logf[2]=(2*logf[1]+logf[3])/3;
        logf[4]=0.0;
        logf[5]=-3.5;
    } else {
        // compute endpoint derivatives for clamped spline initialization
        double E=pot(rin/5);   // inward extrapolation (r->0)
        params.w=E;
        gsl_integration_qags(&F, E, unscaledE(spl_logpotential->y[0]), 0, 1e-7, 1000, w, &result, &error); 
        gsl_integration_qag (&F, unscaledE(spl_logpotential->y[0]), 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, w, &result1, &error);
        result+=result1;    // for better accuracy, split interval into two
        result *= (sqrt(0.125)/M_PI/M_PI);
        double logphi_in5= scaledE(E);
        double logphi_in = spl_logpotential->y[0];
        double deriv_in = result>0 && !gsl_isinf(result) && !gsl_isnan(result) ? 
            (log(result)-logf.front())/(logphi_in5-logphi_in) : 0.0;  // derivative at E->Phi(0), or some default value
        E=pot(rout*5);   // extrapolation to E->0
        params.w=E;
        gsl_integration_qags (&F, E, 0, 0, 1e-7, 1000, w, &result, &error); 
        result *= (sqrt(0.125)/M_PI/M_PI);
        double logphi_out5= scaledE(E);
        double logphi_out = spl_logpotential->y[Npoints-1];
        double deriv_out = (log(result)-logf.back())/(logphi_out5-logphi_out);  // derivative at E->0
        logf.push_back(deriv_in);
        logf.push_back(deriv_out);
    }
    // init spline for f(E)
    spl_logdistrf = initSplineInScaledEnergy(phi, logf);
    gsl_integration_workspace_free (w);
}

bool CMassModel::initDistrFuncEnergy(const std::vector<double> &initE, const std::vector<double> &initF)
{
    if(initE.size()!=initF.size() && initE.size()+2!=initF.size()) {
#ifdef DEBUGPRINT
        my_message(FUNCNAME, "Input arrays not equal");
#endif
        return false;
    }
    std::vector<double> logF(initE.size());
    for(size_t i=0; i<initE.size(); i++) {
        if(initE[i]>=0 || (!totalmasscenter && initE[i]<=potentialcenter) || 
            (i>0 && initE[i]<=initE[i-1]) || initF[i]<=0 || !gsl_finite(initF[i]) || !gsl_finite(initE[i]) ) {
#ifdef DEBUGPRINT
                my_message(FUNCNAME, std::string("Distribution function is ")+
                    (initF[i]<=0 ? "negative" : "weird"));
#endif
                return false;   // error
        }
        logF[i]=log(initF[i]);
    }
    if(initE.size()+2==initF.size()) {  // endpoint derivatives were supplied
        logF.push_back(*(initF.end()-2));
        logF.push_back(*(initF.end()-1));
    }
    if(spl_logdistrf!=NULL) gsl_spline_free(spl_logdistrf);
    spl_logdistrf = initSplineInScaledEnergy(initE, logF);
    return true;
}

double int_mass(double r, void* params)
{
    double dens_interp=exp(gsl_spline_evalx(((CHelperParam*)params)->spl, log(r), NULL));
    return 4*M_PI*pow_2(r)*dens_interp;
}

bool CMassModel::computeNewPotential()
{
    if(spl_logdistrf==NULL) return false;
    if(spl_logphasevol==NULL) return false;
    std::vector<double> initE;
    // take the values of E from the distribution function spline knots, 
    // but in addition extrapolate in- and outward by a couple of decades (safety measure to increase accuracy)
    const double safetyFactor=5;   // exp(5) inward extrapolation
    double scaledEmin=spl_logdistrf->x[0]-safetyFactor;
    for(double sE=scaledEmin; sE<spl_logdistrf->x[0]-0.1; sE+=0.5)
        initE.push_back(unscaledE(sE));
    for(size_t i=0; i<spl_logdistrf->size; i++)
        initE.push_back(unscaledE(spl_logdistrf->x[i]));
    initE.push_back(unscaledE(spl_logdistrf->x[spl_logdistrf->size-1]+1.0));
    initE.push_back(unscaledE(spl_logdistrf->x[spl_logdistrf->size-1]+2.0));
    initE.push_back(unscaledE(spl_logdistrf->x[spl_logdistrf->size-1]+3.0));  // exp(3) outward extrapolation
    size_t npoints=initE.size();
    CHelperParam param;
    param.model=this;
    gsl_function Fnc;
    Fnc.params=&param;
    double result, error;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    std::vector<double> logr(npoints), logrho(npoints+2);
    gsl_spline *spl_dens=NULL;
    // compute new density from the given distribution function
    Fnc.function=&int_generic;
    param.memberFunction=&CMassModel::distrf;
    param.m=1;
    param.n=0.5;
    for(size_t i=0; i<npoints; i++) {
        param.w=initE[i];
        gsl_integration_qag(&Fnc, param.w, 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21,  ws, &result, &error);
        logr[i]=log(rmax(param.w));
        logrho[i]=log(4*M_PI*M_SQRT2*result);
    }
    // compute derivatives at endpoints
    param.n=-0.5;
    param.w=initE.front();
    gsl_integration_qags(&Fnc, param.w, 0, 0, 1e-7, 1000, ws, &result, &error);
    double rad=exp(logr[0]);
    double gam_in=-4*M_PI/M_SQRT2*rad/exp(logrho[0])*result*mass(rad)/pow_2(rad);
    param.w=initE.back();
    gsl_integration_qags(&Fnc, param.w, 0, 0, 1e-7, 1000, ws, &result, &error);
    rad=exp(logr[npoints-1]);
    double gam_out=-4*M_PI/M_SQRT2*rad/exp(logrho[npoints-1])*result*mass(rad)/pow_2(rad);

    if(gam_in<=-1.9) {
#ifdef DEBUGPRINT
        my_message(FUNCNAME,
            "Error in potential recomputation: gamma_in="+convertToString(-gam_in)+">2");
#endif
        gam_in=-1.9;
        //return false;  // non-integrable at origin
    }
    if(gam_out>=-3.1) {
#ifdef DEBUGPRINT
        my_message(FUNCNAME,
            "Error in potential recomputation: gamma_out="+convertToString(-gam_out)+"<3");
#endif
        gam_out=-3.1;
        //return false;  // mass is infinite
    }
    logrho[npoints]=gam_in;
    logrho[npoints+1]=gam_out;
    spl_dens=gsl_spline_alloc (gsl_interp_ccspline, npoints);
    gsl_spline_init(spl_dens, &(logr.front()), &(logrho.front()), npoints);
    // then compute new potential by integrating new density
    param.spl=spl_dens;
    Fnc.function=&int_mass;
    std::vector<double> newr(npoints+1,0), newm(npoints+1);
    newm[0]=masscenter;
    for(size_t i=0; i<npoints; i++) {
        newr[i+1]=exp(logr[i]);
        gsl_integration_qag(&Fnc, newr[i], newr[i+1], 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
        newm[i+1]=newm[i]+result;
    }
    initSplineMass(newr, newm);
    initSplinePhi(NULL);
    gsl_spline_free(spl_dens);
    gsl_integration_workspace_free(ws);
    return true;
}

double CMassModel::ekin() const
{
    gsl_function Fnc;
    CHelperParam param;
    Fnc.params=&param;
    Fnc.function=&int_generic_f;
    param.m=1;
    param.model=this;
    double result, error;
    size_t neval;
    param.memberFunction=&CMassModel::phasevol;
    double E=unscaledE(spl_logfhint->x[spl_logfhint->size-1]);
    gsl_integration_qng(&Fnc, E, 0, 0, 1e-7, &result, &error, &neval);
    return (result+exp(computeSplineInScaledEnergy(E, spl_logfhint)))*1.5;
}

//#if 0
double int_epot(double r, void* params)
{
    return 4*M_PI*r*r*((CHelperParam*)params)->model->pot(r)*((CHelperParam*)params)->model->dens(r);
}

double CMassModel::epot() const
{
    gsl_function Fnc;
    CHelperParam param;
    Fnc.params=&param;
    Fnc.function=&int_epot;
    param.model=this;
    double result, error;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    gsl_integration_qagiu(&Fnc, 0, 0, 1e-7, 1000, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return (result+masscenter*potentialcenter)/2;
}
//#else
double int_epot1(double r, void* params)
{
    double pot=((CHelperParam*)params)->model->pot(r);
    return -pot*r*r*sqrt(2*(((CHelperParam*)params)->v-pot));
}
double int_epot2(double E, void* params)
{
    
    return exp(((CHelperParam*)params)->model->computeSplineInScaledEnergy(E, ((CHelperParam*)params)->spl))
        * ((CHelperParam*)params)->model->distrf(E);
}

double CMassModel::epotf() const
{
    gsl_function Fnc;
    CHelperParam param;
    Fnc.params=&param;
    Fnc.function=&int_epot1;
    param.model=this;
    double result, error;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    std::vector<double> logE(spl_logpotential->size), logZ(spl_logpotential->size);
    for(size_t i=0; i<spl_logpotential->size; i++) {
        param.v=unscaledE(spl_logpotential->y[i]);
        gsl_integration_qag(&Fnc, 0, exp(spl_logpotential->x[i]), 0, 1e-7, 1000, GSL_INTEG_GAUSS21,  ws, &result, &error);
        logE[i]=spl_logpotential->y[i];
        logZ[i]=log(result);
    }
    gsl_spline* spl_Z=gsl_spline_alloc (gsl_interp_cspline, logE.size());
    gsl_spline_init(spl_Z, &(logE.front()), &(logZ.front()), logE.size());
    // then compute new potential by integrating new density
    Fnc.function=&int_epot2;
    param.spl=spl_Z;
    if(totalmasscenter) 
        gsl_integration_qagil(&Fnc, 0, 0, 1e-7, 1000, ws, &result, &error);
    else
        gsl_integration_qag(&Fnc, potentialcenter, 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
    result *= -16*M_PI*M_PI;
    gsl_spline_free(spl_Z);
    gsl_integration_workspace_free(ws);
    return (result+masscenter*potentialcenter)/2;
}
//#endif
double CMassModel::etot() const
{
    gsl_function Fnc;
    CHelperParam param;
    Fnc.params=&param;
    Fnc.function=&int_generic_f;
    param.model=this;
    param.memberFunction=&CMassModel::ded;
    param.m=1;
    param.n=1;
    double result, error;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    if(totalmasscenter) 
        gsl_integration_qagil(&Fnc, 0, 0, 1e-7, 1000, ws, &result, &error);
    else
        gsl_integration_qag(&Fnc, potentialcenter, 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
    gsl_integration_workspace_free(ws);
    return -result+masscenter*potentialcenter;
}

/* --- Calculation of diffusion coefficients --- */

// compute local velocity diffusion coefficients (using approximation from interpolation tables)
void CMassModel::DifCoefVel(double r, double v, double *dvpar, double *dv2par, double *dv2per, bool exact) const
{
    double Phi=pot(r), E=Phi+v*v/2;
    if(E>=0 || (!exact && splineJ1.size()==0)) {   // proceed only if dif.coef. interpolation table was initialized
        *dvpar=0; *dv2par=0; *dv2per=0; 
        return; 
    }   // currently no way of computing coefs for positive energy
    double I0, I12, I32;
    if(exact)
        DifVexact(E, Phi, &I0, &I12, &I32);
    else
        DifVapprox(E, Phi, &I0, &I12, &I32);
    *dvpar =-2.0*I12;
    *dv2par= 2.0/3*(I0+I32);
    *dv2per= 2.0/3*(3*I12-I32+2*I0);
}

// generic integrand for r^n vel^m (r1*I0+r2*I12+r3*I32)
// NB: params->v is energy, so velocity is sqrt(v-Phi(r))
double int_generic_dif(double r, void* params)
{
    double E=((CHelperParam*)params)->v;
    double Phi=((CHelperParam*)params)->model->pot(r);
    if(E<=Phi) return 0;
    double v2=2*(E-Phi);
    double I0, I12, I32;
    if(((CHelperParam*)params)->exact)
        ((CHelperParam*)params)->model->DifVexact(E, Phi, &I0, &I12, &I32);
    else
        ((CHelperParam*)params)->model->DifVapprox(E, Phi, &I0, &I12, &I32);
    return pow(r, ((CHelperParam*)params)->n)*pow(v2, ((CHelperParam*)params)->m/2) *
        (((CHelperParam*)params)->r1*I0 + ((CHelperParam*)params)->r2*I12 + ((CHelperParam*)params)->r3*I32);
}

// generic integrand for (q2+q3*v^-2)*v*<delta v_par> + (r1*v^2+r2+r3*v^-2)*<delta v^2_par> + (p2+p3*v^-2+w*r^2)*<delta v^2_per>
// params->v is energy, so velocity is sqrt(v-Phi(r))
// s is a scaled variable of Henon(1971): r = (r_peri+r_apo)/2 + (r_apo-r_peri)/4*(3*s-s^3), -1<=s<=1
// params->n=(r_peri+r_apo)/2;  params->a=(r_apo-r_peri)/4
double int_generic_dif1(double s, void* params)
{
    double r=((CHelperParam*)params)->n + ((CHelperParam*)params)->a*s*(3-s*s);
    double E=((CHelperParam*)params)->v;
    double Phi=((CHelperParam*)params)->model->pot(r);
    if(E<=Phi) return 0;
    double v2=2*(E-Phi);
    double I0, I12, I32;
    if(((CHelperParam*)params)->exact)
        ((CHelperParam*)params)->model->DifVexact(E, Phi, &I0, &I12, &I32);
    else
        ((CHelperParam*)params)->model->DifVapprox(E, Phi, &I0, &I12, &I32);
    double vdvpar=-(1+((CHelperParam*)params)->m)*I12;
    double dv2par=(I0+I32)*2/3;
    double dv2per=(2*I0+3*I12-I32)*2/3;
    double vr2=v2-((CHelperParam*)params)->L2/(r*r);
    return vr2<=0 ? 0 : 1/sqrt(vr2) * 3*((CHelperParam*)params)->a*(1-s*s) * (
        (((CHelperParam*)params)->q2 + ((CHelperParam*)params)->q3/v2)*vdvpar +
        (((CHelperParam*)params)->r1*v2 + ((CHelperParam*)params)->r2 + ((CHelperParam*)params)->r3/v2)*dv2par +
        (((CHelperParam*)params)->p2 + ((CHelperParam*)params)->p3/v2 + ((CHelperParam*)params)->w*r*r)*dv2per );
}

// fully-averaged (ergodic) dif.coefs (depend on E only)
void CMassModel::DifCoefEL(double E, double *DEE, double *DE, double *DLL, double *DRRoverR, bool exact) const
{
    if(E>=0) {   // currently no provision for computing coefs for positive energy
        if(DEE) *DEE=0;
        if(DE ) *DE =0;
        if(DLL) *DLL=0;
        if(DRRoverR) *DRRoverR=0;
        return;
    }
    gsl_integration_workspace* ws_difcoefs = NULL;
    if(exact || DLL!=NULL || DRRoverR!=NULL)
        ws_difcoefs = gsl_integration_workspace_alloc(1000);
    gsl_function F;
    CHelperParam params;
    F.params=&params;
    params.model=this;
    double norm=ded(E);
    double res0, result, error;
    bool failed=false;
    if(DEE || DE) {
        if(exact) {
            F.function=&int_generic_f;
            params.memberFunction=NULL;
            gsl_integration_qag(&F, E, 0, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, &res0, &error);
        } else 
            res0=exp(computeSplineInScaledEnergy(E, spl_logfint));
    }
    if(DEE){
        if(exact) {
            F.function=&int_generic_f;
            params.memberFunction=&CMassModel::phasevol;
            params.m=1;
            params.n=0;
            params.v=0;
            if(!totalmasscenter)
                gsl_integration_qag(&F, potentialcenter, E, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, &result, &error);
            else
                gsl_integration_qagil(&F, E, 0, 1e-7, 1000, ws_difcoefs, &result, &error);
        } else
            result=exp(computeSplineInScaledEnergy(E, spl_logfhint));
        result += res0*phasevol(E);
        *DEE = 2*16*M_PI*M_PI*result*(/*totalmass-masscenter*/1) / norm;
        if(!gsl_finite(*DEE)) failed=true;
    }
    if(DE){
        if(exact) {
            F.function=&int_generic_f;
            params.memberFunction=&CMassModel::ded;
            params.m=1;
            params.n=0;
            params.v=0;
            if(!totalmasscenter)
                gsl_integration_qag(&F, potentialcenter, E, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, &result, &error);
            else
                gsl_integration_qagil(&F, E, 0, 1e-7, 1000, ws_difcoefs, &result, &error);
        } else
            result=exp(computeSplineInScaledEnergy(E, spl_logfgint));
        result -= res0*norm;
        *DE = -result *16*M_PI*M_PI*(/*totalmass-masscenter*/1)/norm;
        if(!gsl_finite(*DE)) failed=true;
    }
    double rmx=(DLL || DRRoverR) ? rmax(E) : 0;
    if(DLL){
        F.function=&int_generic_dif;
        params.exact=exact;
        params.v=E;
        params.n=4;
        params.m=1;
        params.r1=2;
        params.r2=1;
        params.r3=1;
        gsl_integration_qags (&F, 0, rmx, 0, 1e-7, 1000, ws_difcoefs, &result, &error); 
        *DLL = result *16*M_PI*M_PI/norm/3;
        if(!gsl_finite(*DLL)) failed=true;
    }
    if(DRRoverR){
        F.function=&int_generic_dif;
        params.exact=exact;
        params.v=E;
        params.n=2;
        params.m=-1;
        params.r1=2;
        params.r2=3;
        params.r3=-1;
        gsl_integration_qags (&F, 0, rmx, 0, 1e-4, 1000, ws_difcoefs, &result, &error); 
        *DRRoverR = result *16*M_PI*M_PI/norm/3;
        if(!gsl_finite(*DRRoverR)) failed=true;
    }
    if(ws_difcoefs) 
        gsl_integration_workspace_free(ws_difcoefs);
#ifdef DEBUGPRINT
    if(failed)
        my_message(FUNCNAME, "Error in computing dif.coefs at E="+convertToString(E));
#endif
}

void CMassModel::DifVexact(double E, double Phi, double *I0, double *I12, double *I32) const
{
    //gsl_integration_workspace* ws_difcoefs = gsl_integration_workspace_alloc(1000);
    gsl_integration_cquad_workspace *ws_difcoefs = gsl_integration_cquad_workspace_alloc(1000);
    double error;
    size_t neval;
    gsl_function F;
    CHelperParam params;
    F.params=&params;
    params.model=this;
    F.function=&int_generic;
    params.memberFunction=&CMassModel::distrf;
    params.m=1;
    params.n=0;
    params.w=Phi;
    //gsl_integration_qag(&F, E, 0, 0, 1e-4, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, I0, &error);
    gsl_integration_cquad(&F, E, 0, 0, 1e-4, ws_difcoefs, I0, &error, &neval);
    *I0 *= 16*M_PI*M_PI*(totalmass-masscenter);
    params.n=0.5;
    //gsl_integration_qag(&F, Phi, E, 0, 1e-4, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, I12, &error);
    gsl_integration_cquad(&F, Phi, E, 0, 1e-4, ws_difcoefs, I12, &error, &neval);
    *I12 *= 16*M_PI*M_PI*(totalmass-masscenter)/sqrt(E-Phi);
    params.n=1.5;
    //gsl_integration_qag(&F, Phi, E, 0, 1e-4, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, I32, &error);
    gsl_integration_cquad(&F, Phi, E, 0, 1e-4, ws_difcoefs, I32, &error, &neval);
    *I32 *= 16*M_PI*M_PI*(totalmass-masscenter)/pow(E-Phi,1.5);
    //gsl_integration_workspace_free(ws_difcoefs);
    gsl_integration_cquad_workspace_free(ws_difcoefs);
}

const size_t splJ_sizex=31;
const size_t splJ_sizey=100;
const double splJ_facty=5.0;

void CMassModel::initSplineDifCoefV()
{
    gsl_integration_workspace* ws_difcoefs = gsl_integration_workspace_alloc(1000);
    double J0, J1, J3, error;
    gsl_function F;
    CHelperParam params;
    F.params=&params;
    params.model=this;
    F.function=&int_generic;
    params.memberFunction=&CMassModel::distrf;
    params.m=1;
    splineJ1.resize(splJ_sizex+1);
    splineJ3.resize(splJ_sizex+1);
    size_t sizey = totalmasscenter==0 ? splJ_sizey : spl_logdistrf->size;
    for(size_t ix=0; ix<=splJ_sizex; ix++) {
        double ox=(ix+0.0)/splJ_sizex;
        if(ox<0.1/splJ_sizex) ox=0.1/splJ_sizex;
        if(ox>1-0.1/splJ_sizex) ox=1-0.1/splJ_sizex;
        // first variable: E/Phi, scaled in such a way as to increase resolution at the ends of interval [0..1]
        double x=(1-cos(M_PI*ox))/2;
        if(x>=0.9999) x=0.9999;
        if(x<0.0001) x=0.0001;
        splineJ1[ix].resize(sizey);
        splineJ3[ix].resize(sizey);
        for(size_t iy=0; iy<sizey; iy++) {
            double E, Phi;
            if(!totalmasscenter) {  // second variable:  (E-Phi(0))/(Phi-Phi(0)), scaled in a logarithmic way
                double lny=-(iy+0.0)/sizey*splJ_facty;
                if(lny==0.0) lny=-0.001;
                double y=exp(lny);
                Phi=potentialcenter*(1-y)/(1-x*y);
                E=x*Phi;
            } else {  // second variable:  E, scaled in a logarithmic way (same as distribution function)
                E=unscaledE(spl_logdistrf->x[iy]);
                Phi=E/x;
            }
            params.w=Phi;
            params.n=0;
            gsl_integration_qag(&F, Phi, E, 0, 1e-4, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, &J0, &error);
            params.n=0.5;
            gsl_integration_qag(&F, Phi, E, 0, 1e-4, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, &J1, &error);
            splineJ1[ix][iy] = log(J1/sqrt(E-Phi)/J0);
            params.n=1.5;
            gsl_integration_qag(&F, Phi, E, 0, 1e-4, 1000, GSL_INTEG_GAUSS21, ws_difcoefs, &J3, &error);
            splineJ3[ix][iy] = log(J3/pow(E-Phi,1.5)/J0);
        }
    }
    gsl_integration_workspace_free(ws_difcoefs);
}

void CMassModel::DifVapprox(double E, double Phi, double *I0, double *I12, double *I32) const
{
    *I0 = 16*M_PI*M_PI*(totalmass-masscenter)* exp(computeSplineInScaledEnergy(E, spl_logfint));
    double J0 = 16*M_PI*M_PI*(totalmass-masscenter)* exp(computeSplineInScaledEnergy(Phi, spl_logfint)) - *I0;
    // 2d interpolation on x,y grid
    size_t ix, iy;  // indices of grid node just below the interpolation point
    double dx, dy;  // fractional distance from the (ix,iy) grid node in units of cell size [0..1]
    double x=E/Phi;
    double ox=acos(1-2*x)/M_PI;
    ix=static_cast<size_t>(ox*splJ_sizex);
    if(ix>=splJ_sizex) ix=splJ_sizex-1;
    dx=(ox*splJ_sizex-ix);
    if(!totalmasscenter) {
        double lny=-log( (Phi-potentialcenter)/(E-potentialcenter) );
        iy=static_cast<size_t>(lny*splJ_sizey/splJ_facty);
        if(iy>=splJ_sizey-1) iy=splJ_sizey-2;
        dy=(lny*splJ_sizey/splJ_facty-iy);
    } else {
        double y=scaledE(E);
        if(y<spl_logdistrf->x[0]) iy=0; 
        else if(y>=spl_logdistrf->x[spl_logdistrf->size-1]) iy=spl_logdistrf->size-2;
        else iy=gsl_interp_bsearch(spl_logdistrf->x, y, 0, spl_logdistrf->size-1);
        dy=(y-spl_logdistrf->x[iy])/(spl_logdistrf->x[iy+1]-spl_logdistrf->x[iy]);
    }
    double V00=splineJ1[ix][iy];
    double V01=splineJ1[ix][iy+1];
    double V10=splineJ1[ix+1][iy];
    double V11=splineJ1[ix+1][iy+1];
    double V=V00*(1-dx)*(1-dy)+V01*(1-dx)*dy+V10*dx*(1-dy)+V11*dx*dy;
    *I12 = J0 * exp(V);
    V00=splineJ3[ix][iy];
    V01=splineJ3[ix][iy+1];
    V10=splineJ3[ix+1][iy];
    V11=splineJ3[ix+1][iy+1];
    V=V00*(1-dx)*(1-dy)+V01*(1-dx)*dy+V10*dx*(1-dy)+V11*dx*dy;
    *I32 = J0 * exp(V);
}

void CMassModel::initSplineDifCoefE()
{
    if(spl_logdistrf==NULL) return;
    if(spl_logphasevol==NULL) return;
    std::vector<double> initE;
    // take the values of E from the distribution function spline knots, 
    // but in addition extrapolate in- and outward by a couple of decades (safety measure to increase accuracy)
    const double safetyFactor=5;   // exp(5) inward extrapolation
    double scaledEmin=std::min<double>(spl_logdistrf->x[0], spl_logphasevol->x[0])-safetyFactor;
    for(double sE=scaledEmin; sE<spl_logdistrf->x[0]-0.1; sE+=0.5)
        initE.push_back(unscaledE(sE));
    for(size_t i=0; i<spl_logdistrf->size; i++)
        initE.push_back(unscaledE(spl_logdistrf->x[i]));
    initE.push_back(unscaledE(spl_logdistrf->x[spl_logdistrf->size-1]+1.0));
    initE.push_back(unscaledE(spl_logdistrf->x[spl_logdistrf->size-1]+2.0));
    initE.push_back(unscaledE(spl_logdistrf->x[spl_logdistrf->size-1]+3.0));  // exp(3) outward extrapolation
    size_t Npoints=initE.size();
    std::vector<double> logf(Npoints), logfg(Npoints), logfh(Npoints);
    gsl_function Fnc;
    CHelperParam param;
    Fnc.params=&param;
    Fnc.function=&int_generic_f;
    param.m=1;
    param.model=this;
    double result, error;
    gsl_integration_workspace * ws = gsl_integration_workspace_alloc(1000);
    for(size_t i=0; i<Npoints; i++)
    {
        double f, b;   // f(E) = A*(-E)^b for E-> -infinity
        double E=initE[i];

        param.memberFunction=&CMassModel::ded;
        if(totalmasscenter && i==0) {  // integrate exactly
            computeSplineDerivInScaledEnergy(E, spl_logdistrf, &f, &b, NULL);
            if(b>=0.45) {
#ifdef DEBUGPRINT
                my_message(FUNCNAME, "Distribution function is too steep as E->-infinity");
#endif
                b=0.45;
            }
            b*= E;
            f = exp(f);
            // g(E) ~ E^-5/2
            result=-E/(1.5-b)*f*ded(E);
        }
        else
            gsl_integration_qag(&Fnc, i==0?potentialcenter:initE[i-1], E, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
        if(i>0) result+=logfg[i-1];
        logfg[i]=result;

        param.memberFunction=&CMassModel::phasevol;
        if(totalmasscenter && i==0) {
            result=-E/(0.5-b)*f*phasevol(E);
        }
        else
            gsl_integration_qag(&Fnc, i==0?potentialcenter:initE[i-1], E, 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
        if(i>0) result+=logfh[i-1];
        logfh[i]=result;

        param.memberFunction=NULL;
        E=initE[Npoints-1-i];
        gsl_integration_qag(&Fnc, E, i==0?0:initE[Npoints-i], 0, 1e-7, 1000, GSL_INTEG_GAUSS21, ws, &result, &error);
        if(i>0) result+=logf[Npoints-i];
        logf[Npoints-1-i]=result;
    }
    for(size_t i=0; i<Npoints; i++)
    {
        logf[i]=log(logf[i]);
        logfg[i]=log(logfg[i]);
        logfh[i]=log(logfh[i]);
    }
    gsl_integration_workspace_free(ws);
    if(spl_logfint)  gsl_spline_free(spl_logfint);
    if(spl_logfgint) gsl_spline_free(spl_logfgint);
    if(spl_logfhint) gsl_spline_free(spl_logfhint);
    spl_logfint = initSplineInScaledEnergy(initE, logf);
    spl_logfgint= initSplineInScaledEnergy(initE, logfg);
    spl_logfhint= initSplineInScaledEnergy(initE, logfh);
}

double find_diffusion_limit(double scaledE, void* param)
{
    const CMassModel* model=((CHelperParam*)param)->model;
    double E=model->unscaledE(scaledE);
    double Torb=((CHelperParam*)param)->model->trad(E);
    double dc, norm;
    if( ((CHelperParam*)param)->v==0) {  // compare energy dif.coef.
        model->DifCoefEL(E, &dc);
        norm = model->mass(0) ? -E : std::min<double>(-E, E-model->getPotentialAtCenter());
    } else {  // compare angular momentum dif.coef.
        model->DifCoefEL(E, NULL, NULL, &dc);
        norm = model->lcirc(E);
    }
    return sqrt(dc * ((CHelperParam*)param)->a * Torb) / norm - 1;
}

double CMassModel::getMinEnergyOrbitAvg(double dcmult) const
{
    if(totalmasscenter || dcmult==0) return gsl_neginf();
    // estimate the energy below which  sqrt( <DeltaE^2>*P ) >~ (E-potentialcenter)
    gsl_function F;
    CHelperParam param;
    F.params=&param;
    F.function=&find_diffusion_limit;
    param.model=this;
    param.a=dcmult;
    param.v=0;  // compare energy diff.coef.
    double Eupp=scaledE(pot(rad(totalmass/2)));
    double Elow=spl_logfint->x[0];
    double E=Eupp, val=gsl_posinf(), valprev;
    do {
        valprev=val;
        val=F.function(E, F.params);
        if(val<0) { Eupp=E; E-=0.5; }
    } while(val<0 && (E>Elow || val>valprev) && E>Elow-10.0);
    if(val<0) return potentialcenter;  // no such value exists
    E=findRoot(&F, E, Eupp, E, 0.01);
    return unscaledE(E);
}

// copy constructor taking into account the need to make deep copy of splines data
CMassModel::CMassModel(const smile::CMassModel &src) :
  // copy all trivial members by value
  errorcode(src.errorcode),
  potentialcenter(src.potentialcenter),
  totalmass(src.totalmass),
  masscenter(src.masscenter),
  totalmasscenter(src.totalmasscenter),
  rin(src.rin), rout(src.rout),
  gammain(src.gammain), gammaout(src.gammaout),
  coefmrout(src.coefmrout),
  potentialin(src.potentialin),
  densconst(src.densconst),
  alpha(src.alpha), rho0(src.rho0), coefdrhodr(src.coefdrhodr), coefdrhodphi(src.coefdrhodphi),
  // use standard copy constructor for vectors
  splineJ1(src.splineJ1), splineJ3(src.splineJ3),
  // manually clone GSL splines
  spl_loginmass   (copyspline(src.spl_loginmass)), 
  spl_logrmass    (copyspline(src.spl_logrmass)), 
  spl_logpotential(copyspline(src.spl_logpotential)), 
  spl_logdenspot  (copyspline(src.spl_logdenspot)), 
  spl_logdistrf   (copyspline(src.spl_logdistrf)), 
  spl_logphasevol (copyspline(src.spl_logphasevol)), 
  spl_logtrad     (copyspline(src.spl_logtrad)), 
  spl_logrcirc    (copyspline(src.spl_logrcirc)), 
  spl_logfint     (copyspline(src.spl_logfint)), 
  spl_logfgint    (copyspline(src.spl_logfgint)),
  spl_logfhint    (copyspline(src.spl_logfhint))
{}

// constructor:
// create model from array of { r, m(r) } pairs, initialize all relevant splines
CMassModel::CMassModel(const std::vector<double> &initRadii, 
                       const std::vector<double> &initMass, 
                       const CMassModel* model_pot)
{
    spl_loginmass=NULL;
    spl_logrmass=NULL;
    spl_logpotential=NULL;
    spl_logdenspot=NULL;
    spl_logdistrf=NULL;
    spl_logphasevol=NULL;
    spl_logtrad=NULL;
    spl_logrcirc=NULL;
    spl_logfint=NULL;
    spl_logfgint=NULL;
    spl_logfhint=NULL;
    errorcode=MS_OK;
    if(initRadii.size()!=initMass.size())
    {
        errorcode=MS_ARRAY_NOT_EQ;  // input arrays not equal
        return;
    }
    for(size_t i=1; i<initRadii.size(); i++)
    {
        if(initRadii[i]<=initRadii[i-1]){
            errorcode=MS_RADII_NOT_ASC; // "Radii not in ascending order!\n";
            return;
        }
        if(initMass[i]<=initMass[i-1]){
            errorcode=MS_MASS_NOT_ASC; // "Mass not increasing with radius!\n";
            return;
        }
    }
    if(initRadii.size()<5) {
        errorcode=MS_ARRAY_SHORT;  // Too few points
        return;
    }
    initSplineMass(initRadii, initMass);  // initialize mass profile
    initSplinePhi(model_pot);  // compute self-consistent potential or use external one
    initEddingtonDF();    // compute distribution function
    initSplineDifCoefE(); // compute orbit-averaged diffusion coefficients
    initSplineDifCoefV(); // compute local velocity diffusion coefficients
}

CMassModel::~CMassModel()
{
    if(spl_loginmass!=NULL)   gsl_spline_free(spl_loginmass);
    if(spl_logrmass!=NULL)    gsl_spline_free(spl_logrmass);
    if(spl_logpotential!=NULL)gsl_spline_free(spl_logpotential);
    if(spl_logdenspot!=NULL)  gsl_spline_free(spl_logdenspot);
    if(spl_logdistrf!=NULL)   gsl_spline_free(spl_logdistrf);
    if(spl_logphasevol!=NULL) gsl_spline_free(spl_logphasevol);
    if(spl_logtrad!=NULL)     gsl_spline_free(spl_logtrad);
    if(spl_logrcirc!=NULL)    gsl_spline_free(spl_logrcirc);
    if(spl_logfint!=NULL)     gsl_spline_free(spl_logfint);
    if(spl_logfgint!=NULL)    gsl_spline_free(spl_logfgint);
    if(spl_logfhint!=NULL)    gsl_spline_free(spl_logfhint);
}

//-------------- penalized spline fitting for creating mass models from N-body snapshots --------------//

// sort pairs of floating-point values by the first element
template<typename NumT>
bool sortfnc(const std::pair<NumT, NumT> &elem1, const std::pair<NumT, NumT> &elem2)
{   return elem1.first<elem2.first; }

/// \cond INTERNAL_DOCS
typedef struct{
    double* xdata;
    double* ydata;
    size_t  npoints;
} fitdata;
/// \endcond

// helper routine to find extrapolated mass of an Nbody snapshot at infinity Mtot 
// by requiring that log(M(r)/(Mtot-M(r))) vs log(r) is best described by a linear relation close to outer radius
double getfitquality(double totalmass, void* params)
{
    size_t npoints=static_cast<fitdata*>(params)->npoints;
    double* xdata=static_cast<fitdata*>(params)->xdata;
    double* ydata=static_cast<fitdata*>(params)->ydata;
    std::vector<double> y1data(npoints);
    for(size_t i=0; i<npoints; i++)
        y1data[i]=log(ydata[i]/(totalmass-ydata[i]));
    double cov00, cov01, cov11, sumsq, a, b;
    gsl_fit_linear(xdata, 1, &(y1data.front()), 1, npoints, &a, &b, &cov00, &cov01, &cov11, &sumsq);
    double norm=b*(xdata[npoints-1]-xdata[0]);  // normalization factor: difference between first and last point according to linear fit
    return sumsq/(norm*norm);
}

template<typename NumT>
bool fitMassModel(const std::vector<NumT> &particlecoord, const std::vector<NumT> &particlemass,
                  std::vector<double> *gridcoord, std::vector<double> *gridmass, double smoothing, 
                  size_t gridsize, size_t nskip, CMassModel::FITTYPE ftype, const CMassModel* model)
{
    if(particlecoord.size()!=particlemass.size() || gridcoord==NULL || gridmass==NULL || 
        (ftype!=CMassModel::FT_MASS && model==NULL))
        return false;
    size_t nbody_in=particlecoord.size();
    std::vector< std::pair<NumT, NumT> > coordmass(nbody_in);
    for(size_t i=0; i<nbody_in; i++) {
        coordmass[i].first=particlecoord[i];
        coordmass[i].second=particlemass[i];
    }
    // sort particules in radius and initialize M(r) grid by fitting a smoothing spline to 
    // log(r), log[M(r)]
    std::sort(coordmass.begin(), coordmass.end(), sortfnc<NumT>);
    double totalmass=0, bhmass=0;
    std::vector<double> logcoord, logmassrel;
    logcoord.reserve(nbody_in);
    logmassrel.reserve(nbody_in);
    for(size_t i=0; i<nbody_in; i++)
    {
        double coord;
        switch(ftype) {   // perform suitable type of scaling for coordinate
            case CMassModel::FT_MASS:
                coord=log(coordmass[i].first);
                if(coordmass[i].first==0) 
                    bhmass+=coordmass[i].second;
                break;
            case CMassModel::FT_DF_ENERGY:
                coord=model->scaledE(coordmass[i].first);
                break;
            case CMassModel::FT_DF_PHASEVOL:
                coord=log(model->phasevol(coordmass[i].first));
                break;
            default:
                continue;
        }
        if(!gsl_isnan(coord) && !gsl_isinf(coord) && coordmass[i].second>0)
        {
            totalmass+=coordmass[i].second;
            logcoord.push_back(coord);
            logmassrel.push_back(totalmass);  // not log-mass yet; will require one more step to find extrapolated Mtot 
        }
    }
    size_t nbody1=logcoord.size();
    std::vector<double> knots;           // (almost equidistant) grid in log radius
    std::vector<double> gridlogmassrel;  // values of log-scaled mass returned by the fitting routine
    // create equidistant grid in log r
    size_t minbin=static_cast<size_t>(10 + 5*log10(nbody1/1000.));
    if(minbin<5) minbin=5;
    if(gridsize==0)
        gridsize=static_cast<size_t>(10*pow(nbody1/1000., 1./3));   // some heuristical recipe
    if(gridsize*minbin>nbody1) gridsize=nbody1/minbin;
    // assign knots in radial grid trying to achieve equal log-r intervals 
    // but requiring that each interval holds at least minbin points
    size_t nskipfront=nskip, nskipback=nskip;
    do{
#ifdef DEBUGPRINT
        if(nskipfront+nskipback>0)
            my_message(FUNCNAME,
            "Skipping "+convertToString(nskipfront)+" points from front and "+convertToString(nskipback)+" from back");
#endif
        if(nskipback>0) {
            logcoord.resize(logcoord.size()-nskipback);
            logmassrel.resize(logmassrel.size()-nskipback);
        }
        if(nskipfront>0) {
            logcoord.erase(logcoord.begin(), logcoord.begin()+nskipfront);
            logmassrel.erase(logmassrel.begin(), logmassrel.begin()+nskipfront);
        }
        nbody1-=(nskipfront+nskipback);
        smile::createAlmostUniformGrid(logcoord, minbin, &gridsize, &knots);
        // check that 1st grid cell is not too 'sparse' (leftmost point(s) are not too widely spaced), otherwise ignore first points
        nskipfront=0;
        double tolerable_spacing = (knots[1]-knots[0])/minbin * 3.0;
        for(size_t i=1; i<minbin; i++)
            if(logcoord[i]-logcoord[0] > i*tolerable_spacing) 
                nskipfront=i;
        // same for outermost points
        nskipback=0;
        tolerable_spacing = (knots[gridsize-1]-knots[gridsize-2])/minbin * 3.0;
        for(size_t i=1; i<minbin; i++)
            if(logcoord[nbody1-1]-logcoord[nbody1-1-i] > i*tolerable_spacing) 
                nskipback=i;
    } while(nskipfront+nskipback>0);
    double totalmassfit=totalmass;
    if(nskip==0) {
        // find "totalmassfit" such that log(m[i]/(totalmassfit-m[i])) is best described by a linear function in log(r)
        gsl_function Fmin;
        fitdata Fdata;
        size_t nfit=1; 
        while(nfit<nbody1 && logcoord[nbody1-nfit]>knots[knots.size()-3]) nfit++;
        Fdata.xdata=&(logcoord[nbody1-nfit]);
        Fdata.ydata=&(logmassrel[nbody1-nfit]);
        Fdata.npoints=nfit;
        Fmin.function=&getfitquality;
        Fmin.params=&Fdata;
        totalmassfit=smile::findMinGuess(&Fmin, totalmass*1.0000001, totalmass*1.1);
    }   // otherwise simply use total mass of all input particles (including those that were skipped)
    for(size_t i=0; i<nbody1; i++)
        logmassrel[i]=log( logmassrel[i]/(totalmassfit-logmassrel[i]));  // scale to M(r)/(Mtot-M(r))

    // init penalized spline fitting routine
    smile::CSplineApprox fit(logcoord, knots);
    if(fit.status()!=smile::CSplineApprox::AS_OK && fit.status()!=smile::CSplineApprox::AS_SINGULAR) {
#ifdef DEBUGPRINT
        my_message(FUNCNAME, "Error in creating smoothing spline for M(r)");
#endif
        return false;
    }
    fit.loadyvalues(logmassrel);
    double rmserr, edf, lambda;
    smile::CSplineApprox::APPROXSTATUS status = fit.fitDataOversmooth(smoothing, &gridlogmassrel, &rmserr, &edf, &lambda);
    if(status != smile::CSplineApprox::AS_OK && status != smile::CSplineApprox::AS_SINGULAR) {
#ifdef DEBUGPRINT
        my_message(FUNCNAME, "Error in fitting smoothing spline for M(r)");
#endif
        return false;
    }
#ifdef DEBUGPRINT
    else {
        my_message(FUNCNAME, "nbody mass="+convertToString(totalmass)+
        ", fit mass="+convertToString((totalmassfit/(1+exp(-gridlogmassrel[knots.size()-1]))))+
        ", extrapolated mass="+convertToString(totalmassfit)+
        ", rmserr="+convertToString(rmserr)+
        ", edf(# independent radial pts)="+convertToString(edf)+
        ", lambda="+convertToString(lambda));
        if(ftype!=CMassModel::FT_DF_ENERGY) 
            my_message(FUNCNAME, "grid in coord from "+convertToString(exp(knots.front()))+
            " to "+convertToString(exp(knots.back())));
        else
            my_message(FUNCNAME, "grid in coord from "+convertToString(model->unscaledE(knots.front()))+
            " to "+convertToString(model->unscaledE(knots.back())));
    }
#endif
    // scale back from log-scaling
    if(gridcoord->size()<5)  {  // no user-defined grid in radius was provided, use the internally created grid
        gridcoord->assign(knots.size(),0);
        gridmass->assign(knots.size(),0);
        for(size_t k=0; k<knots.size(); k++) 
            switch(ftype) {   // perform suitable type of scaling for coordinate
            case CMassModel::FT_MASS:
                gridcoord->at(k)=exp(knots[k]);
                break;
            case CMassModel::FT_DF_ENERGY:
                gridcoord->at(k)=model->unscaledE(knots[k]);
                break;
            case CMassModel::FT_DF_PHASEVOL:
                gridcoord->at(k)=exp(knots[k]);
                break;
            default:
                continue;
        }
        if(ftype==CMassModel::FT_MASS && bhmass>0)
            gridcoord->insert(gridcoord->begin(), 0);
    }
    gridmass->assign(gridcoord->size(), 0);
    // create auxiliary clamped cubic spline to provide derivatives and extrapolation outside knot definition region if necessary
    gsl_spline* spl=gsl_spline_alloc(gsl_interp_ccspline, knots.size());
    gsl_spline_init(spl, &(knots.front()), &(gridlogmassrel.front()), knots.size());
    switch(ftype) {
    // Warning, no check is performed if the mass is increasing with radius, or if the distribution function is non-negative
    case CMassModel::FT_MASS:
        // check for inner density profile slope to be reasonable
        if(gridlogmassrel[knots.size()]>=3)  {
            my_message(FUNCNAME, 
                "Warning, inner density was dropping towards the center, a constant-density core is assumed");
            gridlogmassrel[knots.size()]=3;
        }
        if(gridlogmassrel[knots.size()]<=1.05)  {
            my_message(FUNCNAME,
                "Warning, inner density was rising steeper than r^-2, slope is corrected");
            gridlogmassrel[knots.size()]=1.05;
        }
        // compute extrapolated mass at provided radial points
        for(size_t k=0; k<gridcoord->size(); k++) {
            gridmass->at(k) = bhmass + (gridcoord->at(k)==0 ? 0 :
                totalmassfit/(1+exp(-gsl_spline_evalx(spl, log(gridcoord->at(k)), NULL))) );
        }
        break;
    case CMassModel::FT_DF_ENERGY:
        // f(E) = 1/g(E) * d M(E) / d E = 1/g(E) * dMscaled/dEscaled * dEscaled/dE / (dMscaled/dM)
        for(size_t k=0; k<gridcoord->size(); k++) {
            double Escaled=model->scaledE(gridcoord->at(k));
            double Mscaled = gsl_spline_evalx(spl, Escaled, NULL);
            double dE_over_dEscaled = 1/model->scaledEder(gridcoord->at(k));
            double dM_over_dMscaled = totalmassfit * exp(Mscaled)/pow_2(exp(Mscaled)+1);
            double dMscaled_over_dEscaled = gsl_spline_eval_derivx(spl, Escaled, NULL);
            gridmass->at(k) = dMscaled_over_dEscaled * dM_over_dMscaled / dE_over_dEscaled / model->ded(gridcoord->at(k));
        }
        break;
    case CMassModel::FT_DF_PHASEVOL:
        // f(h) = d M(h) / d h = dMscaled/dHscaled * dHscaled/dH / (dMscaled/dM)
        for(size_t k=0; k<gridcoord->size(); k++) {
            double Hscaled = log(gridcoord->at(k));
            double Mscaled = gsl_spline_evalx(spl, Hscaled, NULL);
            double dM_over_dMscaled = totalmassfit * exp(Mscaled)/pow_2(exp(Mscaled)+1);
            double dH_over_dHscaled = gridcoord->at(k);  // H
            double dMscaled_over_dHscaled = gsl_spline_eval_derivx(spl, Hscaled, NULL);
            gridmass->at(k) = dMscaled_over_dHscaled * dM_over_dMscaled / dH_over_dHscaled;
        }
        break;
    default:
        gsl_spline_free(spl);
        return false;
    }
    gsl_spline_free(spl);
    return true;
}

template
bool fitMassModel(const std::vector<double> &particlecoord, const std::vector<double> &particlemass,
                  std::vector<double> *gridcoord, std::vector<double> *gridmass, double smoothing, 
                  size_t gridsize, size_t nskip, CMassModel::FITTYPE ftype, const CMassModel* model);
template
bool fitMassModel(const std::vector<float> &particlecoord, const std::vector<float> &particlemass,
                  std::vector<double> *gridcoord, std::vector<double> *gridmass, double smoothing, 
                  size_t gridsize, size_t nskip, CMassModel::FITTYPE ftype, const CMassModel* model);

}