#include "odeint.h"
#include <cmath>
#include <gsl/gsl_math.h>

namespace smile{

/* ----------------- Orbit integrators ------------- */

/* --- leap-frog integrator for frozen-N treecode potential --- */
void COdeIntegratorLeapfrog::integrateToTime(double timeEnd)
{
    if(!odeSystem->isStdHamiltonian())
        return;
    OdeStateType derivPrev, derivCurr;
    odeSystem->initState(&stateCurr);
    statePrev.assign(stateCurr.size(), 0);
    derivCurr.assign(stateCurr.size()+1, 0);
    derivPrev.assign(stateCurr.size()+1, 0);
    timePrev=timeCurr;
    double hprev=0, h=hprev;
    size_t numVars2=stateCurr.size()/2;
    unsigned int nstep=0;
    if(numVars2!=N_DIM)
        return;  // only works in 3d case, for simplicity
    odeSystem->odeFnc(timeCurr, stateCurr, &derivCurr);
    bool needToTerminate=false;
    while(!needToTerminate)
    {
        for(size_t d=0; d<numVars2*2; d++) {
            statePrev[d] = stateCurr[d];
            derivPrev[d] = derivCurr[d];
        }
        timePrev=timeCurr;
        if(symmetrizeTimestep)
            h=accTreecode;   ///!!! provision for constant timestep
        else {
        // estimate timestep based on the distance to the nearest particle, which was computed during tree-walk
        double dist= derivCurr.back();
        double vel = sqrt(pow_2(stateCurr[numVars2+0])+pow_2(stateCurr[numVars2+1])+pow_2(stateCurr[numVars2+2]));
        double acc = sqrt(pow_2(derivCurr[numVars2+0])+pow_2(derivCurr[numVars2+1])+pow_2(derivCurr[numVars2+2]));
        double ts1 = vel>0 ? dist/vel : timeEnd;
        double ts2 = acc>0 ? sqrt(dist/acc) : timeEnd;
        //if(potential->Mbh>0) ts1=std::min<double>(ts1, sqrt(pow_2(y[0])+pow_2(y[1])+pow_2(y[2]))/vel*TREECODE_TIMESTEP_BH_FACTOR);
        h = std::min<double> (ts1, ts2) * accTreecode;
        }
        /*if(symmetrizeTimestep){
            // iteratively estimate timestep according to (Hut, Makino & McMillan 1995)
            for(unsigned int d=0; d<N_DIM; d++)  // first step for position
                y[d] = yprev[d] + yprev[d+N_DIM]*h + aprev[d]*h*h/2;
            potential->Force(y, timePrev+h, force, NULL);
            nfcn++;
            for(unsigned int d=0; d<N_DIM; d++)
                y[d+N_DIM] = yprev[d+N_DIM] + (aprev[d]+force[d])*h/2;
            vel = sqrt(pow_2(y[N_DIM])+pow_2(y[N_DIM+1])+pow_2(y[N_DIM+2]));
            acc = sqrt(pow_2(force[0])+pow_2(force[1])+pow_2(force[2]));
            ts1 = vel>0 ? force[N_DIM]/vel : timeEnd;
            ts2 = acc>0 ? sqrt(force[N_DIM]/acc) : timeEnd;
            if(potential->Mbh>0) ts1=std::min<double>(ts1, sqrt(pow_2(y[0])+pow_2(y[1])+pow_2(y[2]))/vel*TREECODE_TIMESTEP_BH_FACTOR);
            double h1 = std::min<double> (ts1, ts2) * accTreecode;
            h = (h+h1)/2;
        }*/
        if(timeCurr<timeEnd && timeCurr+h>timeEnd) { h=timeEnd-timeCurr; timeCurr=timeEnd; }   // last step exactly at x_end
        else timeCurr+=h;
        for(unsigned int d=0; d<numVars2; d++)  // first step for position
            stateCurr[d] = statePrev[d] + statePrev[d+numVars2]*h + derivPrev[d+numVars2]*h*h/2;
        odeSystem->odeFnc(timeCurr, stateCurr, &derivCurr);
        nstep++;
        for(unsigned int d=0; d<numVars2; d++)  // second step for velocity
            stateCurr[d+numVars2] = statePrev[d+numVars2] + (derivPrev[d+numVars2]+derivCurr[d+numVars2])/2*h;
        hprev=h;
        COdeSystem::STEPRESULT result = odeSystem->processStep(timePrev, timeCurr, &stateCurr);
        if(result == COdeSystem::SR_TERMINATE) 
            needToTerminate=true;
        if(!gsl_finite(h) || fabs(h) <= fabs(timeCurr) * GSL_DBL_EPSILON*10) needToTerminate=true;   // timestep too small
        if(nstep>NUMSTEP_MAX) needToTerminate=true;  // too many steps or stepsize adjustments
    }
}

double COdeIntegratorLeapfrog::getInterpolatedSolution(unsigned int d, double x) const
{
    if(x<timePrev || x>timeCurr) return gsl_nan();
    double dt=x-timePrev, timeStep=timeCurr-timePrev, dtt=dt/timeStep;
    unsigned int c=d%N_DIM;  // index of coordinate
    if(d<N_DIM)  // interpolate position
        return statePrev[c] + statePrev[c+N_DIM]*dt*(1-dtt) + (stateCurr[c]-statePrev[c])*dtt*dtt;
    else         // interpolate velocity
        return statePrev[c+N_DIM]*(1-dtt) + stateCurr[c+N_DIM]*dtt;
}

#ifdef HAVE_ODEINT
// Fancy C++ ODE integrator from boost

// some machinery to be able to use member functions in templated class
template< class Obj>
class ode_wrapper
{
    Obj* m_obj;
public:
    ode_wrapper( Obj* obj ) : m_obj( obj ) { }

    template< class State , class Deriv , class Time >
    void operator()( const State &x , Deriv &dxdt , Time t )
    {
        m_obj->odeFnc( t, x , &dxdt );
    }
};

template< class Obj >
ode_wrapper< Obj > make_ode_wrapper( Obj* obj )
{
    return ode_wrapper< Obj >( obj );
}

template< class Obj, class TempStorage >
class ode_wrapper_sympl
{
    Obj* obj;
    TempStorage* tmpx, *tmpdx;
public:
    ode_wrapper_sympl( Obj* _obj, TempStorage* _tmpx, TempStorage* _tmpdx) : 
      obj(_obj), tmpx(_tmpx), tmpdx(_tmpdx) { }

    template< class State , class Deriv >
    void operator()( const State &q , Deriv &dpdt )
    {
        size_t numVars2=tmpx->size()/2;
        for(size_t i=0; i<numVars2; i++)
            tmpx->at(i)=q[i];
        obj->odeFnc( 0, *tmpx , tmpdx );
        for(size_t i=0; i<numVars2; i++)
            dpdt[i]=tmpdx->at(i+numVars2);
    }
};

template< class Obj, class TempStorage >
ode_wrapper_sympl< Obj, TempStorage > make_ode_wrapper_sympl( Obj* obj, 
    TempStorage* tempStorageX, TempStorage* tempStorageDXDT)
{
    return ode_wrapper_sympl< Obj, TempStorage >(obj, tempStorageX, tempStorageDXDT);
}

// Adaptive, dense-output 5th order Runge-Kutta (Dormand-Prince)
template<>
COdeIntegratorOdeint<StepperDP5>::COdeIntegratorOdeint(
    COdeSystem* _odeSystem, double _accAbs, double _accRel) :
  CBasicOdeIntegrator(_odeSystem),
  stepper(boost::numeric::odeint::controlled_runge_kutta< 
          boost::numeric::odeint::runge_kutta_dopri5< OdeStateType >, my_error_checker >
          (my_error_checker(_odeSystem, _accAbs, _accRel),
          boost::numeric::odeint::runge_kutta_dopri5< OdeStateType >() ) ),
  isStdHamiltonian(odeSystem->isStdHamiltonian())
{}

template<> 
std::string COdeIntegratorOdeint<StepperDP5>::myName() { return "DormandPrince5"; };

// Adaptive 5th order Runge-Kutta (Cash-Karp)
template<>
COdeIntegratorOdeint<StepperCK5>::COdeIntegratorOdeint(
    COdeSystem* _odeSystem, double _accAbs, double _accRel) :
  CBasicOdeIntegrator(_odeSystem),
  stepper(my_error_checker(_odeSystem, _accAbs, _accRel),
          boost::numeric::odeint::runge_kutta_cash_karp54< OdeStateType >() ),
  isStdHamiltonian(odeSystem->isStdHamiltonian())
{}

template<> 
std::string COdeIntegratorOdeint<StepperCK5>::myName() { return "CashKarp5"; };

// Fixed-step 4th order Runge-Kutta
template<>
COdeIntegratorOdeint<StepperRK4>::COdeIntegratorOdeint(
    COdeSystem* _odeSystem, double _timeStep, double /*_accRel*/) :
  CBasicOdeIntegrator(_odeSystem), 
  stepper(), 
  timeStep(_timeStep),
  isStdHamiltonian(odeSystem->isStdHamiltonian())
{}

template<> 
std::string COdeIntegratorOdeint<StepperRK4>::myName() { return "RungeKutta4"; };

// Bulirsch-Stoer (adaptive with dense output)
template<>
COdeIntegratorOdeint<StepperBS>::COdeIntegratorOdeint(
    COdeSystem* _odeSystem, double _accAbs, double _accRel) :
  CBasicOdeIntegrator(_odeSystem),
  stepper(_accAbs, _accRel), 
  isStdHamiltonian(odeSystem->isStdHamiltonian())
{}

template<> 
std::string COdeIntegratorOdeint<StepperBS>::myName() { return "BulirschStoer"; };

// Symplectic 4-th order Runge-Kutta (fixed-step)
template<>
COdeIntegratorOdeint<StepperSympl4>::COdeIntegratorOdeint(
    COdeSystem* _odeSystem, double _accAbs, double /*_accRel*/) :
  CBasicOdeIntegrator(_odeSystem), 
  stepper(), 
  timeStep(_accAbs),
  isStdHamiltonian(odeSystem->isStdHamiltonian())
{}

template<> 
std::string COdeIntegratorOdeint<StepperSympl4>::myName() { return "SymplecticRK4"; };

// generic integration function
template< class Stepper >
void COdeIntegratorOdeint<Stepper>::integrateToTime(double timeEnd)
{
    odeSystem->initState(&stateCurr);
    timeIntermediate=timeCurr;
    stateIntermediate=stateCurr;
    // do actual work differently for dense- and non-dense-output steppers
    integrateToTimeImpl(timeEnd, stepper_category());
}

// integration for fixed-timestep symplectic steppers
template<>
void COdeIntegratorOdeint<StepperSympl4>::integrateToTime(double timeEnd)
{
    if(!odeSystem->isStdHamiltonian()) return;   // cannot deal with systems which are not separable Hamiltonian in standard form
    odeSystem->initState(&stateCurr);
    size_t numVars=stateCurr.size();
    OdeStateType state_q(numVars/2), state_p(numVars/22);
    OdeStateType tmp1(numVars), tmp2(numVars);  // reserve space for intermediate computations
    size_t nstep = 0;
    bool needToTerminate=false;
    double tstep=timeStep;
    COdeSystem::STEPRESULT result = COdeSystem::SR_REINIT;
    while(!needToTerminate)
    {
        timePrev=timeCurr;
        statePrev=stateCurr;
        if(timeCurr<timeEnd && timeCurr+tstep > timeEnd) 
        {   // ensure that last step ends exactly at t_end
            tstep=timeEnd-timeCurr;
        } else
            tstep=timeStep;

        if(result == COdeSystem::SR_REINIT) {   // split combined phase space into coordinates and momenta
            state_q.assign(stateCurr.begin(), stateCurr.begin()+numVars/2);
            state_p.assign(stateCurr.begin()+numVars/2, stateCurr.end());
        }
        stepper.do_step( 
                make_ode_wrapper_sympl(odeSystem, &tmp1, &tmp2), 
                state_q, state_p, timeCurr , tstep );
        nstep++;
        timeCurr+=tstep;
        // map symplectic vars back onto the combined phase space
        stateCurr.assign(state_q.begin(), state_q.end());
        stateCurr.insert(stateCurr.end(), state_p.begin(), state_p.end());
        result = odeSystem->processStep(timePrev, timeCurr, &stateCurr);
        if(result == COdeSystem::SR_TERMINATE) 
            needToTerminate=true;
        if(!gsl_finite(timeCurr) || fabs(timeCurr-timePrev) <= fabs(timeCurr) * GSL_DBL_EPSILON*10) needToTerminate=true;   // timestep too small
        if(nstep>NUMSTEP_MAX) needToTerminate=true;  // too many steps or stepsize adjustments
    }
}

// integration for fixed-timestep steppers
template< class Stepper >
void COdeIntegratorOdeint<Stepper>::integrateToTimeImpl(double timeEnd, 
    boost::numeric::odeint::stepper_tag)
{
    size_t nstep = 0;
    bool needToTerminate=false;
    double tstep=timeStep;
    while(!needToTerminate)
    {
        timePrev=timeCurr;
        statePrev=stateCurr;
        if(timeCurr<timeEnd && timeCurr+tstep > timeEnd) 
        {   // ensure that last step ends exactly at t_end
            tstep=timeEnd-timeCurr;
        } else
            tstep=timeStep;

        stepper.do_step( 
                make_ode_wrapper(odeSystem ), 
                stateCurr, timeCurr , tstep );
        nstep++;
        timeCurr+=tstep;
        COdeSystem::STEPRESULT result = odeSystem->processStep(timePrev, timeCurr, &stateCurr);
        if(result == COdeSystem::SR_TERMINATE) 
            needToTerminate=true;
        if(!gsl_finite(timeCurr) || fabs(timeCurr-timePrev) <= fabs(timeCurr) * GSL_DBL_EPSILON*10) needToTerminate=true;   // timestep too small
        if(nstep>NUMSTEP_MAX) needToTerminate=true;  // too many steps or stepsize adjustments
    }
}

// integration for adaptive-timestep steppers
template< class Stepper >
void COdeIntegratorOdeint<Stepper>::integrateToTimeImpl(double timeEnd, 
    boost::numeric::odeint::controlled_stepper_tag)
{
    double tstep=std::max<double>(timeEnd*1e-4, 1e-4);
    size_t nstep = 0;
    bool needToTerminate=false;
    while(!needToTerminate)
    {
        timePrev=timeCurr;
        statePrev=stateCurr;
        if(timeCurr+tstep > timeEnd) 
        {   // ensure that last step ends exactly at t_end
            tstep=timeEnd-timeCurr;
        }

        const size_t max_attempts=100;
        size_t trials = 0;
        boost::numeric::odeint::controlled_step_result res;
        do {
            res = stepper.try_step( 
                make_ode_wrapper(odeSystem ), 
                stateCurr, timeCurr , tstep );
            trials++;
        }
        while( ( res == boost::numeric::odeint::fail ) && ( trials < max_attempts ) );
        nstep++;
        COdeSystem::STEPRESULT result = odeSystem->processStep(timePrev, timeCurr, &stateCurr);
        if(result == COdeSystem::SR_TERMINATE) 
            needToTerminate=true;
        if(!gsl_finite(timeCurr) || fabs(timeCurr-timePrev) <= fabs(timeCurr) * GSL_DBL_EPSILON*10) needToTerminate=true;   // timestep too small
        if(nstep>NUMSTEP_MAX || trials>=max_attempts) needToTerminate=true;  // too many steps or stepsize adjustments
    }
}

// integration for dense-output adaptive timestep
template< class Stepper >
void COdeIntegratorOdeint<Stepper>::integrateToTimeImpl(double timeEnd, 
    boost::numeric::odeint::dense_output_stepper_tag)
{
    double tstep=std::max<double>(timeEnd*1e-4, 1e-4);
    size_t nstep = 0;
    stepper.initialize(stateCurr, 0, tstep);
    bool needToTerminate=false;
    //double t_add=0;
    while(!needToTerminate)
    {
        timePrev=timeCurr;
        if(timeCurr+stepper.current_time_step() > timeEnd) 
        {   // ensure that last step ends exactly at t_end
            stepper.initialize( stepper.current_state(), timeCurr, timeEnd-timeCurr );
        }
        paird timeInterval = stepper.do_step( make_ode_wrapper(odeSystem ) );
        nstep++;
        timeCurr=stepper.current_time();
        stateCurr=stepper.current_state();
        COdeSystem::STEPRESULT result = odeSystem->processStep(timePrev, timeCurr, &stateCurr);
        if(result == COdeSystem::SR_TERMINATE) 
            needToTerminate=true;
        if(result == COdeSystem::SR_REINIT)
            stepper.initialize( stateCurr, timeCurr, stepper.current_time_step() );
        if(!gsl_finite(timeCurr) || fabs(timeCurr-timePrev) <= fabs(timeCurr) * GSL_DBL_EPSILON*10) needToTerminate=true;   // timestep too small
        if(nstep>NUMSTEP_MAX) needToTerminate=true;  // too many steps
    }
}

// generic interpolation routine (calls the appropriate implementation based on whether dense output is available or not)
template< class Stepper >
double COdeIntegratorOdeint<Stepper>::getInterpolatedSolution(unsigned int c, double t) const
{
    if(t<timePrev || t>timeCurr || c>=stateCurr.size()) return gsl_nan();
    if(t==timeCurr) return stateCurr[c];
    return getInterpolatedSolutionImpl(c, t, stepper_category());
}

// linear interpolation
template< class Stepper >
double COdeIntegratorOdeint<Stepper>::getInterpolatedSolutionGeneric(unsigned int c, double t) const 
{
    return (statePrev[c] * (timeCurr-t) + stateCurr[c] * (t-timePrev)) / (timeCurr-timePrev);
}

// generic 3(2) order interpolation in coordinate(velocity) for Hamiltonian systems in the standard form, 
// using only the phase-space coordinates at the ends of interval
template< class Stepper >
double COdeIntegratorOdeint<Stepper>::getInterpolatedSolutionStdHamiltonian(unsigned int c, double t) const 
{
    unsigned int dim=stateCurr.size()/2;  // dimension of either coordinate or momentum vector
    double x_old=statePrev[c%dim];
    double x_new=stateCurr[c%dim];
    double v_old=statePrev[c%dim+dim];
    double v_new=stateCurr[c%dim+dim];
    double timestep=timeCurr-timePrev;
    double tau=(t-timePrev)/timestep;
    // compute acceleration and its derivative for the given coordinate
    double accel2 = 3*(x_new-x_old)-timestep*(2*v_old+v_new);
    double acder6 = 2*(x_old-x_new)+timestep*(v_old+v_new);
    // do 3rd order interpolation for coordinate or 2nd order for velocity
    if(c<dim)
        return x_old + tau*(v_old*timestep + tau*(accel2 + tau*acder6));
    else
        return v_old + tau/timestep*(2*accel2 + 3*tau*acder6);
}

// high-order interpolation for dense output methods
template< class Stepper >
double COdeIntegratorOdeint<Stepper>::getInterpolatedSolutionImpl(unsigned int c, double t, 
    boost::numeric::odeint::dense_output_stepper_tag) const
{
    if(timeIntermediate!=t) { // perform interpolation of the entire system to a given time
        stepper.calc_state(t, stateIntermediate);
        timeIntermediate=t;   // store (cache) state for subsequent calls
    }
    return stateIntermediate[c];
}
#endif

/* --- DOP853 high-accuracy Runge-Kutta integrator --- */
void COdeIntegratorDOP853::integrateToTime(double timeEnd)
{
    OdeStateType initVars;
    odeSystem->initState(&initVars);
    dop853(timePrev, initVars, timeEnd, accAbs, accRel);
}

/* ------ below follow dop853.c encapsulated routines ------- */

double COdeIntegratorDOP853::hinit (unsigned int n, double x, OdeStateType& y,
              double posneg, OdeStateType& f0, OdeStateType& f1, OdeStateType& yy1, 
              int iord, double hmax, double atoli, double rtoli)
{
  double   dnf, dny, sk, h, h1, der2, der12, sqre;
  unsigned int i;

  dnf = 0.0;
  dny = 0.0;

  for (i = 0; i < n; i++)
  {
      sk = atoli + rtoli * fabs(y[i]);
      if(sk==0) continue;
      sqre = f0[i] / sk;
      dnf += sqre*sqre;
      sqre = y[i] / sk;
      dny += sqre*sqre;
  }

  if (!gsl_finite(dnf+dny) || (dnf <= 1.0E-10) || (dny <= 1.0E-10))
    h = 1.0E-6;
  else
    h = sqrt (dny/dnf) * 0.01;

  h = std::min<double>(h, hmax);
  h *= posneg>=0?1:-1;

  /* perform an explicit Euler step */
  for (i = 0; i < n; i++)
    yy1[i] = y[i] + h * f0[i];
  fcn (n, x+h, yy1, f1);
  nfcn++;

  /* estimate the second derivative of the solution */
  der2 = 0.0;
  for (i = 0; i < n; i++)
  {
      sk = atoli + rtoli * fabs(y[i]);
      sqre = (f1[i] - f0[i]) / sk;
      if(sk!=0) der2 += sqre*sqre;
  }
  der2 = sqrt (der2) / h;

  /* step size is computed such that h**iord * max_d(norm(f0),norm(der2)) = 0.01 */
  der12 = std::max<double> (fabs(der2), sqrt(dnf));
  if (der12 <= 1.0E-15)
    h1 = std::max<double> (1.0E-6, fabs(h)*1.0E-3);
  else
    h1 = pow (0.01/der12, 1.0/(double)iord);
  h = std::min<double> (100.0 * h, std::min<double> (h1, hmax));

  return h*(posneg>=0?1:-1);

} /* hinit */

/* core integrator */
int COdeIntegratorDOP853::dop853(
    double x, OdeStateType &y, double timeEnd, double atoli, double rtoli)
{
  stateCurr = y;  // store the current values of variables
  /* initialisations */
  nfcn = nstep = naccpt = nrejct = 0;
  /* get dimensions of the system */
  unsigned int n=y.size();
  /* nstiff, parameter for stiffness detection */
  int nstiff = 1000;
  /* safety factor */
  double safe = 0.9;
  /* fac1, fac2, parameters for step size selection */
  double fac1 = 0.333;
  double fac2 = 6.0;
  /* beta for step control stabilization */
  double beta = 0.0;
  /* current timestep */
  double h = 0.0;
  /* max timestep */
  double hmax = timeEnd-x;

  double   facold, expo1, fac, facc1, facc2, fac11, posneg;
  double   hlamb, err, sk, hnew, ydiff, bspl;
  double   stnum, stden, sqre, err2, erri, deno;
  int      iasti, iord, reject, nonsti;
  unsigned int i;
  /* storate for intermediate steps - set size equal to the size of system */
  OdeStateType yy1(y), k1(y), k2(y), k3(y), k4(y), k5(y), k6(y), k7(y), k8(y), k9(y), k10(y);
  /* storage for dense output interpolation - assign size */
  rcont1=y; rcont2=y; rcont3=y; rcont4=y; rcont5=y; rcont6=y; rcont7=y; rcont8=y; 
  /* initializations */

      const double   // define the original numerical coefficients as constants, to speed up computations
      c2  = 0.526001519587677318785587544488E-01,
      c3  = 0.789002279381515978178381316732E-01,
      c4  = 0.118350341907227396726757197510E+00,
      c5  = 0.281649658092772603273242802490E+00,
      c6  = 0.333333333333333333333333333333E+00,
      c7  = 0.25E+00,
      c8  = 0.307692307692307692307692307692E+00,
      c9  = 0.651282051282051282051282051282E+00,
      c10 = 0.6E+00,
      c11 = 0.857142857142857142857142857142E+00,
      b1 =   5.42937341165687622380535766363E-2,
      b6 =   4.45031289275240888144113950566E0,
      b7 =   1.89151789931450038304281599044E0,
      b8 =  -5.8012039600105847814672114227E0,
      b9 =   3.1116436695781989440891606237E-1,
      b10 = -1.52160949662516078556178806805E-1,
      b11 =  2.01365400804030348374776537501E-1,
      b12 =  4.47106157277725905176885569043E-2,
      bhh1 = 0.244094488188976377952755905512E+00,
      bhh2 = 0.733846688281611857341361741547E+00,
      bhh3 = 0.220588235294117647058823529412E-01,
      er1  =  0.1312004499419488073250102996E-01,
      er6  = -0.1225156446376204440720569753E+01,
      er7  = -0.4957589496572501915214079952E+00,
      er8  =  0.1664377182454986536961530415E+01,
      er9  = -0.3503288487499736816886487290E+00,
      er10 =  0.3341791187130174790297318841E+00,
      er11 =  0.8192320648511571246570742613E-01,
      er12 = -0.2235530786388629525884427845E-01,
      a21 =    5.26001519587677318785587544488E-2,
      a31 =    1.97250569845378994544595329183E-2,
      a32 =    5.91751709536136983633785987549E-2,
      a41 =    2.95875854768068491816892993775E-2,
      a43 =    8.87627564304205475450678981324E-2,
      a51 =    2.41365134159266685502369798665E-1,
      a53 =   -8.84549479328286085344864962717E-1,
      a54 =    9.24834003261792003115737966543E-1,
      a61 =    3.7037037037037037037037037037E-2,
      a64 =    1.70828608729473871279604482173E-1,
      a65 =    1.25467687566822425016691814123E-1,
      a71 =    3.7109375E-2,
      a74 =    1.70252211019544039314978060272E-1,
      a75 =    6.02165389804559606850219397283E-2,
      a76 =   -1.7578125E-2,
      a81 =    3.70920001185047927108779319836E-2,
      a84 =    1.70383925712239993810214054705E-1,
      a85 =    1.07262030446373284651809199168E-1,
      a86 =   -1.53194377486244017527936158236E-2,
      a87 =    8.27378916381402288758473766002E-3,
      a91 =    6.24110958716075717114429577812E-1,
      a94 =   -3.36089262944694129406857109825E0,
      a95 =   -8.68219346841726006818189891453E-1,
      a96 =    2.75920996994467083049415600797E1,
      a97 =    2.01540675504778934086186788979E1,
      a98 =   -4.34898841810699588477366255144E1,
      a101 =   4.77662536438264365890433908527E-1,
      a104 =  -2.48811461997166764192642586468E0,
      a105 =  -5.90290826836842996371446475743E-1,
      a106 =   2.12300514481811942347288949897E1,
      a107 =   1.52792336328824235832596922938E1,
      a108 =  -3.32882109689848629194453265587E1,
      a109 =  -2.03312017085086261358222928593E-2,
      a111 =  -9.3714243008598732571704021658E-1,
      a114 =   5.18637242884406370830023853209E0,
      a115 =   1.09143734899672957818500254654E0,
      a116 =  -8.14978701074692612513997267357E0,
      a117 =  -1.85200656599969598641566180701E1,
      a118 =   2.27394870993505042818970056734E1,
      a119 =   2.49360555267965238987089396762E0,
      a1110 = -3.0467644718982195003823669022E0,
      a121 =   2.27331014751653820792359768449E0,
      a124 =  -1.05344954667372501984066689879E1,
      a125 =  -2.00087205822486249909675718444E0,
      a126 =  -1.79589318631187989172765950534E1,
      a127 =   2.79488845294199600508499808837E1,
      a128 =  -2.85899827713502369474065508674E0,
      a129 =  -8.87285693353062954433549289258E0,
      a1210 =  1.23605671757943030647266201528E1,
      a1211 =  6.43392746015763530355970484046E-1,
#ifndef DOP853_DENSE_OUTPUT_FAST  // original coefficients for 7th order interpolation
      a141 =  5.61675022830479523392909219681E-2,
      a147 =  2.53500210216624811088794765333E-1,
      a148 = -2.46239037470802489917441475441E-1,
      a149 = -1.24191423263816360469010140626E-1,
      a1410 =  1.5329179827876569731206322685E-1,
      a1411 =  8.20105229563468988491666602057E-3,
      a1412 =  7.56789766054569976138603589584E-3,
      a1413 = -8.298E-3,
      a151 =  3.18346481635021405060768473261E-2,
      a156 =  2.83009096723667755288322961402E-2,
      a157 =  5.35419883074385676223797384372E-2,
      a158 = -5.49237485713909884646569340306E-2,
      a1511 = -1.08347328697249322858509316994E-4,
      a1512 =  3.82571090835658412954920192323E-4,
      a1513 = -3.40465008687404560802977114492E-4,
      a1514 =  1.41312443674632500278074618366E-1,
      a161 = -4.28896301583791923408573538692E-1,
      a166 = -4.69762141536116384314449447206E0,
      a167 =  7.68342119606259904184240953878E0,
      a168 =  4.06898981839711007970213554331E0,
      a169 =  3.56727187455281109270669543021E-1,
      a1613 = -1.39902416515901462129418009734E-3,
      a1614 =  2.9475147891527723389556272149E0,
      a1615 = -9.15095847217987001081870187138E0,
      c14 = 0.1E+00,
      c15 = 0.2E+00,
      c16 = 0.777777777777777777777777777778E+00,
      d41  = -0.84289382761090128651353491142E+01,
      d46  =  0.56671495351937776962531783590E+00,
      d47  = -0.30689499459498916912797304727E+01,
      d48  =  0.23846676565120698287728149680E+01,
      d49  =  0.21170345824450282767155149946E+01,
      d410 = -0.87139158377797299206789907490E+00,
      d411 =  0.22404374302607882758541771650E+01,
      d412 =  0.63157877876946881815570249290E+00,
      d413 = -0.88990336451333310820698117400E-01,
      d414 =  0.18148505520854727256656404962E+02,
      d415 = -0.91946323924783554000451984436E+01,
      d416 = -0.44360363875948939664310572000E+01,
      d51  =  0.10427508642579134603413151009E+02,
      d56  =  0.24228349177525818288430175319E+03,
      d57  =  0.16520045171727028198505394887E+03,
      d58  = -0.37454675472269020279518312152E+03,
      d59  = -0.22113666853125306036270938578E+02,
      d510 =  0.77334326684722638389603898808E+01,
      d511 = -0.30674084731089398182061213626E+02,
      d512 = -0.93321305264302278729567221706E+01,
      d513 =  0.15697238121770843886131091075E+02,
      d514 = -0.31139403219565177677282850411E+02,
      d515 = -0.93529243588444783865713862664E+01,
      d516 =  0.35816841486394083752465898540E+02,
      d61 =  0.19985053242002433820987653617E+02,
      d66 = -0.38703730874935176555105901742E+03,
      d67 = -0.18917813819516756882830838328E+03,
      d68 =  0.52780815920542364900561016686E+03,
      d69 = -0.11573902539959630126141871134E+02,
      d610 =  0.68812326946963000169666922661E+01,
      d611 = -0.10006050966910838403183860980E+01,
      d612 =  0.77771377980534432092869265740E+00,
      d613 = -0.27782057523535084065932004339E+01,
      d614 = -0.60196695231264120758267380846E+02,
      d615 =  0.84320405506677161018159903784E+02,
      d616 =  0.11992291136182789328035130030E+02,
      d71  = -0.25693933462703749003312586129E+02,
      d76  = -0.15418974869023643374053993627E+03,
      d77  = -0.23152937917604549567536039109E+03,
      d78  =  0.35763911791061412378285349910E+03,
      d79  =  0.93405324183624310003907691704E+02,
      d710 = -0.37458323136451633156875139351E+02,
      d711 =  0.10409964950896230045147246184E+03,
      d712 =  0.29840293426660503123344363579E+02,
      d713 = -0.43533456590011143754432175058E+02,
      d714 =  0.96324553959188282948394950600E+02,
      d715 = -0.39177261675615439165231486172E+02,
      d716 = -0.14972683625798562581422125276E+03;
#else  // modified coefficients for 6th order interpolation
      d41  = -5.40685903845352664250302,
      d46  = +367.268892700041893590281,
      d47  = +154.609958204083905482676,
      d48  = -505.920283865412564024766,
      d49  = +15.5975154819608130688200,
      d410 = -26.1936204184402805956691,
      d411 = -.740035123641222308447206,
      d412 = +1.11776539319431476294221,
      d413 = -.333333333333333333333333,
      d51  = +6.51987095363079615048119,
      d56  = -1066.34956011730205278592,
      d57  = -351.864047514639508625601,
      d58  = +1363.51955696662884408368,
      d59  = -112.727669432657582669864,
      d510 = +159.796191868560289612921,
      d511 = -2.13865100308788816220259,
      d512 = -3.75569172113289760348584,
      d513 = +7.00000000000000000000000,
      d61  = +10.4698004763293477204238,
      d66  = -1380.01473607038123167155,
      d67  = -531.219827862514074379012,
      d68  = +1866.98964341870892451324,
      d69  = -53.3302605020547902574560,
      d610 = +82.4147560258671369782481,
      d611 = +7.38443654502992069572676,
      d612 = +.417299080125877511498426,
      d613 = -3.11111111111111111111111,
      d71  = -16.6338582677165354330709,
      d76  = +4516.16568914956011730205,
      d77  = +1393.85185384057776465219,
      d78  = -5687.52042419481539670071,
      d79  = +473.965563750151263163661,
      d710 = -661.810776942355889724311,
      d711 = -18.0180473354013232598119,
      d712 = 0, d713 = 0;
#endif
  facold = 1.0E-4;
  expo1 = 1.0/8.0 - beta * 0.2;
  facc1 = 1.0 / fac1;
  facc2 = 1.0 / fac2;
  posneg = timeEnd>x ? 1.0 : -1.0;

  /* initial preparations */
  double xadd=0;  // addendum for time variable (to avoid roundoff errors if timestep is very small and simulation time is already large)
  hlamb = 0.0;
  iasti = 0;
  nonsti= 0;
  timePrev = x;
  timeCurr = x;
  hmax = fabs (hmax);
  iord = 8;
  fcn (n, x, y, k1);
  nfcn++;
  h = hinit (n, x, y, posneg, k1, k2, k3, iord, hmax, atoli, rtoli);
  reject = 0;

  /* basic integration step */
  while (1)
  {
    if (nstep > NUMSTEP_MAX || !gsl_finite(h) || h==0)
    {
      timeCurr = x+xadd;
      hout = h;
      return -2;
    }

    if (fabs(h) <= 10 * fabs(x) * GSL_DBL_EPSILON)
    {
      double add=1.0;
      while(add*2<x) add*=2;   // find the largest multiple of 2 that is less than x
      while(add>x) add/=2; 
      if(add<=xadd*GSL_DBL_EPSILON) {  // cannot increase addendum without roundoff becoming too large
        timeCurr = x+xadd;
        hout = h;
        return -2;
      }
      xadd+=add;
      x-=add;
      // reinitialize derivative for the current value of y
      fcn (n, x+xadd, y, k1);
      nfcn++;
    }

    if(x+xadd<timeEnd && x+xadd+h>=timeEnd) 
        h=timeEnd-(x+xadd);    // last step exactly at x_end

    nstep++;

    /* the twelve stages */
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * a21 * k1[i];
    fcn (n, x+xadd+c2*h, yy1, k2);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a31*k1[i] + a32*k2[i]);
    fcn (n, x+xadd+c3*h, yy1, k3);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a41*k1[i] + a43*k3[i]);
    fcn (n, x+xadd+c4*h, yy1, k4);
    for (i = 0; i <n; i++)
      yy1[i] = y[i] + h * (a51*k1[i] + a53*k3[i] + a54*k4[i]);
    fcn (n, x+xadd+c5*h, yy1, k5);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a61*k1[i] + a64*k4[i] + a65*k5[i]);
    fcn (n, x+xadd+c6*h, yy1, k6);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a71*k1[i] + a74*k4[i] + a75*k5[i] + a76*k6[i]);
    fcn (n, x+xadd+c7*h, yy1, k7);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a81*k1[i] + a84*k4[i] + a85*k5[i] + a86*k6[i] +
                          a87*k7[i]);
    fcn (n, x+xadd+c8*h, yy1, k8);
    for (i = 0; i <n; i++)
      yy1[i] = y[i] + h * (a91*k1[i] + a94*k4[i] + a95*k5[i] + a96*k6[i] +
                          a97*k7[i] + a98*k8[i]);
    fcn (n, x+xadd+c9*h, yy1, k9);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a101*k1[i] + a104*k4[i] + a105*k5[i] + a106*k6[i] +
                          a107*k7[i] + a108*k8[i] + a109*k9[i]);
    fcn (n, x+xadd+c10*h, yy1, k10);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a111*k1[i] + a114*k4[i] + a115*k5[i] + a116*k6[i] +
                          a117*k7[i] + a118*k8[i] + a119*k9[i] + a1110*k10[i]);
    fcn (n, x+xadd+c11*h, yy1, k2);
    for (i = 0; i < n; i++)
      yy1[i] = y[i] + h * (a121*k1[i] + a124*k4[i] + a125*k5[i] + a126*k6[i] +
                          a127*k7[i] + a128*k8[i] + a129*k9[i] +
                          a1210*k10[i] + a1211*k2[i]);
    fcn (n, x+xadd+h, yy1, k3);
    nfcn += 11;
    for (i = 0; i < n; i++)
    {
      k4[i] = b1*k1[i] + b6*k6[i] + b7*k7[i] + b8*k8[i] + b9*k9[i] +
              b10*k10[i] + b11*k2[i] + b12*k3[i];
      k5[i] = y[i] + h * k4[i];
    }
     
    /* error estimation */
    err = 0.0;
    err2 = 0.0;
    double rtol=rtoli * odeSystem->getRelErrorToleranceFactor(y);  // provision for having smaller tolerance if necessary

    for (i = 0; i < n; i++)
    {
        sk = atoli + rtol * std::max<double> (fabs(y[i]), fabs(k5[i]));
        if(sk==0) continue;
        erri = k4[i] - bhh1*k1[i] - bhh2*k9[i] - bhh3*k3[i];
        sqre = erri / sk;
        err2 += sqre*sqre;
        erri = er1*k1[i] + er6*k6[i] + er7*k7[i] + er8*k8[i] + er9*k9[i] +
               er10 * k10[i] + er11*k2[i] + er12*k3[i];
        sqre = erri / sk;
        err += sqre*sqre;
    }
    deno = err + 0.01 * err2;
    if (deno <= 0.0)
      deno = 1.0;
    err = fabs(h) * err * sqrt (1.0 / (deno*(double)n));

    /* computation of hnew */
    fac11 = pow (err, expo1);
    /* Lund-stabilization */
    fac = fac11 / pow(facold,beta);
    /* we require fac1 <= hnew/h <= fac2 */
    fac = std::max<double> (facc2, std::min<double> (facc1, fac/safe));
    hnew = h / fac;

    if (err <= 1.0)
    {
      /* step accepted */

      facold = std::max<double> (err, 1.0E-4);
      naccpt++;
      fcn (n, x+xadd+h, k5, k4);
      nfcn++;
      
      /* stiffness detection */
      if (!(naccpt % nstiff) || (iasti > 0))
      {
        stnum = 0.0;
        stden = 0.0;
        for (i = 0; i < n; i++)
        {
          sqre = k4[i] - k3[i];
          stnum += sqre*sqre;
          sqre = k5[i] - yy1[i];
          stden += sqre*sqre;
        }
        if (stden > 0.0)
          hlamb = h * sqrt (stnum / stden);
        if (hlamb > 6.1)
        {
          nonsti = 0;
          iasti++;
          if (iasti == 15)
            /*if (fileout)
              fprintf (fileout, "The problem seems to become stiff at x = %.16e\r\n", x);
            else*/
            {
              timeCurr = x+xadd;
              hout = h;
              return -4;
            }
        }
        else
        {
          nonsti++;
          if (nonsti == 6)
            iasti = 0;
        }
      }
      /* final preparation for dense output */
        /* save the first function evaluations */
          for (i = 0; i < n; i++)
          {
            rcont1[i] = y[i];
            ydiff = k5[i] - y[i];
            rcont2[i] = ydiff;
            bspl = h * k1[i] - ydiff;
            rcont3[i] = bspl;
            rcont4[i] = ydiff - h*k4[i] - bspl;
            rcont5[i] = h * (d41*k1[i] + d46*k6[i] + d47*k7[i] + d48*k8[i] +
                d49*k9[i] + d410*k10[i]+ d411*k2[i]+ d412*k3[i]+ d413*k4[i]);
            rcont6[i] = h * (d51*k1[i] + d56*k6[i] + d57*k7[i] + d58*k8[i] +
                d59*k9[i] + d510*k10[i]+ d511*k2[i]+ d512*k3[i]+ d513*k4[i]);
            rcont7[i] = h * (d61*k1[i] + d66*k6[i] + d67*k7[i] + d68*k8[i] +
                d69*k9[i] + d610*k10[i]+ d611*k2[i]+ d612*k3[i]+ d613*k4[i]);
            rcont8[i] = h * (d71*k1[i] + d76*k6[i] + d77*k7[i] + d78*k8[i] +
                d79*k9[i] + d710*k10[i]+ d711*k2[i]+ d712*k3[i]+ d713*k4[i]);
          }
#ifndef DOP853_DENSE_OUTPUT_FAST
        /* the next three function evaluations */
        for (i = 0; i < n; i++)
          yy1[i] = y[i] + h * (a141*k1[i] + a147*k7[i] + a148*k8[i] +
                              a149*k9[i] + a1410*k10[i] + a1411*k2[i] +
                              a1412*k3[i] + a1413*k4[i]);
        fcn (n, x+xadd+c14*h, yy1, k10);
        for (i = 0; i < n; i++)
          yy1[i] = y[i] + h * (a151*k1[i] + a156*k6[i] + a157*k7[i] + a158*k8[i] +
                              a1511*k2[i] + a1512*k3[i] + a1513*k4[i] +
                              a1514*k10[i]);
        fcn (n, x+xadd+c15*h, yy1, k2);
        for (i = 0; i < n; i++)
          yy1[i] = y[i] + h * (a161*k1[i] + a166*k6[i] + a167*k7[i] + a168*k8[i] +
                              a169*k9[i] + a1613*k4[i] + a1614*k10[i] +
                              a1615*k2[i]);
        fcn (n, x+xadd+c16*h, yy1, k3);
        nfcn += 3;
        /* final preparation */
          for (i = 0; i < n; i++)
          {
            rcont5[i] += h * (d414*k10[i] + d415*k2[i] + d416*k3[i]);
            rcont6[i] += h * (d514*k10[i] + d515*k2[i] + d516*k3[i]);
            rcont7[i] += h * (d614*k10[i] + d615*k2[i] + d616*k3[i]);
            rcont8[i] += h * (d714*k10[i] + d715*k2[i] + d716*k3[i]);
          }
#endif
      //memcpy (k1, k4, n * sizeof(double)); 
      //memcpy (y, k5, n * sizeof(double));
      k1=k4;
      y =k5;
      x += h;
      hout = h;
      timePrev = timeCurr;
      timeCurr = x+xadd;   // this may introduce roundoff errors!
      stateCurr = y;  // store the current values of variables
      COdeSystem::STEPRESULT result =   // ensure that the timestep function will be called only if rounded-off time interval is positive
          timeCurr>timePrev ? odeSystem->processStep(timePrev, timeCurr, &y) : COdeSystem::SR_CONTINUE;
      if(result == COdeSystem::SR_TERMINATE)
          return 1;  /* normal exit */

      if(result == COdeSystem::SR_REINIT)
      {
          // copy back the new current state (the variable "y" is internally used in this function 
          // while "stateCurr" is used in getInterpolatedSolution. 
          // For the end point of the trajectory to hold the true end point after all runtime function 
          // processed the step, it is necessary for stateCurr to contain this final value.
          stateCurr = y;
          // recompute derivative for the current value of y
          fcn (n, x+xadd, y, k1);
          nfcn++;
      }

      if (fabs(hnew) > hmax)
        hnew = posneg * hmax;
      if (reject)
        hnew = posneg * std::min<double> (fabs(hnew), fabs(h));

      reject = 0;

    }
    else
    {
      /* step rejected */
      hnew = h / std::min<double> (facc1, fac11/safe);
      reject = 1;
      if (naccpt >= 1)
        nrejct=nrejct + 1;
    }

    h = hnew;
  }

} /* dop853 */

/* dense output function */
double COdeIntegratorDOP853::getInterpolatedSolution (unsigned int i, double t) const
{
  if(t<timePrev || t>timeCurr || i>=rcont1.size()) return gsl_nan();
  if(t==timeCurr) return stateCurr[i];
  double s = (t - timePrev) / hout;
  double s1= 1.0 - s;
  return rcont1[i]+s*(rcont2[i]+s1*(rcont3[i]+s*(rcont4[i]
      +s1*(rcont5[i]+s*(rcont6[i]+s1*(rcont7[i]+s*rcont8[i]))))));
} /* contd8 */

/* ----- IAS15 ----- */

void COdeIntegratorIAS15::integrateToTime(double timeEnd)
{
    if(!odeSystem->isStdHamiltonian()) {
       // my_error(FUNCNAME, "IAS15 can only deal with standard Hamiltonian variables (coordinate+velocity)");
        return;
    }
    odeSystem->initState(&stateCurr);
    statePrev=stateCurr;
    deriv.assign(stateCurr.size(),0);
    int N3 = stateCurr.size()/2;
    for (int l=0;l<7;++l) {
        g[l].assign(N3, 0);
        b[l].assign(N3, 0);
        e[l].assign(N3, 0);
        br[l].assign(N3,0);
        er[l].assign(N3,0);
    }
    at.resize(N3);
    a0.resize(N3);
    compsum.assign(stateCurr.size(), 0);
    nstep=naccpt=nrejct=nfcn=0;
    dt_last_success=dt=std::max<double>(timeEnd*1e-4, 1e-4); // maybe assign initial timestep in a more fancy way?
    while(timeCurr<timeEnd) {
        nstep++;
        if(integrator_ias15_step()) {  // step successful
            naccpt++;
            COdeSystem::STEPRESULT result =   // ensure that the timestep function will be called only if rounded-off time interval is positive
                timeCurr>timePrev ? odeSystem->processStep(timePrev, timeCurr, &stateCurr) : COdeSystem::SR_CONTINUE;
            if(result == COdeSystem::SR_TERMINATE)
                return;
            if(timeCurr==timePrev)  // timestep is too small, integration does not advance
                return;
            if(timeCurr<timeEnd && timeCurr+dt>timeEnd)
                dt=timeEnd-timeCurr; // end exactly at desired time
        } else {
            nrejct++;  // will repeat with a smaller timestep
        }
    }
}

int COdeIntegratorIAS15::integrator_ias15_step() 
{
    // Gauss Radau spacings
    const double h[8]  = { 0.0, 0.0562625605369221464656521910, 0.1802406917368923649875799428, 0.3526247171131696373739077702, 0.5471536263305553830014485577, 0.7342101772154105410531523211, 0.8853209468390957680903597629, 0.9775206135612875018911745004}; 
    // Other constants
    const double r[28] = {0.0562625605369221464656522, 0.1802406917368923649875799, 0.1239781311999702185219278, 0.3526247171131696373739078, 0.2963621565762474909082556, 0.1723840253762772723863278, 0.5471536263305553830014486, 0.4908910657936332365357964, 0.3669129345936630180138686, 0.1945289092173857456275408, 0.7342101772154105410531523, 0.6779476166784883945875001, 0.5539694854785181760655724, 0.3815854601022409036792446, 0.1870565508848551580517038, 0.8853209468390957680903598, 0.8290583863021736216247076, 0.7050802551022034031027798, 0.5326962297259261307164520, 0.3381673205085403850889112, 0.1511107696236852270372074, 0.9775206135612875018911745, 0.9212580530243653554255223, 0.7972799218243951369035946, 0.6248958964481178645172667, 0.4303669872307321188897259, 0.2433104363458769608380222, 0.0921996667221917338008147};
    const double c[21] = {-0.0562625605369221464656522, 0.0101408028300636299864818, -0.2365032522738145114532321, -0.0035758977292516175949345, 0.0935376952594620658957485, -0.5891279693869841488271399, 0.0019565654099472210769006, -0.0547553868890686864408084, 0.4158812000823068616886219, -1.1362815957175395318285885, -0.0014365302363708915610919, 0.0421585277212687082291130, -0.3600995965020568162530901, 1.2501507118406910366792415, -1.8704917729329500728817408, 0.0012717903090268677658020, -0.0387603579159067708505249, 0.3609622434528459872559689, -1.4668842084004269779203515, 2.9061362593084293206895457, -2.7558127197720458409721005};
    const double d[21] = {0.0562625605369221464656522, 0.0031654757181708292499905, 0.2365032522738145114532321, 0.0001780977692217433881125, 0.0457929855060279188954539, 0.5891279693869841488271399, 0.0000100202365223291272096, 0.0084318571535257015445000, 0.2535340690545692665214616, 1.1362815957175395318285885, 0.0000005637641639318207610, 0.0015297840025004658189490, 0.0978342365324440053653648, 0.8752546646840910912297246, 1.8704917729329500728817408, 0.0000000317188154017613665, 0.0002762930909826476593130, 0.0360285539837364596003871, 0.5767330002770787313544596, 2.2485887607691598182153473, 2.7558127197720458409721005};
    const double safety_factor = 0.5;    // Maximum increase/deacrease of consecutve timesteps. (rebound orig.=0.25)
    const double max_predictor_corrector_error = 1e-8;  // convergence criterion (rebound orig.=1e-16)
    const bool integrator_epsilon_global=false;  // left in place from rebound
    double s[9];                // Summation coefficients 

    timePrev=timeCurr;
    statePrev=stateCurr;
    int N3=stateCurr.size()/2;

    odeSystem->odeFnc(timeCurr, stateCurr, &deriv);
    nfcn++;
    for(int k=0; k<N3; k++) {
        a0[k] = deriv[k+N3];
    }

    for(int k=0;k<N3;k++) {
        g[0][k] = b[6][k]*d[15] + b[5][k]*d[10] + b[4][k]*d[6] + b[3][k]*d[3]  + b[2][k]*d[1]  + b[1][k]*d[0]  + b[0][k];
        g[1][k] = b[6][k]*d[16] + b[5][k]*d[11] + b[4][k]*d[7] + b[3][k]*d[4]  + b[2][k]*d[2]  + b[1][k];
        g[2][k] = b[6][k]*d[17] + b[5][k]*d[12] + b[4][k]*d[8] + b[3][k]*d[5]  + b[2][k];
        g[3][k] = b[6][k]*d[18] + b[5][k]*d[13] + b[4][k]*d[9] + b[3][k];
        g[4][k] = b[6][k]*d[19] + b[5][k]*d[14] + b[4][k];
        g[5][k] = b[6][k]*d[20] + b[5][k];
        g[6][k] = b[6][k];
    }

    double predictor_corrector_error = 1e300;
    double predictor_corrector_error_last = 2;
    int iterations = 0;    
    // Predictor corrector loop
    // Stops if one of the following conditions is satisfied: 
    //   1) predictor_corrector_error better than the threshold 
    //   2) predictor_corrector_error starts to oscillate
    //   3) more than 12 iterations
    while(1) {
        if(predictor_corrector_error<max_predictor_corrector_error) {
            break;
        }
        if(iterations > 2 && predictor_corrector_error_last <= predictor_corrector_error) {
            break;
        }
        if (iterations>=12){
            /*integrator_iterations_max_exceeded++;
            const int integrator_iterations_warning = 10;
            if (integrator_iterations_max_exceeded==integrator_iterations_warning ){
                my_message(FUNCNAME, "At least "+convertToString(integrator_iterations_warning)+
                    " predictor corrector loops in integrator IAS15 did not converge. This is typically an indication of the timestep being too large.");
            }*/
            break;   // Quit predictor corrector loop, will result in abandoning this step and repeating with a smaller timestep
        }
        predictor_corrector_error_last = predictor_corrector_error;
        predictor_corrector_error = 0;
        iterations++;

        for(int n=1;n<8;n++) {                            // Loop over interval using Gauss-Radau spacings

            s[0] = dt * h[n];
            s[1] = s[0] * s[0] / 2.;
            s[2] = s[1] * h[n] / 3.;
            s[3] = s[2] * h[n] / 2.;
            s[4] = 3. * s[3] * h[n] / 5.;
            s[5] = 2. * s[4] * h[n] / 3.;
            s[6] = 5. * s[5] * h[n] / 7.;
            s[7] = 3. * s[6] * h[n] / 4.;
            s[8] = 7. * s[7] * h[n] / 9.;

            // Prepare particles arrays for force calculation
            for(int k=0;k<N3;k++) {  // Predict positions at interval n using b values
                double xk = compsum[k] + (s[8]*b[6][k] + s[7]*b[5][k] + s[6]*b[4][k] + s[5]*b[3][k] + 
                    s[4]*b[2][k] + s[3]*b[1][k] + s[2]*b[0][k] + s[1]*a0[k] + s[0]*statePrev[N3+k] );
                stateCurr[k] = xk + statePrev[k];
            }
            
            if(true /*problem_additional_forces && integrator_force_is_velocitydependent*/){
                s[0] = dt * h[n];
                s[1] =      s[0] * h[n] / 2.;
                s[2] = 2. * s[1] * h[n] / 3.;
                s[3] = 3. * s[2] * h[n] / 4.;
                s[4] = 4. * s[3] * h[n] / 5.;
                s[5] = 5. * s[4] * h[n] / 6.;
                s[6] = 6. * s[5] * h[n] / 7.;
                s[7] = 7. * s[6] * h[n] / 8.;

                for(int k=0;k<N3;k++) {  // Predict velocities at interval n using b values
                    double vk =  compsum[k+N3] + s[7]*b[6][k] + s[6]*b[5][k] + s[5]*b[4][k] + s[4]*b[3][k] + 
                        s[3]*b[2][k] + s[2]*b[1][k] + s[1]*b[0][k] + s[0]*a0[k];
                    stateCurr[k+N3] = vk + statePrev[N3+k];
                }
            }

            odeSystem->odeFnc(timeCurr+dt*h[n], stateCurr, &deriv);
            nfcn++;
            for(int k=0; k<N3; k++) {
                at[k] = deriv[k+N3];
            }

            switch(n) {                            // Improve b and g values
                case 1: 
                    for(int k=0;k<N3;++k) {
                        double tmp = g[0][k];
                        g[0][k]  = (at[k] - a0[k]) / r[0];
                        b[0][k] += g[0][k] - tmp;
                    } break;
                case 2: 
                    for(int k=0;k<N3;++k) {
                        double tmp = g[1][k];
                        const double gk = at[k] - a0[k];
                        g[1][k] = (gk/r[1] - g[0][k])/r[2];
                        tmp = g[1][k] - tmp;
                        b[0][k] += tmp * c[0];
                        b[1][k] += tmp;
                    } break;
                case 3: 
                    for(int k=0;k<N3;++k) {
                        double tmp = g[2][k];
                        const double gk = at[k] - a0[k];
                        g[2][k] = ((gk/r[3] - g[0][k])/r[4] - g[1][k])/r[5];
                        tmp = g[2][k] - tmp;
                        b[0][k] += tmp * c[1];
                        b[1][k] += tmp * c[2];
                        b[2][k] += tmp;
                    } break;
                case 4:
                    for(int k=0;k<N3;++k) {
                        double tmp = g[3][k];
                        const double gk = at[k] - a0[k];
                        g[3][k] = (((gk/r[6] - g[0][k])/r[7] - g[1][k])/r[8] - g[2][k])/r[9];
                        tmp = g[3][k] - tmp;
                        b[0][k] += tmp * c[3];
                        b[1][k] += tmp * c[4];
                        b[2][k] += tmp * c[5];
                        b[3][k] += tmp;
                    } break;
                case 5:
                    for(int k=0;k<N3;++k) {
                        double tmp = g[4][k];
                        const double gk = at[k] - a0[k];
                        g[4][k] = ((((gk/r[10] - g[0][k])/r[11] - g[1][k])/r[12] - g[2][k])/r[13] - g[3][k])/r[14];
                        tmp = g[4][k] - tmp;
                        b[0][k] += tmp * c[6];
                        b[1][k] += tmp * c[7];
                        b[2][k] += tmp * c[8];
                        b[3][k] += tmp * c[9];
                        b[4][k] += tmp;
                    } break;
                case 6:
                    for(int k=0;k<N3;++k) {
                        double tmp = g[5][k];
                        const double gk = at[k] - a0[k];
                        g[5][k] = (((((gk/r[15] - g[0][k])/r[16] - g[1][k])/r[17] - g[2][k])/r[18] - g[3][k])/r[19] - g[4][k])/r[20];
                        tmp = g[5][k] - tmp;
                        b[0][k] += tmp * c[10];
                        b[1][k] += tmp * c[11];
                        b[2][k] += tmp * c[12];
                        b[3][k] += tmp * c[13];
                        b[4][k] += tmp * c[14];
                        b[5][k] += tmp;
                    } break;
                case 7:
                {
                    double maxak = 0.0;
                    double maxb6ktmp = 0.0;
                    for(int k=0;k<N3;++k) {
                        double tmp = g[6][k];
                        const double gk = at[k] - a0[k];
                        g[6][k] = ((((((gk/r[21] - g[0][k])/r[22] - g[1][k])/r[23] - g[2][k])/r[24] - g[3][k])/r[25] - g[4][k])/r[26] - g[5][k])/r[27];
                        tmp = g[6][k] - tmp;    
                        b[0][k] += tmp * c[15];
                        b[1][k] += tmp * c[16];
                        b[2][k] += tmp * c[17];
                        b[3][k] += tmp * c[18];
                        b[4][k] += tmp * c[19];
                        b[5][k] += tmp * c[20];
                        b[6][k] += tmp;
                        
                        // Monitor change in b[6][k] relative to at[k]. The predictor corrector scheme is converged if it is close to 0.
                        if(integrator_epsilon_global) {
                            const double ak  = fabs(at[k]);
                            if(gsl_finite(ak) && ak>maxak) {
                                maxak = ak;
                            }
                            const double b6ktmp = fabs(tmp);  // change of b6ktmp coefficient
                            if(gsl_finite(b6ktmp) && b6ktmp>maxb6ktmp) {
                                maxb6ktmp = b6ktmp;
                            }
                        } else {
                            const double ak  = at[k];
                            const double b6ktmp = tmp; 
                            const double errork = fabs(b6ktmp/ak);
                            if(gsl_finite(errork) && errork>predictor_corrector_error) {
                                predictor_corrector_error = errork;
                            }
                        }
                    } 
                    if(integrator_epsilon_global) {
                        predictor_corrector_error = maxb6ktmp/maxak;
                    }
                    
                    break;
                }
            }
        }
    }
    // Find new timestep
    const double dt_done = dt;
    
    if(integrator_epsilon>0) {
        // Estimate error (given by last term in series expansion) 
        // There are two options:
        // integrator_epsilon_global==1  (default)
        //   First, we determine the maximum acceleration and the maximum of the last term in the series. 
        //   Then, the two are divided.
        // integrator_epsilon_global==0
        //   Here, the fractional error is calculated for each particle individually and we use the maximum of the fractional error.
        //   This might fail in cases where a particle does not experience any (physical) acceleration besides roundoff errors. 
        double integrator_error = 0.0;
        if(integrator_epsilon_global) {
            /*double maxak = 0.0;
            double maxb6k = 0.0;
            for(int i=0;i<N;i++){ // Looping over all particles and all 3 components of the acceleration. 
                const double v2 = particles[i].vx*particles[i].vx+particles[i].vy*particles[i].vy+particles[i].vz*particles[i].vz;
                const double x2 = particles[i].x*particles[i].x+particles[i].y*particles[i].y+particles[i].z*particles[i].z;
                // Skip slowly varying accelerations
                if (fabs(v2*dt*dt/x2) < 1e-16) continue;
                for(int k=3*i;k<3*(i+1);k++) { 
                    const double ak  = fabs(at[k]);
                    if (gsl_finite(ak) && ak>maxak){
                        maxak = ak;
                    }
                    const double b6k = fabs(b[6][k]); 
                    if (gsl_finite(b6k) && b6k>maxb6k){
                        maxb6k = b6k;
                    }
                }
            }
            integrator_error = maxb6k/maxak;*/
        } else {
            for(int k=0;k<N3;k++) {
                const double ak  = at[k];
                const double b6k = b[6][k]; 
                const double errork = fabs(b6k/ak);
                if (gsl_finite(errork) && errork>integrator_error){
                    integrator_error = errork;
                }
            }
        }

        double dt_new;
        if(gsl_finite(integrator_error)) {
            // if error estimate is available increase by more educated guess
             dt_new = pow(integrator_epsilon/integrator_error,1./7.)*dt_done;
        } else {                    // In the rare case that the error estimate doesn't give a finite number (e.g. when all forces accidentally cancel up to machine precission).
             dt_new = dt_done/safety_factor; // by default, increase timestep a little
        }
        
        if(fabs(dt_new/dt_done) < safety_factor) {    // New timestep is significantly smaller.
            // Reset particles
            stateCurr=statePrev;
            dt = dt_new;
            double ratio = dt/dt_last_success;
            predict_next_step(ratio, N3, er, br);
            
            return 0; // Step rejected. Do again. 
        }        
        if(fabs(dt_new/dt_done) > 1.0) {    // New timestep is larger.
            if(dt_new/dt_done > 1./safety_factor) 
                dt_new = dt_done /safety_factor;    // Don't increase the timestep by too much compared to the last one.
        }
        dt = dt_new;
    }

    // Find new position and velocity values at end of the sequence
    const double dt_done2 = dt_done * dt_done;
    for(int k=0;k<N3;++k) {
        double a = statePrev[k];
        compsum[k] += (b[6][k]/72. + b[5][k]/56. + b[4][k]/42. + b[3][k]/30. + b[2][k]/20. + b[1][k]/12. + 
            b[0][k]/6. + a0[k]/2.) * dt_done2 + statePrev[N3+k] * dt_done;
        stateCurr[k] = a + compsum[k];
        compsum[k] += a - stateCurr[k]; 

        a = statePrev[k+N3]; 
        compsum[k+N3] += (b[6][k]/8. + b[5][k]/7. + b[4][k]/6. + b[3][k]/5. + b[2][k]/4. + b[1][k]/3. +
            b[0][k]/2. + a0[k]) * dt_done;
        stateCurr[N3+k] = a + compsum[k+N3];
        compsum[k+N3] += a - stateCurr[N3+k];
    }

    timeCurr += dt_done;
    dt_last_success = dt_done;
    for(int j=0; j<7; j++) {
        er[j]=e[j];  // copy entire array
        br[j]=b[j];
    }
    double ratio = dt/dt_done;
    predict_next_step(ratio, N3, e, b);
    return 1; // Success.
}

void COdeIntegratorIAS15::predict_next_step(double ratio, int N3, vectord _e[7], vectord _b[7])
{
    // Predict new B values to use at the start of the next sequence. The predicted
    // values from the last call are saved as E. The correction, BD, between the
    // actual and predicted values of B is applied in advance as a correction.
    const double q1 = ratio;
    const double q2 = q1 * q1;
    const double q3 = q1 * q2;
    const double q4 = q2 * q2;
    const double q5 = q2 * q3;
    const double q6 = q3 * q3;
    const double q7 = q3 * q4;

    for(int k=0;k<N3;++k) {
        double be0 = _b[0][k] - _e[0][k];
        double be1 = _b[1][k] - _e[1][k];
        double be2 = _b[2][k] - _e[2][k];
        double be3 = _b[3][k] - _e[3][k];
        double be4 = _b[4][k] - _e[4][k];
        double be5 = _b[5][k] - _e[5][k];
        double be6 = _b[6][k] - _e[6][k];

        e[0][k] = q1*(_b[6][k]* 7.0 + _b[5][k]* 6.0 + _b[4][k]* 5.0 + _b[3][k]* 4.0 + _b[2][k]* 3.0 + _b[1][k]*2.0 + _b[0][k]);
        e[1][k] = q2*(_b[6][k]*21.0 + _b[5][k]*15.0 + _b[4][k]*10.0 + _b[3][k]* 6.0 + _b[2][k]* 3.0 + _b[1][k]);
        e[2][k] = q3*(_b[6][k]*35.0 + _b[5][k]*20.0 + _b[4][k]*10.0 + _b[3][k]* 4.0 + _b[2][k]);
        e[3][k] = q4*(_b[6][k]*35.0 + _b[5][k]*15.0 + _b[4][k]* 5.0 + _b[3][k]);
        e[4][k] = q5*(_b[6][k]*21.0 + _b[5][k]* 6.0 + _b[4][k]);
        e[5][k] = q6*(_b[6][k]* 7.0 + _b[5][k]);
        e[6][k] = q7* _b[6][k];

        b[0][k] = e[0][k] + be0;
        b[1][k] = e[1][k] + be1;
        b[2][k] = e[2][k] + be2;
        b[3][k] = e[3][k] + be3;
        b[4][k] = e[4][k] + be4;
        b[5][k] = e[5][k] + be5;
        b[6][k] = e[6][k] + be6;
    }
}

double COdeIntegratorIAS15::getInterpolatedSolution(unsigned int c, double t) const 
{
    unsigned int dim=stateCurr.size()/2;  // dimension of either coordinate or momentum vector
    unsigned int k=c%dim;
    double timestep=timeCurr-timePrev;
    double tau=(t-timePrev)/timestep;
    double s[9]; // temp.array
    if(c<dim) {  // interpolate position
        s[0] = timestep * tau;
        s[1] = s[0] * s[0] / 2.;
        s[2] = s[1] * tau / 3.;
        s[3] = s[2] * tau / 2.;
        s[4] = 3. * s[3] * tau / 5.;
        s[5] = 2. * s[4] * tau / 3.;
        s[6] = 5. * s[5] * tau / 7.;
        s[7] = 3. * s[6] * tau / 4.;
        s[8] = 7. * s[7] * tau / 9.;
        return compsum[c] + (s[8]*br[6][k] + s[7]*br[5][k] + s[6]*br[4][k] + s[5]*br[3][k] + 
            s[4]*br[2][k] + s[3]*br[1][k] + s[2]*br[0][k] + s[1]*a0[k] + s[0]*statePrev[k+dim] ) + statePrev[k];
    } else {  // interpolate velocity
        s[0] = timestep * tau;
        s[1] =      s[0] * tau / 2.;
        s[2] = 2. * s[1] * tau / 3.;
        s[3] = 3. * s[2] * tau / 4.;
        s[4] = 4. * s[3] * tau / 5.;
        s[5] = 5. * s[4] * tau / 6.;
        s[6] = 6. * s[5] * tau / 7.;
        s[7] = 7. * s[6] * tau / 8.;
        return compsum[c] + s[7]*br[6][k] + s[6]*br[5][k] + s[5]*br[4][k] + s[4]*br[3][k] + 
            s[3]*br[2][k] + s[2]*br[1][k] + s[1]*br[0][k] + s[0]*a0[k] + statePrev[k+dim];
    }
}

/* ------ utility functions ------- */
// return the name for a given integrator type
std::string getIntegratorNameByType(CBasicOdeIntegrator::STEPPERKIND integratorType)
{
    switch(integratorType) {
    case CBasicOdeIntegrator::SK_LEAPFROG_NB:   return COdeIntegratorLeapfrog::myName();
    case CBasicOdeIntegrator::SK_DOP853:        return COdeIntegratorDOP853::myName();
    case CBasicOdeIntegrator::SK_IAS15:         return COdeIntegratorIAS15::myName();
#ifdef HAVE_ODEINT
    case CBasicOdeIntegrator::SK_ODEINT_CK5:    return COdeIntegratorOdeint<StepperCK5>::myName();
    case CBasicOdeIntegrator::SK_ODEINT_DP5:    return COdeIntegratorOdeint<StepperDP5>::myName();
    case CBasicOdeIntegrator::SK_ODEINT_RK4:    return COdeIntegratorOdeint<StepperRK4>::myName();
    case CBasicOdeIntegrator::SK_ODEINT_BS:     return COdeIntegratorOdeint<StepperBS>::myName();
    case CBasicOdeIntegrator::SK_ODEINT_SYMPL4: return COdeIntegratorOdeint<StepperSympl4>::myName();
#endif
    default: return "Default";
    }
};

// return integrator type associated with the text name
CBasicOdeIntegrator::STEPPERKIND getIntegratorTypeByName(std::string integratorName)
{
    if(integratorName==COdeIntegratorLeapfrog          ::myName()) return CBasicOdeIntegrator::SK_LEAPFROG_NB;
    if(integratorName==COdeIntegratorDOP853            ::myName()) return CBasicOdeIntegrator::SK_DOP853;
    if(integratorName==COdeIntegratorIAS15             ::myName()) return CBasicOdeIntegrator::SK_IAS15;
#ifdef HAVE_ODEINT
    if(integratorName==COdeIntegratorOdeint<StepperCK5>::myName()) return CBasicOdeIntegrator::SK_ODEINT_CK5;
    if(integratorName==COdeIntegratorOdeint<StepperDP5>::myName()) return CBasicOdeIntegrator::SK_ODEINT_DP5;
    if(integratorName==COdeIntegratorOdeint<StepperRK4>::myName()) return CBasicOdeIntegrator::SK_ODEINT_RK4;
    if(integratorName==COdeIntegratorOdeint<StepperBS> ::myName()) return CBasicOdeIntegrator::SK_ODEINT_BS;
    if(integratorName==COdeIntegratorOdeint<StepperSympl4>::myName()) return CBasicOdeIntegrator::SK_ODEINT_SYMPL4;
#endif
    return CBasicOdeIntegrator::SK_DEFAULT;
};

};  // namespace