/** \file    odeint.h
    \brief   ODE integration classes
    \author  EV
    \date    2008-2014

    Classes for orbit integration, derived from smile::CBasicOdeIntegrator, 
    are declared in this module.
    The integrators call smile::CPotential::DiffEq for obtaining accelerations, 
    adaptively adjust timestep to maintain desired accuracy, and call 
    runtime functions derived from smile::CBasicOrbitRuntimeFnc after each timestep.

    The COrbitIntegratorDOP853 contains code of the 8th order Runge-Kutta integrator from 
    Hairer, Norsett & Wanner, "Solving ordinary differential equations", 1987, Berlin:Springer.
    Based on the C version (by J.Colinge) of the original Fortran code by E.Hairer & G.Wanner.

    The COrbitIntegratorIAS15 is an adapted version of the 15th order integrator from Rebound:
    Rein & Spiegel, 2014, MNRAS.
*/

#pragma once
#include "common.h"
// disable MSVC warning about unreferenced formal parameter in odeint templates
#ifdef _WIN32
#pragma warning( push )
#pragma warning( disable:4100 )
#endif
#ifdef HAVE_ODEINT
#include <boost/numeric/odeint.hpp>
#endif

namespace smile{

/** 8th order Runge-Kutta integrator from Hairer,Norsett&Wanner.
    It is suitable for integrating orbits in a smooth potential (everything except the treecode potential). */
class COdeIntegratorDOP853: public CBasicOdeIntegrator
{
public:
    COdeIntegratorDOP853(COdeSystem* _odeSystem, double _accAbs, double _accRel):
      CBasicOdeIntegrator(_odeSystem), accAbs(_accAbs), accRel(_accRel) {};

    virtual void integrateToTime(double timeEnd);

    /// computes interpolated coordinate at time t within current timestep
    virtual double getInterpolatedSolution(unsigned int c, double t) const;

    static std::string myName() { return "DOP853"; };
private:
    double accAbs, accRel;    ///< relative and absolute accuracy for the integrator
    unsigned int nfcn,        ///< statistics: number of force evaluations
        nstep, naccpt, nrejct;///< statistics: number of timesteps (total, accepted, rejected)
    double   hout;            ///< current timestep
    /** temporary storage for dense output interpolation */
    OdeStateType rcont1, rcont2, rcont3, rcont4, rcont5, rcont6, rcont7, rcont8;
    OdeStateType stateCurr;   ///< current values of variables inside the integrator

    /// driver routine of the DOP853 integrator
    int dop853(
      double tstart,    ///< initial t-value
      OdeStateType &y,  ///< initial values for variables
      double tend,      ///< final t-value (tend-tstart may be positive or negative)
      double atoler,    ///< absolute error tolerance
      double rtoler     ///< relative error tolerance
    );

    /// shortcut for calling the function that provides the derivative of the ODE system 
    inline void fcn(unsigned int /*n*/, double t, const OdeStateType &vars, OdeStateType &derivs) const
    { odeSystem->odeFnc(t, vars, &derivs); }

    /// internal integration initialization routine
    double hinit (unsigned int n, double x, OdeStateType& y,
      double posneg, OdeStateType& f0, OdeStateType& f1, OdeStateType& yy1, 
      int iord, double hmax, double atoler, double rtoler);
    
};

/** 15-th order implicit Runge-Kutta scheme from Rein & Spiegel, 2014, MNRAS (adapted from Rebound).
    It has its own accuracy parameter, with typical values 10^-4..10^-3 providing relative accuracy 
    in the range 10^-15(almost machine precision)..10^-10. 
    It works only for "Standard Hamiltonian" systems (position+velocity variables). */
class COdeIntegratorIAS15: public CBasicOdeIntegrator
{
public:
    COdeIntegratorIAS15(COdeSystem* _odeSystem, double _acc):
      CBasicOdeIntegrator(_odeSystem), integrator_epsilon(_acc) {};

    virtual void integrateToTime(double timeEnd);

    virtual double getInterpolatedSolution(unsigned int c, double t) const;

    static std::string myName() { return "IAS15"; };

private:
    double integrator_epsilon;  ///< accuracy for the integrator
    unsigned int nfcn,          ///< statistics: number of force evaluations
        nstep, naccpt, nrejct;  ///< statistics: number of timesteps (total, accepted, rejected)
    OdeStateType stateCurr;     ///< ODE system state at the end of timestep
    OdeStateType statePrev;     ///< ODE system state at the beginning of timestep
    OdeStateType deriv;         ///< temp.buffer for computing rhs of equations of motion
    vectord at;                 ///< Temp.buffer for accelerations
    vectord a0;                 ///< Accelerations at the beginning of timestep
    vectord compsum;            ///< Temp.buffer for compensated summation
    vectord g[7],b[7],e[7];     ///< coefficients of approximation in the integrator
    vectord br[7], er[7];       ///< coefficients of approximation at the beginning of timestep
    double dt;                  ///< current timestep length
    double dt_last_success;     ///< Last accepted timestep (corresponding to br and er)
    /// Helper function for resetting the b and e coefficients
    void predict_next_step(double ratio, int N3, vectord _e[7], vectord _b[7]);
    /// Actual integration for one timestep
    int integrator_ias15_step();
};

#ifdef HAVE_ODEINT

/** A template class for integrators based on the boost::numeric::odeint library.
    The argument of template class is the class name of the stepper algorithm from odeint. */
template<class Stepper >
class COdeIntegratorOdeint: public CBasicOdeIntegrator
{
public:
    /// category of the integrator (could be fixed-timestep, adaptive-timestep, adaptive with dense output, etc)
    typedef typename boost::numeric::odeint::unwrap_reference< Stepper >::type::stepper_category stepper_category;

    COdeIntegratorOdeint(COdeSystem* _odeSystem, double _accAbs, double _accRel);
    virtual void integrateToTime(double timeEnd);
    virtual double getInterpolatedSolution(unsigned int c, double t) const;
    static std::string myName();
private:
    Stepper stepper;                         ///< the instance of stepper that does actual integration
    OdeStateType stateCurr, statePrev;       ///< ODE system state at the beginning and the end of timestep
    mutable OdeStateType stateIntermediate;  ///< temporary cached storage for interpolated trajectory
    mutable double timeIntermediate;         ///< temporary cache for interpolation is for this particular value of time
    double timeStep;                         ///< current timestep length
    bool isStdHamiltonian;                   ///< whether the ODE system is a "standard Hamiltonian" (positions, then velocities)
    /// driver routine for generic (fixed-timestep) steppers
    void integrateToTimeImpl(double timeEnd, boost::numeric::odeint::stepper_tag);
    /// driver routine for adaptive-timestep steppers
    void integrateToTimeImpl(double timeEnd, boost::numeric::odeint::controlled_stepper_tag);
    /// driver routine for adaptive-timestep steppers with dense output
    void integrateToTimeImpl(double timeEnd, boost::numeric::odeint::dense_output_stepper_tag);
    /// interpolate trajectory for a generic stepper
    double getInterpolatedSolutionImpl(unsigned int c, double t, boost::numeric::odeint::stepper_tag) const
    { return isStdHamiltonian ? getInterpolatedSolutionStdHamiltonian(c, t) : getInterpolatedSolutionGeneric(c, t); }
    /// interpolate trajectory for an adaptive-timestep stepper
    double getInterpolatedSolutionImpl(unsigned int c, double t, boost::numeric::odeint::controlled_stepper_tag) const
    { return isStdHamiltonian ? getInterpolatedSolutionStdHamiltonian(c, t) : getInterpolatedSolutionGeneric(c, t); }
    /// interpolate trajectory for a stepper that provides dense output
    double getInterpolatedSolutionImpl(unsigned int c, double t, boost::numeric::odeint::dense_output_stepper_tag) const;
    /// simple first-order interpolation of trajectory for an arbitrary stepper and ODE system of unknown structure
    double getInterpolatedSolutionGeneric(unsigned int c, double t) const;
    /// interpolation of trajectory for an ODE system that is a "standard Hamiltonian" -- 3rd order in position, 2nd order in velocity
    double getInterpolatedSolutionStdHamiltonian(unsigned int c, double t) const;
};

/** Helper class that provides timestep control for odeint steppers */
class my_error_checker
{
public:
    typedef double value_type;
    typedef boost::numeric::odeint::range_algebra algebra_type;
    typedef boost::numeric::odeint::default_operations operations_type;
    my_error_checker(
        const COdeSystem* _odeSystem,
        value_type eps_abs = static_cast< value_type >( 1.0e-6 ) ,
        value_type eps_rel = static_cast< value_type >( 1.0e-6 ) ,
        value_type a_x = static_cast< value_type >( 1 ) ,
        value_type a_dxdt = static_cast< value_type >( 1 ) )
    : odeSystem(_odeSystem), m_eps_abs(eps_abs) , m_eps_rel(eps_rel) , m_a_x(a_x) , m_a_dxdt(a_dxdt)
    { }

    template< class State , class Deriv , class Err , class Time >
    value_type error( const State &x_old , const Deriv &dxdt_old , Err &x_err , Time dt ) const
    {
        return error( algebra_type() , x_old , dxdt_old , x_err , dt );
    }

    template< class State , class Deriv , class Err , class Time >
    value_type error( algebra_type &algebra , const State &x_old , const Deriv &dxdt_old , Err &x_err , Time dt ) const
    {
        double eps_rel=m_eps_rel * odeSystem->getRelErrorToleranceFactor(x_old);
        algebra.for_each3( x_err , x_old , dxdt_old ,
            typename operations_type::template rel_error< value_type >( m_eps_abs , eps_rel , m_a_x , m_a_dxdt * boost::numeric::odeint::get_unit_value( dt ) ) );
        return algebra.norm_inf( x_err );
    }

private:
    const COdeSystem* odeSystem;
    value_type m_eps_abs;
    value_type m_eps_rel;
    value_type m_a_x;
    value_type m_a_dxdt;
};

/// Implementation of Bulirsch-Stoer method with dense output
typedef boost::numeric::odeint::bulirsch_stoer_dense_out< OdeStateType >  StepperBS;
/// Implementation of the Dormand-Prince 5th order Runge-Kutta method with adaptive timestep and dense output
typedef boost::numeric::odeint::dense_output_runge_kutta<
        boost::numeric::odeint::controlled_runge_kutta< 
        boost::numeric::odeint::runge_kutta_dopri5< OdeStateType >, my_error_checker > >  StepperDP5;
/// Implementation of the Cash-Karp 5th order Runge-Kutta method with adaptive timestep
typedef boost::numeric::odeint::controlled_runge_kutta< 
        boost::numeric::odeint::runge_kutta_cash_karp54< OdeStateType >, my_error_checker > StepperCK5;
/// Implementation of the standard fixed-timestep 4th order Runge-Kutta method
typedef boost::numeric::odeint::runge_kutta4< OdeStateType >  StepperRK4;
/// Implementation of the fixed-timestep 4th order symplectic Runge-Kutta method
typedef boost::numeric::odeint::symplectic_rkn_sb3a_mclachlan< OdeStateType >  StepperSympl4;
#endif

/** low-order leapfrog method provided for treecode, which has discontinuous potential/forces 
    and is poorly handled by Runge-Kutta integrator. It does not compute Lyapunov exponent either. */
class COdeIntegratorLeapfrog: public CBasicOdeIntegrator
{
public:
    COdeIntegratorLeapfrog(COdeSystem* _odeSystem, double _accTreecode, bool _symmetrizeTimestep):
      CBasicOdeIntegrator(_odeSystem), accTreecode(_accTreecode), symmetrizeTimestep(_symmetrizeTimestep) {};
    virtual void integrateToTime(double timeEnd);
    virtual double getInterpolatedSolution(unsigned int c, double t) const;
    static std::string myName() { return "Leapfrog"; };
private:
    double accTreecode;        ///< accuracy parameter (timestep is accTreecode*timeToCollision)
    bool symmetrizeTimestep;   ///< whether to use Hut et al.(1995) timestep symmetrization algorithm
    OdeStateType stateCurr, statePrev;  ///< ODE system state at the beginning and the end of timestep
};

///@{  Correspondence between ODE integrator names and types 
/// return the name for a given integrator type
std::string getIntegratorNameByType(CBasicOdeIntegrator::STEPPERKIND integratorType);
/// return integrator type associated with the text name
CBasicOdeIntegrator::STEPPERKIND getIntegratorTypeByName(std::string integratorName);
///@}
};  // namespace
#ifdef _WIN32
#pragma warning( pop )
#endif