#include "orbitint.h"
#include <cassert>
#include <gsl/gsl_integration.h>
#include "potential.h"
#include "utils.h"
#include "odeint.h"

namespace smile{

/** default parameters for orbit integration and classification **/
const CConfigOrbit defaultConfigOrbit = {
    3,     // N_dim
    0.0,   // Omega
    CBasicOdeIntegrator::SK_DOP853,  // integratorType
    1e-9,  // accuracyRel
    1e-9,  // accuracyAbs
    1e-4,  // accuracyIAS15
    0.4,   // accuracyTreecode
    false, // treecodeSymmetrizeTimestep
    0.005, // accuracyFixedTimestep
    0.0,   // relaxRate
};

/*-----------------   Below follows COrbit class  -------------*/

/* constructor for an empty orbit */
template<typename NumT> COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const CPosVelPoint<NumT> &_initCond, const vectorRuntimeFncCreators* creators,
    double _timeEnd, double _timeMax) :
  potential(_potential),
  configOrbit(_configOrbit),
  initCond(_initCond), endCond(initCond)
{
    //timeUnit = -2;  // will be computed when necessary
    timeEnd=_timeEnd; 
    timeMax=_timeMax;
    if(timeMax<timeEnd) timeMax=timeEnd;
    state=OS_INITIALIZED;
    odeInt=NULL;
    if(creators!=NULL)  // construct timestep function objects using creator classes
        for(vectorRuntimeFncCreators::const_iterator iter=creators->begin(); iter!=creators->end(); iter++)
        {
            CBasicOrbitRuntimeFnc* fnc=(*iter)->createRuntimeFnc(this);
            if(fnc!=NULL) runtimeFncs.push_back(fnc);
        }
}
// explicit instantiations of constructor for two variants of template argument
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const CPosVelPoint<float> &_initCond, const vectorRuntimeFncCreators* creators,
    double _timeEnd, double _timeMax);
template COrbit::COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
    const CPosVelPoint<double> &_initCond, const vectorRuntimeFncCreators* creators,
    double _timeEnd, double _timeMax);

COrbit::~COrbit()
{
    // if there were any timestep functions attached, destroy 'em all!!
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
        delete (*iter);
}

std::string COrbit::toString(bool runtimeFncInfo) const 
{
    std::string strInfo;
    if(runtimeFncInfo)  
        for(size_t rf=0; rf<getInfoNum(); rf++)
        {
            const CBasicInformation* info=getNewInfo(rf);
            if(info!=NULL) {
                strInfo += info->toString();
                delete info;
            }
        }
    strInfo+="nStep="+convertToString(numSteps)+", nForceEval="+convertToString(numFncEval)+"\n";
    return strInfo;
}

double COrbit::getIntTime() const
{ 
    return state==OS_INITIALIZED ? 0 : odeInt!=NULL ? odeInt->getTime() : timeEnd; 
}

void COrbit::run()
{
    if(odeInt!=NULL) 
        return;   // can't run orbit integrator twice
    CBasicOdeIntegrator::STEPPERKIND stepperKind=configOrbit->integratorType;
    if(stepperKind==CBasicOdeIntegrator::SK_DEFAULT) 
        stepperKind=CBasicOdeIntegrator::SK_DOP853;
    needToTerminate=false;
    numSteps=numFncEval=0;
    Etoler=fabs(totalEnergy(potential, initCond)*10);  // used for energy conservation accuracy in determining timestep length
    if(Etoler==0) Etoler=(pow_2(initCond.Vel[0])+pow_2(initCond.Vel[1])+pow_2(initCond.Vel[2]))*5;  // somewhat arbitrary setting for non-finite potentials (which could have Phi>0)
    try{
        switch(stepperKind)
        {
        case CBasicOdeIntegrator::SK_DOP853:
            odeInt = new COdeIntegratorDOP853(this, 
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_IAS15:
            odeInt = new COdeIntegratorIAS15(this, 
                configOrbit->accuracyIAS15);
            break;
#ifdef HAVE_ODEINT
        case CBasicOdeIntegrator::SK_ODEINT_DP5:
            odeInt = new COdeIntegratorOdeint<StepperDP5>(this,
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_CK5:
            odeInt = new COdeIntegratorOdeint<StepperCK5>(this,
                configOrbit->accuracyAbs, configOrbit->accuracyRel);
            break;
        /*case CBasicOdeIntegrator::SK_ODEINT_RK4:
            odeInt = new COdeIntegratorOdeint<StepperRK4>(this,
                timeUnit*configOrbit->accuracyFixedTimestep, 0);
            break;
        case CBasicOdeIntegrator::SK_ODEINT_SYMPL4:
            odeInt = new COdeIntegratorOdeint<StepperSympl4>(this,
                timeUnit*configOrbit->accuracyFixedTimestep, 0);
            break;*/
#endif
        default:
            my_error(FUNCNAME, "Unknown orbit integrator type requested!");
            return;
        }
    }
    catch(std::exception& e) {
        my_error(FUNCNAME, std::string("An exception occured during creation of ODE integrator: ")+e.what());
        return;
    }
    state=OS_RUNNING;
    try{
        odeInt->integrateToTime(timeEnd);
    }
    catch(std::exception& e) {
        my_error(FUNCNAME, std::string("An exception occured during orbit integration: ")+e.what());
        return;
    }
    try{
        timeEnd=odeInt->getTime();
        const size_t N_dim=configOrbit->N_dim;
        const size_t numVars2 = N_dim;
        endCond=CPosVelPoint<double>(
            odeInt->getInterpolatedSolution(0, timeEnd),
            odeInt->getInterpolatedSolution(1, timeEnd),
            N_dim==N_DIM ? odeInt->getInterpolatedSolution(2, timeEnd) : 0,
            odeInt->getInterpolatedSolution(numVars2+0, timeEnd),
            odeInt->getInterpolatedSolution(numVars2+1, timeEnd),
            N_dim==N_DIM ? odeInt->getInterpolatedSolution(numVars2+2, timeEnd) : 0);
        for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
            (*iter)->Finish();
    }
    catch(std::exception& e) {
        my_error(FUNCNAME, std::string("An exception occured during finalizing an orbit: ")+e.what());
        return;
    }
    state=OS_DONE;
    delete odeInt;
    odeInt=NULL;
}

void COrbit::initState(OdeStateType* state)
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim;
    state->assign(2*numVars2, 0);
    for(unsigned int d=0; d<N_dim; d++)
    {
        (*state)[d]         = initCond.Pos[d];
        (*state)[d+numVars2]= initCond.Vel[d];
    }
}

void COrbit::odeFnc(double t, const OdeStateType& x, OdeStateType* dx)
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim;
    double force[N_DIM+1] = {0,0,0,0};
    double coord[N_DIM] = {x[0], x[1], N_dim==N_DIM?x[2]:0};
    potential->Force(coord, t, force, NULL);
    numFncEval++;
    for(size_t i=0; i<numVars2; i++)
        dx->at(i)=x[i+numVars2];  // d x_i/dt = v_i
    dx->at(numVars2+0)=force[0];
    dx->at(numVars2+1)=force[1];
    if(N_dim==N_DIM) dx->at(numVars2+2)=force[2];
}

double COrbit::getRelErrorToleranceFactor(const OdeStateType& state) const
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim;
    double vel2=0;
    for(size_t i=numVars2; i<numVars2+N_dim; i++)
        vel2+=pow_2(state[i]);
    return std::min<double>(1.0, Etoler/vel2);
}

COrbit::STEPRESULT COrbit::processStep(double timePrev, double timeCurr, OdeStateType* state)
{
    const size_t N_dim=configOrbit->N_dim;
    const size_t numVars2 = N_dim;
    numSteps++;
    CPosVelPoint<double> pt(
        state->at(0), state->at(1), N_dim==N_DIM ? state->at(2) : 0,
        state->at(numVars2+0), state->at(numVars2+1), N_dim==N_DIM ? state->at(numVars2+2) : 0);
    bool needToContinue=false;
    bool needToReinit=false;
    for(vectorRuntimeFncs::iterator iter=runtimeFncs.begin(); iter!=runtimeFncs.end(); iter++)
    {   
        // note that if runtime funtions modify pt, their order is important! 
        // in fact, it doesn't make sense to have more than one function modifying the current state,
        // and this only one should be the last in sequence.
        CBasicOrbitRuntimeFnc::STEPRESULT sr = (*iter)->Timestep(timePrev, timeCurr, &pt);
        if(sr==CBasicOrbitRuntimeFnc::SR_CONTINUE) needToContinue=true;
        if(sr==CBasicOrbitRuntimeFnc::SR_REINIT) needToReinit=true;
    }
    if(needToReinit) 
    {   // store modified phase-space variables into ODE state vector
        state->at(0)=pt.Pos[0];
        state->at(1)=pt.Pos[1];
        state->at(numVars2)=pt.Vel[0];
        state->at(numVars2+1)=pt.Vel[1];
        if(N_dim==N_DIM) {
            state->at(2)=pt.Pos[2];
            state->at(numVars2+2)=pt.Vel[2];
        }
    }
    // check for escaping orbits (very crude!!)
    if(pow_2(pt.Pos[0])+pow_2(pt.Pos[1])+pow_2(pt.Pos[2]) > 
       (pow_2(initCond.Pos[0])+pow_2(initCond.Pos[1])+pow_2(initCond.Pos[2]))*1e6 )  // could be escaping, check total energy
        if(totalEnergy(potential, pt)>0)
            needToTerminate=true;
    if(needToTerminate || (timeCurr>=timeEnd && !needToContinue) || timeCurr>=timeMax)
        return SR_TERMINATE;
    return needToReinit ? SR_REINIT : SR_CONTINUE;
}

double COrbit::getPos(unsigned int c, double t) const 
{
    if(odeInt==NULL || c>=configOrbit->N_dim) return 0;
    return odeInt->getInterpolatedSolution(c, t);
}

double COrbit::getVel(unsigned int c, double t) const 
{
    if(odeInt==NULL || c>=configOrbit->N_dim) return 0;
    size_t numVars2 = configOrbit->N_dim;
    return odeInt->getInterpolatedSolution(c+numVars2, t);
}

} // namespace