/** \file    orbitint.h
    \brief   Orbit integration class
    \author  EV
    \date    2009-2013

    This file defines smile::COrbit class which serves as an interface between orbit integrators
    (implemented in odeint.h) and orbit analysis methods (implemented in orbitan.h).
*/

#pragma once
#include "common.h"

namespace smile{

/** general parameters for orbit integration and classification */
struct CConfigOrbit
{
    // integration accuracy and parameters
    unsigned int N_dim;              ///< spatial dimensions of current problem (2d or 3d)
    double Omega;                    ///< pattern speed of rotating potential
    CBasicOdeIntegrator::STEPPERKIND integratorType;  ///< type of ODE integrator
    double accuracyRel;              ///< relative accuracy parameter for variable-timestep integrators (except IAS15)
    double accuracyAbs;              ///< absolute accuracy parameter for variable-timestep integrators (except IAS15)
    double accuracyIAS15;            ///< accuracy parameter for the IAS15 integrator
    double accuracyTreecode;         ///< factor in leap-frog timestep selection for the treecode potential
    bool treecodeSymmetrizeTimestep; ///< whether to use time-symmetrized leapfrog (Hut, Makino & McMillan 1995) - doubles the number of evaluations but better conserves energy
    double accuracyFixedTimestep;    ///< for fixed-timestep integrators, this is the timestep in units of orbital time
    double relaxRate;                ///< magnitude of velocity perturbation term (logLambda/N, with N being the number of discrete stars in galaxy)
};
extern const CConfigOrbit defaultConfigOrbit;

/** Class for performing a single orbit integration.
    COrbit is an interface class that stores the orbit initial data, calls one of the available 
    orbit integrators (derived from CBasicOdeIntegrator), and stores and provides access 
    to orbit runtime functions and their information objects 
    (e.g. trajectory and spectrum analysis, Poincare section, etc.).
    Its interface with ODE integrators is derived from the COdeSystem class, 
    which provides the way to initialize and, if necessary, modify variables 
    in the course of integration, and provide the r.h.s. of the equations of motion.
    The exact "meaning" of integration variables is defined by this class: 
    it provides methods to access position and velocity as a function of time 
    during integration, which are, in turn, provided by ODE integrator and transformed 
    according to the internal representation.
*/
class COrbit: public COdeSystem
{
public:
    /** three stages of orbit integration */
enum ORBITSTATE{
    OS_INITIALIZED,
    OS_RUNNING,
    OS_DONE
};
    /** construct from given initial conditions and optionally list of runtime function creators */
    template<typename NumT> COrbit(const CConfigOrbit* _configOrbit, const CPotential* _potential, 
        const CPosVelPoint<NumT> &_initCond, const vectorRuntimeFncCreators* creators,
        double _timeEnd, double _timeMax=0);

    /** destructor deletes all associated runtime functions */
    virtual ~COrbit();
    
    /** initializes the working variables of ODE integrator 
        (using InitData as input, assigning values to the output vector "state") */
    virtual void initState(OdeStateType* state);

    /** provides the time derivative of ODE integration variables (equations of motion) */
    virtual void odeFnc(double t, const OdeStateType& x, OdeStateType* dx);

    /** defines whether the ODE variables are canonical and represent a separable hamiltonian system.
        Presently, the internal representation of integration variables is always 
        in canonical coordinates -- first come positions, then velocities; 
        this allows for some simplifications in ODE integrator */
    virtual bool isStdHamiltonian() { return true; }

    /** allows to tighten error tolerance for ODE integrator near the black hole */
    virtual double getRelErrorToleranceFactor(const OdeStateType& state) const;
    
    /** do some data recording and processing after each integration timestep.
        Calls each of the runtime functions at each timestep and performs some other checks. 
        \param[in,out] state is the vector of integration variables at the end of timestep;
        if necessary, it may be transformed (e.g.rescaled).
        \return SR_TERMINATE if the requested integration time is achieved and 
        all runtime functions can stop right now, or if maximum integration time is exceeded;
        SR_CONTINUE if need to continue integration;
        SR_REINIT if the internal system state has been changed (need to redo ODE integrator's 
        internal initialization). */
    virtual STEPRESULT processStep(double timePrev, double timeCurr, OdeStateType* state);

    /** integrate orbit up to the timeEnd provided in the constructor; 
        if any of runtime functions returns false then continue until all of them return true 
        or timeMax is reached; timeMax=0 means timeMax=timeEnd */
    void run();
    
    /** called if it is necessary to stop integration immediately; raises internal termination flag */
    void halt() { needToTerminate=true; }

    /** returns c-th coordinate interpolated at time t within currect timestep interval */
    double getPos(unsigned int c, double t) const;

    /** returns c-th velocity interpolated at time t within currect timestep interval */
    double getVel(unsigned int c, double t) const;

    /** returns Ln(w) - logarithm of deviation vector at time t within current timestep */
//    double getLogDevVec(double t) const;

    // get functions
    /** return current state (unstarted/running/done) */
    ORBITSTATE getState() const { return state; };

    /** if the orbit is running, return integration time to which it has already progressed, 
        otherwise return the total integration time if it is finished, or 0 if it hasn't started yet */
    double getIntTime() const;

    /** return initial conditions */
    const CPosVelPoint<double>& getInitCond() const { return initCond; };

    /** return end point (after integration is finished) */
    const CPosVelPoint<double>& getEndCond() const { return endCond; };

    /** return pointer to potential used in integration */
    const CPotential* getPotential() const { return potential; };

    /** return the period of long-axis orbit with the same energy as this one */
//    double getTimeUnit() const;

    /** return pointer to configuration parameters (may be needed by runtime functions) */
    const CConfigOrbit* getConfig() const { return configOrbit; };

    /** return pointer to the runtime function at the given index */
    const CBasicOrbitRuntimeFnc* getRuntimeFnc(size_t index) const { return runtimeFncs.at(index); };

    /** convenience function: create and return result information object for the given runtime function index;
        warning: the returned value may be NULL */
    const CBasicInformation * getNewInfo(size_t index) const 
    { return index<runtimeFncs.size() ? runtimeFncs.at(index)->getData() : NULL; };

    /** return the total number of runtime functions */
    size_t getInfoNum() const { return runtimeFncs.size(); };

    /** return some textual information about the orbit, optionally with contributions 
        from all runtime function information objects */
    std::string toString(bool withRuntimeFncInfo=true) const;

private:
    const CPotential* potential;         ///< pointer to the potential used in integration
    const CConfigOrbit* configOrbit;     ///< pointer to the configuration structure keeping various integration and analysis parameters
    const CPosVelPoint<double> initCond; ///< initial conditions for the orbit
    CPosVelPoint<double> endCond;        ///< end point of the orbit
//    mutable double timeUnit;             ///< time unit = period of radial orbit with the same energy as this one (computed at most once if needed, then cached)
    volatile ORBITSTATE state;           ///< current state of affairs
    volatile bool needToTerminate;       ///< a flag raised internally or externally if the integration should be finished immediately
    CBasicOdeIntegrator* odeInt;         ///< pointer to orbit integrator if it is running, otherwise NULL
    vectorRuntimeFncs runtimeFncs;       ///< holds all attached 'timestep functions' (if any). They are destroyed along with COrbit!
    double timeEnd;                      ///< integration time - 0 if it hasn't started yet, or the actual duration when finished
    double timeMax;                      ///< maximum allowed integration time (0 means equal to timeEnd)
    size_t numFncEval;                   ///< count the number of force evaluations
    size_t numSteps;                     ///< count the number of timesteps taken by ODE integrator
    double Etoler;                       ///< factor determining the accuracy of energy conservation in timestep selection
};

/// Utility function to find approximate L^2 of a circular orbit with given energy
//double findLcirc2(const CPotential* potential, const CPosVelPoint<double> &initCond, double Omega=0);

/// Utility function to get the period of long-axis orbit in the given potential with the energy corresponding to initial conditions 
//template<typename NumT>
//double computeTimeUnit(const CPotential* potential, const CPosVelPoint<NumT> &initCond, double Omega=0);

}  // namespace