/** \file    potential.h
    \brief   Classes for computing potential, forces and density for various mass models
    \author  EV
    \date    2009-2013

    This file defines several classes of two kind:
    - density profiles (descendants of smile::CDensity class), which provide only density 
      as a function of coordinates, the corresponding potential needs not be known;
    - potential solvers (derived from smile::CPotential class, which inherits CDensity), 
      which in addition compute potential and forces, and provide several utility functions 
      like finding an intersection with equipotential surface.

    There are several flexible general-purpose potential representations, which can take an arbitrary 
    density profile as an input; moreover there are at least two flexible density parametrizations suitable 
    to represent almost anything one may need. 
    If, nevertheless, you want to add a new density model, the necessary steps are summarized below:
    - Add new identifier to CDensity::POTENTIALTYPE enum;
    - Implement a new class derived from CDensity, which provides the density profile in the virtual function 
      Rho(x,y,z), and, optionally, returns the asymptotic power-law slope at r->0 in getGamma();
      also it should return correct PotentialType and name;
    - Add item to DensityNames list in initPotentialAndSymmetryNameMap(); 
    - Add corresponding line to CSmileCore::initPotential() in core.cpp under the section PT_BSE, PT_SPLINE;
    - If the density model has additional parameters besides axis ratio, one needs also to modify 
      CSmileCore::loadSettings()/saveSettings() in core.cpp and potentialChanged() in gui.cpp to show/hide 
      and interactively modify these parameters, and probably additional visual elements in GUI (smile.ui)
**/
#pragma once
#include "common.h"
#include <gsl/gsl_spline.h>
#include <map>

namespace smile{

/// \name Density models
///@{

/** Base class defining density model (without corresponding potential). 
    May be used for initializing density for general-purpose potential approximations (BSE and Spline). 
**/
class CDensity
{
public:
    /// list of all existing types of density or density/potential models, each of them implemented in its own class
enum POTENTIALTYPE   
{
    PT_UNKNOWN,     ///< means that potential will be initialized by default, currently as PT_LOG
    PT_COMPOSITE,   ///< a superposition of multiple potential instances:  CPotentialComposite
    PT_NB,          ///< density initialized from a discrete point mass set
    PT_COEFS,       ///< not an actual density model, but a way to load pre-computed coefficients of potential expansion
    PT_BSE,         ///< basis-set expansion for infinite systems:  CPotentialBSE
    PT_BSECOMPACT,  ///< basis-set expansion for systems with non-singular density and finite extent:  CPotentialBSECompact
    PT_SPLINE,      ///< spline spherical-harmonic expansion:  CPotentialSpline
    PT_LOG,         ///< logaritmic potential:  CPotentialLog
    PT_HARMONIC,    ///< simple harmonic oscillator:  CPotentialHarmonic
    PT_SPHERICAL,   ///< arbitrary spherical mass model:  CPotentialSpherical
    PT_DEHNEN,      ///< Dehnen(1993) density model:  CPotentialDehnen
    PT_FERRERS,     ///< Ferrers finite-extent profile:  CDensityFerrers
    PT_PLUMMER,     ///< Plummer model:  CDensityPlummer
    PT_ISOCHRONE,   ///< isochrone model:  CDensityIsochrone
    PT_PERFECTELLIPSOID,  ///< Kuzmin/de Zeeuw integrable potential:  CDensityPerfectEllipsoid
    PT_NFW,         ///< Navarro-Frenk-White profile:  CDensityNFW
    PT_SERSIC,      ///< Sersic density profile:  CDensitySersic
    PT_ELLIPSOIDAL, ///< a generalization of spherical mass profile with arbitrary axis ratios:  CDensityEllipsoidal
    PT_MGE          ///< Multi-Gaussian expansion:  CDensityMGE
};
    /// Type of symmetry. 
    /// used mainly in spherical-harmonic expansion potentials to compute and use only 
    /// a subset of coefficients based on the symmetry properties of the underlying potential)
enum SYMMETRYTYPE{     
    ST_NONE = 0,       ///< all coefficients should be considered
    ST_REFLECTION = 1, ///< reflection about origin (change of sign of all coordinates simultaneously) - terms with odd l are zero
    ST_PLANESYM = 2,   ///< reflection about principal planes (change of sign of any coordinate) - no terms with sin(m phi) or odd m
    ST_ZROTSYM = 4,    ///< axisymmetric model (all terms with m>0 are zero)
    ST_SPHSYM = 8,     ///< spherical symmetry (all terms with l>0 are zero)
    ST_TRIAXIAL = ST_REFLECTION | ST_PLANESYM,    ///< triaxial symmetry (only even l,m terms are used)
    ST_AXISYMMETRIC = ST_TRIAXIAL | ST_ZROTSYM,   ///< axial symmetry (only m=0 terms are used for even l)
    ST_SPHERICAL = ST_AXISYMMETRIC | ST_SPHSYM    ///< spherical symmetry (only l=0 terms)
};
    CDensity() {};
    virtual ~CDensity() {};
    virtual POTENTIALTYPE PotentialType() const=0;  ///< enumerable potential type
    virtual std::string PotentialName() const=0;    ///< string representation of potential type
    virtual SYMMETRYTYPE symmetry() const=0;        ///< returns symmetry type of this potential
    virtual double Rho(double X, double Y, double Z, double t=0) const=0;    ///< returns density at given coordinates, this should obviously be overriden in derivative classes
    virtual double Mass(const double r) const;      ///< returns mass inside given radius (approximately! not necessary to integrate density over sphere, just a rough estimate used e.g. in choosing radial nodes of Schwarzschild grid)
    virtual double totalMass() const;               ///< returns estimated M(r=infinity) or -1 if mass is infinite
    double getRadiusByMass(const double m) const;   ///< solves for Mass(r)=m
    void getRadiiByMass(const vectord &masses, vectord *radii) const;  ///< solves for Mass(r)=m for an array of sorted values of m (more efficient than doing it one-by-one)
    bool checkMassMonotonic() const;                ///< safety measure: check (roughly) that mass is increasing with radius
    bool checkDensityNonzero() const;               ///< another safety measure: check that density doesn't drop to zero along any of three axes (important to assess spherical-harmonic approximation quality)
    virtual double getGamma() const;                ///< returns inner density slope estimate (only used in BSE potential expansion for the automatic selection of shape parameter Alpha)
private:
    CDensity& operator=(const CDensity& );          ///< assignment forbidden, use clone() for CPotential
};

/** Plummer density profile, triaxial **/
class CDensityPlummer: public CDensity
{
public:
    CDensityPlummer(double _mass, double _scalerad, double _q, double _p):
      CDensity(), mass(_mass), scalerad(_scalerad), q(_q), p(_p) {};
    virtual POTENTIALTYPE PotentialType() const { return PT_PLUMMER; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Plummer"; };
    virtual SYMMETRYTYPE symmetry() const    // returns symmetry type of this potential based on p, q being unity or not
    { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
//    virtual SYMMETRYTYPE symmetry() const { return ST_NONE; };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 0; };
 //   virtual double totalMass() const { return mass; }
private:
    const double mass;          ///< total mass of the model
    const double scalerad;      ///< scale radius
    const double q, p;          ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** Isochrone density profile, triaxial **/
class CDensityIsochrone: public CDensity
{
public:
    CDensityIsochrone(double _mass, double _scalerad, double _q, double _p):
      CDensity(), mass(_mass), scalerad(_scalerad), q(_q), p(_p) {};
    virtual POTENTIALTYPE PotentialType() const { return PT_ISOCHRONE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Isochrone"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 0; };
 //   virtual double totalMass() const { return mass; }
private:
    const double mass;          ///< total mass of the model
    const double scalerad;      ///< scale radius
    const double q, p;          ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** Triaxial Perfect Ellipsoid density profile **/
class CDensityPerfectEllipsoid: public CDensity
{
public:
    CDensityPerfectEllipsoid(double _mass, double _scalerad, double _q, double _p): 
      CDensity(), mass(_mass), scalerad(_scalerad), q(_q), p(_p) {};
    virtual POTENTIALTYPE PotentialType() const { return PT_PERFECTELLIPSOID; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Perfect Ellipsoid"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 0; };
 //   virtual double totalMass() const { return mass; }
private:
    const double mass;          ///< total mass of the model
    const double scalerad;      ///< scale radius
    const double q, p;          ///< axis ratio (y/x and z/x) of equidensity surfaces
};

/** modified Navarro-Frenk-White density profile, triaxial.
    The density is given by rho ~ 1/(r*(1+r)^2*(1+r/r_cut)), i.e. modified to have a steeper 
    r^{-4} outer slope beyond the cutoff radius, to make the total mass finite. 
    The cutoff radius is computed from the condition that the total mass is the same as for 
    a sharply cut NFW profile with a given concentration parameter, but the gentle transition 
    to the outer slope instead of a sharp cut makes it more suitable to use in the potential 
    expansions. **/
class CDensityNFW: public CDensity
{
public:
    CDensityNFW(double _mass, double _scalerad, double _q, double _p, double c);  // c == concentration parameter
    virtual POTENTIALTYPE PotentialType() const { return PT_NFW; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "NFW"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double getGamma() const { return 1.0; };
 //   virtual double totalMass() const { return mass; }
private:
    const double mass;          ///< total mass of the model
    const double scalerad;      ///< scale radius
    const double q, p; ///< axis ratio (y/x and z/x) of equidensity surfaces
    double rcutoff;    ///< cutoff radius
    double norm;       ///< normalization factor for density
};

/** Sersic profile, triaxial */
class CDensitySersic: public CDensity
{
public:
    CDensitySersic(double _mass, double _Reff, double _q, double _p, double _n);
    virtual ~CDensitySersic();
    virtual POTENTIALTYPE PotentialType() const { return PT_SERSIC; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Sersic"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    //virtual double getGamma() const { return 1.0; };
private:
    const double mass; ///< total mass of the model
    const double Reff; ///< scale radius
    const double q, p; ///< axis ratio (y/x and z/x) of equidensity surfaces
    const double n;    ///< Sersic index, must be >0
    double b;          ///< scaling parameter for radius
    //double rcutoff;    ///< cutoff radius
    //double norm;       ///< normalization factor for density
    gsl_spline* dens;  ///< spline interpolated density (log-log scaling)
};

/** Ellipsoidal model.
    Arbitrary spline-interpolated radial density profile with radially varying axis ratios 
    which are also spline-interpolated. **/
class CDensityEllipsoidal: public CDensity
{
public:
    CDensityEllipsoidal(const vectord &_radii, const vectord &_inmass, const vectord &_axesradii, const vectord &_qvals, const vectord &_pvals);
    virtual ~CDensityEllipsoidal();
    virtual POTENTIALTYPE PotentialType() const { return PT_ELLIPSOIDAL; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Ellipsoidal"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
private:
    double q, p;               ///< axis ratio (y/x and z/x) of equidensity surfaces; only make sense if they are constant with radius
    gsl_spline* dens;          ///< spline interpolated density (log-log scaling)
    gsl_spline* qvals, *pvals; ///< spline interpolated axis ratios
};

/** Multi-Gaussian expansion of a triaxial density profile **/
class CDensityMGE: public CDensity
{
public:
    /// defines one component of the multi-gaussian expansion
struct CGaussian {   
    double width;    ///< "sigma" - width of the gaussian in x-direction
    double mass;     ///< total mass in this component
    double q, p;     ///< axes ratio (widths of the gaussian in y and z directions w.r.t x)
    CGaussian(double _width, double _mass, double _q, double _p) : width(_width), mass(_mass), q(_q), p(_p) {};
};
typedef std::vector<CGaussian> vectorg;  ///< an array of gaussian components

    CDensityMGE(const vectorg &_data);
    virtual ~CDensityMGE() {};
    virtual POTENTIALTYPE PotentialType() const { return PT_MGE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "MGE"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
private:
    double q, p;     ///< axis ratio (y/x and z/x) of equidensity surfaces; only make sense if they are constant with radius
    vectorg data;    ///< array of gaussians representing the density distribution
};

///@}
/// \name Potential models
///@{

/** base class for potential-density model including potential derivatives (=forces)  **/
class CPotential: public CDensity
{
public:
    CPotential(): 
      CDensity() {};
    
    /** Return a pointer to a copy of this instance of potential.
        A standard copy constructor or assignment is disabled because of 
        different amount of data needed to be copied in different derived classes). */
    virtual CPotential* clone() const=0;
    
    /** Return potential at a given spatial point (possibly a time-varying one). */
    virtual double Phi(double X, double Y, double Z, double t=0) const=0;

    /** Compute forces and, optionally, force derivatives at a given point.
        \param[in]  xyz - coordinates of the point to compute forces (array of 3 numbers)
        \param[in]  t - time to compute forces (matters only if potential is time-dependent)
        \param[in,out] force - computed values of -d Phi/d x_i; output array must exist and contain N_DIM values (at least).
        The computed force (and its derivative) is added to the one already stored in the array 
        (thus it needs to be initialized before calling this function, most commonly by zeroes).
        \param[in,out] forceDeriv - if not NULL, then also compute the second derivatives of potential
        (which is a symmetric matrix of size N_DIM^2): first 3 values contain -d^2 Phi/d x_i^2, 
        second three contain mixed derivatives -d^2 Phi/d x_1 d x_2, -d^2 Phi/d x_2 d x_3, -d^2 Phi/d x_3 d x_1.
        */
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const=0;
};

/** A composite potential which consists of several other potential types, 
    plus a possible black hole and a rotating reference frame */
class CPotentialComposite: public CPotential
{
public:
    CPotentialComposite(const std::vector<const CPotential*> &_components, double _Mbh=0);
    virtual ~CPotentialComposite();
    virtual CPotential* clone() const { return new CPotentialComposite(components, Mbh); };
    virtual POTENTIALTYPE PotentialType() const { return PT_COMPOSITE; };
    virtual std::string PotentialName() const  { return myName(); };
    static std::string myName() { return "Composite"; };
    virtual SYMMETRYTYPE symmetry() const;
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
    size_t size() const { return components.size(); }
    const CPotential* component(size_t index) const { return index<components.size()?components[index]:NULL; }
    virtual double Mass(const double r) const;
private:
    std::vector<const CPotential*> components;
    const double Mbh;     ///< mass of central black hole(s)
};

/** Logarithmic potential, with possible core **/
class CPotentialLog: public CPotential
{
public:
    CPotentialLog(double _q, double _p, double _Rc): 
      CPotential(), q(_q), p(_p), Rc(_Rc) {};
    virtual CPotential* clone() const { return new CPotentialLog(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_LOG; };
    virtual std::string PotentialName() const  { return myName(); };
    static std::string myName() { return "Logarithmic"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
private:
    const double q, p;    ///< axis ratio (y/x and z/x) of equipotential surfaces
    const double Rc;      ///< core radius of log potential
};

/** Potential of a constant-density (harmonic) core with different frequencies of oscillations in three directions **/
class CPotentialHarmonic: public CPotential 
{
public:
    CPotentialHarmonic(double _q, double _p): 
      CPotential(), q(_q), p(_p) {};
    virtual CPotential* clone() const { return new CPotentialHarmonic(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_HARMONIC; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Harmonic"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
private:
    const double q, p;    ///< axis ratio (y/x and z/x) of equidensity surfaces
};

class CMassModel;  // forward declaration
/** Spherically-symmetric arbitrary density profile potential provided by CMassModel **/
class CPotentialSpherical: public CPotential
{
public:
    /// init from existing mass model (the instance of CMassModel is cloned)
    CPotentialSpherical(const CMassModel& _model);
    /// init potential from a given r,M(r) profile by creating a new mass model
    CPotentialSpherical(const vectord& radii, const vectord& mass);
    /// init potential from N point masses by creating a new mass model; 
    /// may also provide desired grid radii (otherwise assigned automatically)
    template<typename NumT>
    CPotentialSpherical(const CPointMassSet<NumT> &points, double smoothfactor=0, const vectord *_gridradii=NULL);
    ~CPotentialSpherical();
    virtual CPotential* clone() const;
    virtual POTENTIALTYPE PotentialType() const { return PT_SPHERICAL; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Spherical"; };
    virtual SYMMETRYTYPE symmetry() const { return ST_SPHERICAL; };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
    virtual double Mass(const double r) const;
    const CMassModel* getModel() const { return model; }
private:
    const CMassModel* model;
};
    
/** Ferrers density profile (with finite size R and density (1-(r/R)^2)^2).
    The potential is calculated using expressions from Pfenniger(1984) with elliptic integrals, 
    under assumption that p<q<1 strictly (will not work if any of two axes are equal). **/
class CPotentialFerrers: public CPotential
{
public:
    CPotentialFerrers(double _mass, double _R, double _q, double _p);
    virtual CPotential* clone() const { return new CPotentialFerrers(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_FERRERS; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Ferrers"; };
    virtual SYMMETRYTYPE symmetry() const { return (a==b ? (a==c ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
    virtual double getGamma() const { return 0; };
    //   virtual double totalMass() const { return mass; }
private:
    /** compute the array of 20 coefficients W_{ijk} listed in Pfenniger(1984), in order of appearance in that paper; 
        \param[in] lambda is zero inside the model and >0 outside;
        \param[out] W is the array of 20 coefs  */
    void computeW(double lambda, double W[20]) const;
    const double a, b, c;       ///< principal axis of ellipsoidal density 
    const double mass, rho0;    ///< total mass and central density of the model
    double W0[20];              ///< pre-computed coefficients for lambda=0
};
    
/** Dehnen(1993) double power-law model **/
class CPotentialDehnen: public CPotential
{
public:
    CPotentialDehnen(double _mass, double _scalerad, double _q, double _p, double _Gamma);
    virtual CPotential* clone() const { return new CPotentialDehnen(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_DEHNEN; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Dehnen"; };
    virtual SYMMETRYTYPE symmetry() const { return (q==1 ? (p==1 ? ST_SPHERICAL : ST_AXISYMMETRIC) : ST_TRIAXIAL); };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
    virtual double getGamma() const {return Gamma; }
 //   virtual double totalMass() const { return mass; }
private:
    const double mass;       ///< total mass of the model
    const double scalerad;   ///< scale radius
    const double q, p;       ///< axis ratio (y/x and z/x) of equidensity surfaces
    const double Gamma;      ///< cusp exponent for Dehnen potential
};

/** parent class for all potential expansions based on spherical harmonics for angular variables **/
class CPotentialSH: public CPotential
{
public:
    CPotentialSH(size_t _Ncoefs_angular): 
        CPotential(), Ncoefs_angular(std::min<size_t>(MAX_NCOEFS_ANGULAR-1, _Ncoefs_angular)), mysymmetry(ST_TRIAXIAL) { assignlmrange(); };
    virtual SYMMETRYTYPE symmetry() const { return mysymmetry; };
    virtual double Rho(double X, double Y, double Z, double t=0) const;
    virtual double Phi(double X, double Y, double Z, double t=0) const;
    /// common function for all derivative classes
    virtual void Force(const double xyz[N_DIM], const double t, double* force, double* forceDeriv=NULL) const;
    /// return l_max -- the order of angular expansion
    size_t getNcoefs_angular() const { return Ncoefs_angular; }
protected:
    size_t Ncoefs_angular;               ///< l_max, the order of angular expansion (0 means spherically symmetric model)
    SYMMETRYTYPE mysymmetry;             ///< may have different type of symmetry
    int lmax, lstep, mmin, mmax, mstep;  ///< range of angular coefficients used for given symmetry
    void assignlmrange();                ///< assigns the above variables based on mysymmetry, should be called whenever mysymmetry has changed

    /** The function that computes spherical-harmonic coefficients for potential 
        and its radial (first/second) derivative at given radius.
        Must be implemented in derived classes, and is used in evaluation of potential and forces; 
        unnecessary coefficients are indicated by passing NULL for coefs** and should not be computed.
    */
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const = 0;   
};

/** basis-set expansion on the Zhao(1996) basis set (alpha models) **/
class CPotentialBSE: public CPotentialSH
{
public:
    /// init coefficients from a discrete point mass set
    template<typename NumT> 
    CPotentialBSE(double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym=ST_TRIAXIAL, bool checkMonotonic=true);
    /// load coefficients from stored values
    CPotentialBSE(double _Alpha, const std::vector< vectord > &coefs);
    /// init potential coefficients from an analytic mass model
    CPotentialBSE(double _Alpha, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CDensity* density);
    virtual CPotential* clone() const { return new CPotentialBSE(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_BSE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "BSE"; };
    /// a faster estimate of M(r) from the l=0 harmonic only
    virtual double Mass(const double r) const;
    //  get functions:
    /// return BSE coefficients array
    void getCoefs(std::vector< vectord > *coefsArray) const
    { if(coefsArray!=NULL) *coefsArray=SHcoefs; };
    /// return the shape parameter of basis set
    double getAlpha() const { return Alpha; };
    /// return the number of radial basis functions
    size_t getNcoefs_radial() const { return Ncoefs_radial; }
private:
    size_t Ncoefs_radial;          ///< number of radial basis functions [ =SHcoefs.size() ]
    std::vector<vectord> SHcoefs;  ///< array of coefficients A_nlm of potential expansion
    double Alpha;                  ///< shape parameter controlling inner and outer slopes of basis functions

    /// compute angular expansion coefs at the given radius
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const;

    /// compute coefficients from a discrete point mass set; 
    /// if Alpha=0 then it is computed automatically from the data
    template<typename NumT> 
    void prepareCoefsDiscrete(const CPointMassSet<NumT> &points);

    /// compute coefficients from a smooth mass profile; 
    ///if Alpha=0 then it is chosen automatically from density->getGamma()
    void prepareCoefsAnalytic(const CDensity* density);
    
    void initDefault();           ///< called as a default initialization when everything else fails
    void checkSymmetry();         ///< assigns symmetry class if some coefficients are (near-)zero
};

/** basis-set expansion using spherical Bessel functions **/
class CPotentialBSECompact: public CPotentialSH
{
public:
    /// init coefficients from a discrete point mass set
    template<typename NumT> 
    CPotentialBSECompact(double _Rmax, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym=ST_TRIAXIAL /*, bool checkMonotonic=true*/);
    /// load coefficients from stored values
    CPotentialBSECompact(double _Rmax, const std::vector< vectord > &coefs);
    /// init potential coefficients from an analytic mass model
    CPotentialBSECompact(double _Rmax, size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CDensity* density);
    virtual CPotential* clone() const { return new CPotentialBSECompact(*this); };
    virtual POTENTIALTYPE PotentialType() const { return PT_BSECOMPACT; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "BSECompact"; };
    /// a faster estimate of M(r) from the l=0 harmonic only
    virtual double Mass(const double r) const;
    //  get functions:
    /// return BSE coefficients array
    void getCoefs(std::vector< vectord > *coefsArray) const
    { if(coefsArray!=NULL) *coefsArray=SHcoefs; };
    /// return the radial extent of the system
    double getRmax() const { return Rmax; };
    /// return the number of radial basis functions
    size_t getNcoefs_radial() const { return Ncoefs_radial; }
private:
    size_t Ncoefs_radial;          ///< number of radial basis functions [ =SHcoefs.size() ]
    std::vector<vectord> SHcoefs;  ///< array of coefficients A_nlm of potential expansion
    std::vector<vectord> alpha;    ///< array of scale factors for radial coordinate
    double Rmax;                   ///< maximum radius of the system of finite extent

    /// prepare array of scale factors
    void prepareAlpha();

    /// compute angular expansion coefs at the given radius
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const;

    /// compute coefficients from a discrete point mass set; 
    template<typename NumT> 
    void prepareCoefsDiscrete(const CPointMassSet<NumT> &points);

    /// compute coefficients from a smooth mass profile; 
    void prepareCoefsAnalytic(const CDensity* density);
    
    void initDefault();           ///< called as a default initialization when everything else fails
    void checkSymmetry();         ///< assigns symmetry class if some coefficients are (near-)zero
};

/** spherical-harmonic expansion of potential with coefficients being spline functions of radius **/
class CPotentialSpline: public CPotentialSH
{
public:
    /// init potential from N point masses with given assumed symmetry type, 
    /// may also provide desired grid radii (otherwise assigned automatically)
    template<typename NumT>
    CPotentialSpline(size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CPointMassSet<NumT> &points, SYMMETRYTYPE _sym=ST_TRIAXIAL, double smoothfactor=0, 
        const vectord *_gridradii=NULL);
    /// init potential from stored SHE coefficients at given radii
    CPotentialSpline(const vectord &_gridradii, const std::vector< vectord > &_coefs);
    /// init potential from analytic mass model, 
    /// may also provide desired grid radii (if not given, assign automatically)
    CPotentialSpline(size_t _Ncoefs_radial, size_t _Ncoefs_angular, 
        const CDensity* density, const vectord *_gridradii=NULL);
    /** not really a constructor, but a way to compute potential expansion coefficients 
        for a set of discrete point masses at given radii.
        \param[in] points contains the source set of point masses;
        \param[in] _Ncoefs_angular is the order of angular expansion (l_max);
        \param[in] radii contains the array of radii at which the coefficients should be computed;
        \param[out] coefsArray contains the computed coefficients.
        Smoothing is not performed, splines are not initialized and the potential is not usable after this call!
    */
    template<typename NumT> 
    CPotentialSpline(size_t _Ncoefs_angular, const CPointMassSet<NumT> &points, 
        const vectord *radii, std::vector< vectord > *coefsArray);
    ~CPotentialSpline();
    virtual CPotential* clone() const;
    virtual POTENTIALTYPE PotentialType() const { return PT_SPLINE; };
    virtual std::string PotentialName() const { return myName(); };
    static std::string myName() { return "Spline"; };
    virtual double Mass(const double r) const;  ///< faster estimate of M(r) using l=0 terms only
    // get functions
    size_t getNcoefs_radial() const { return Ncoefs_radial; }   ///< return the number of radial points in the spline (excluding r=0)
    /** compute SHE coefficients at given radii.
        \param[in] useNodes tells whether the radii in question are 
                   the original spline nodes (true) or user-specified radii (true);
        \param[in,out] radii: if useNodes=true, this array is filled with spline nodes [output parameter, but must point to an existing array], 
                   otherwise the coefficients are computed at the supplied radii [input parameter];
        \param[out] coefsArray is filled by the computed coefficients (must point to an existing array)
    */
    void getCoefs(vectord *radii, std::vector< vectord > *coefsArray, bool useNodes=true) const;
private:
    size_t Ncoefs_radial;                ///< number of radial coefficients (excluding the one at r=0)
    vectord gridradii;                   ///< defines nodes of radial grid in splines
    double minr, maxr;                   ///< definition range of splines; extrapolation beyond this radius 
    double gammain,  coefin;             ///< slope and coef. for extrapolating potential inside minr (spherically-symmetric part, l=0)
    double gammaout, coefout, der2out;   ///< slope and coef. for extrapolating potential outside maxr (spherically-symmetric part, l=0)
    double potcenter, potmax, potminr;   ///< (abs.value) potential in the center (for transformation of l=0 spline), at the outermost spline node, and at 1st spline node
    std::vector<gsl_spline*> splines;    ///< spline coefficients at each harmonic
    vectord slopein, slopeout;           ///< slope of coefs for l>0 for extrapolating inside rmin/outside rmax

    /// reimplemented function to compute the coefficients of angular expansion of potential at the given radius
    virtual void computeSHCoefs(const double r, double coefsF[], double coefsdFdr[], double coefsd2Fdr2[]) const;

    /// assigns symmetry class if some coefficients are (near-)zero.
    /// called on an intermediate step after computing coefs but before initializing splines
    void checkSymmetry(const std::vector< vectord > &coefsArray);

    /// called as a default initialization when everything else fails
    void initDefault();

    /// create spline objects for all non-zero spherical harmonics from the supplied radii and coefficients.
    /// radii should have "Ncoefs_radial" elements and coefsArray - "Ncoefs_radial * (Ncoefs_angular+1)^2" elements
    void initSpline(const vectord &radii, const std::vector< vectord > &coefsArray);

    /** calculate (non-smoothed) spherical harmonic coefs for a discrete point mass set.
        \param[in] srcpoints contains the positions and masses of particles that serve as the input.
        There are two modes of operation, depending on the content of "outradii" vector: 

        - in the context of initialization of potential from a discrete point set, "srcradii"=NULL;
        "outradii" will be initialized with radii of all particles sorted in ascending order, 
        and "outcoefs" will contain coefficients of SH expansion for all points, 
        "outradii" and "outcoefs" should point to existing arrays; in case of error outcoefs is set to zero length.
        \note: the order of indexes in outcoefs is swapped w.r.t. standard convention in this file:
        outcoefs[coefind][pointind] - this is done to save memory by not allocating arrays for unused coefind'ices.

        - in the context of SHgrid Schwarzschild modelling, this is called from a quasi-constructor 
        for the purpose of computing SH coefs at a given set of radii. 
        This set is passed as "srcradii" array, and coefs are returned in "outcoefs"
        in the standard way: outcoefs[pointind][coefind]. 
        "outcoefs" should point to an existing array, in case of error it is filled with zeroes.
    */
    template<typename NumT> 
    void computeCoefsFromPoints(const CPointMassSet<NumT> &points, const vectord* srcradii, 
        vectord* outradii, std::vector< vectord > *outcoefs);

    /** create smoothing splines from the coefficients computed at each particle's radius.
        \param[in] userradii may contain the radial grid to construct the splines; 
                   if it is NULL then the radial grid is constructed automatically 
                   taking into account the range of radii covered by source points.
        \param[in] smoothfactor determines how much smoothing is applied to the spline 
                   (good results are obtained for smoothfactor=1-2).  
    */
    template<typename NumT>
    void prepareCoefsDiscrete(const CPointMassSet<NumT> &points, double smoothfactor, const vectord* userradii);

    /** compute expansion coefficients from an analytical mass profile.
        \param[in] srcradii may contain the radial grid; if NULL then it is assigned automatically.
    */
    void prepareCoefsAnalytic(const CDensity* density, const vectord *srcradii);

    /// evaluate value and optionally up to two derivatives of l=0 coefficient, 
    /// taking into account extrapolation beyond the grid definition range and log-scaling of splines
    void coef0(double r, double *val, double *der, double *der2) const;

    /// evaluate value, and optionally first and second derivative of l>0 coefficients 
    /// (lm is the combined index of angular harmonic >0); corresponding values for 0th coef must be known
    void coeflm(size_t lm, double r, double xi, double *val, double *der, double *der2, 
        double c0val, double c0der=0, double c0der2=0) const;
};

///@}
/// \name Correspondence between potential/density names and corresponding classes
///@{

/// lists all 'true' potentials, i.e. those providing a complete density-potential(-force) pair
typedef std::map<CDensity::POTENTIALTYPE, std::string> PotentialNameMapType;

/// lists all analytic density profiles (including those that don't have corresponding potential, 
/// but excluding general-purpose expansions)
typedef std::map<CDensity::POTENTIALTYPE, std::string> DensityNameMapType;

/// lists available symmetry types
typedef std::map<CDensity::SYMMETRYTYPE, std::string > SymmetryNameMapType;

extern PotentialNameMapType PotentialNames;
extern DensityNameMapType DensityNames;
extern SymmetryNameMapType SymmetryNames;
CDensity::POTENTIALTYPE getPotentialTypeByName(std::string PotentialName);  ///< perform reverse lookup in PotentialNames map
CDensity::POTENTIALTYPE getDensityTypeByName(std::string DensityName);      ///< same for DensityNames
CDensity::SYMMETRYTYPE getSymmetryTypeByName(std::string SymmetryName);     ///< same for symmetry

/// structure that contains parameters for all possible potentials
struct CConfigPotential
{
    double mass;                             ///< total mass of the model (not applicable to all potential types)
    double scalerad;                         ///< scale radius of the model (if applicable)
    double scalerad2;                        ///< second scale radius of the model (if applicable)
    double q, p;                             ///< axis ratio of the model (if applicable)
    double Gamma;                            ///< central cusp slope (for Dehnen and scale-free models)
    double sersicIndex;                      ///< Sersic index (for Sersic density model)
    size_t Ncoefs_radial, Ncoefs_angular;    ///< number of radial and angular coefficients in spherical-harmonic expansion
    size_t Ncoefs_vertical;                  ///< number of coefficients in z-direction for Cylindrical potential
    double Alpha;                            ///< shape parameter for BSE potential
    double Rmax;                             ///< radius of finite density model for BSECompact potential
    double nbpotEps;                         ///< treecode smooothing length (negative means adaptive smoothing based on local density, absolute value is the proportionality coefficient between eps and mean interparticle distance)
    double nbpotTol;                         ///< tree cell opening angle
    CDensity::POTENTIALTYPE PotentialType;   ///< currently selected potential type
    CDensity::POTENTIALTYPE DensityType;     ///< if pot.type == BSE or Spline, this gives the underlying density profile approximated by these expansions or flags that an Nbody file should be used
    CDensity::SYMMETRYTYPE SymmetryType;     ///< if using Nbody file with the above two potential expansions, may assume certain symmetry on the coefficients (do not compute them but just assign to zero)
    double splineSmoothFactor;               ///< for smoothing Spline potential coefs initialized from discrete point mass set
    double splineRadiusMin, splineRadiusMax; ///< if nonzero, specifies the inner- and outermost grid node radii
    double splineZMin, splineZMax;           ///< if nonzero, gives the grid extent in z direction for Cylindrical spline potential
    std::string NbodyFile;                   ///< name of file with coordinates of points, or coefficients of expansion, or any other external data array
    double Mbh;                              ///< mass of central black hole (in the composite potential)
};
extern const CConfigPotential defaultConfigPotential;

///@}
/// \name Various convenience functions for potential
///@{

/** Convenience function to return the total energy (or Jacobi integral if Omega!=0) for a given position+velocity point. */
template<typename NumT> double totalEnergy(const CPotential* potential, const CPosVelPoint<NumT> &point, double Omega=0, double t=0)
{
    return potential->Phi(point.Pos[0], point.Pos[1], point.Pos[2], t)
    + (pow_2(point.Vel[0])+pow_2(point.Vel[1])+pow_2(point.Vel[2]))/2
    - Omega*(point.Pos[0]*point.Vel[1]-point.Pos[1]*point.Vel[0]); 
}

}