/** \file    raga_core.h
    \brief   Main business logic
    \author  EV
    \date    2013-2014

This module implements the driver routine for the self-consistent secular evolution code RAGA.
The input parameters to the code are:
  - input snapshot serving as initial conditions
  - initial potential (may also be computed from initial snapshot)
  - total integration time (T_final)
  - interval for outputting snapshots of current locations of particles (T_output)
  - interval for updating potential (one step) (T_update)
  - number of sampling points in trajectory of each particle per update step (N_sample)
  - prescribed two-body relaxation rate
  - prescribed symmetry of potential?

  If T_update>=T_output, the run is essentially in a fixed potential. 
  If N_sample=1 and T_update=one integration timestep, this is equivalent to standard SCF approach.
*/
#pragma once
#include "potential.h"
#include "orbitint.h"

namespace smile{

    struct CConfigRaga {
        double timeTotal;       ///< total simulation time
        double episodeLength;   ///< duration of one episode
        double timestepOutput;  ///< interval of storing the trajectories in the output file
        double factorTimestep;  ///< factor in timestep calculation for orbit-averaged diffusion
        double captureRadius;
        double maxPericenterTrackingRadius;
        unsigned int numSamplesPerEpisode;
        unsigned int gridSize;
        std::string fileInput;
        std::string fileOutputParticles;
        std::string fileOutputPotential;
        std::string fileOutputLosscone;
        std::string fileOutputBinary;
        std::string fileLog;
        bool updatePotential;
        bool updateDF;
        enum RAGAMODE {
            RM_FP,        ///< Fokker-Planck
            //RM_SPH_ISO,   ///< Spherical isotropic Monte-Carlo
            //RM_SPH_ANISO, ///< Spherical anisotropic Monte-Carlo
            RM_RAGA       ///< Direct real-space orbit integration
        };
        RAGAMODE mode;
    };
    
    class CBasicRagaTask;  // forward declaration

    class CRagaCore
    {
    public:
        CRagaCore():   
          configPotential(defaultConfigPotential), 
          configOrbit(defaultConfigOrbit),
          //configRaga(defaultConfigRaga),
          potential(NULL), massmodel(NULL), particles(NULL), timeSim(0) {}
        ~CRagaCore();
        bool init(const std::string configFileName="");
        void run();

    private:
        CConfigPotential configPotential;  ///< parameters related to the potential
        CConfigOrbit configOrbit;          ///< parameters related to orbit integration
        CConfigRaga configRaga;            ///< parameters of Raga simulation
        const CPotential* potential;       ///< potential used in orbit integration
        CMassModel* massmodel;             ///< associated spherical mass model
        CPointMassSet<double>* particles;  ///< particle snapshot at the beginning/end of each interval
        std::vector<CBasicRagaTask*> ragaTasks;  ///< array of runtime tasks
        double timeSim;
        
        bool loadSettings(const std::string& configFileName="");
        void runFokkerPlanck();
        void step();
    };
    
    /// utility function to compute total, potential and kinetic energy of a snapshot
    void computeTotalEnergyModel(const CPotential* potential, const CConfigPotential& configPotential, 
        const CPointMassSet<double>* particles, double time, double* Etot, double* Esum);
 
    /// utility function to check for existence of a file
    bool fileExists(const std::string& fileName);
    
}  // namespace