#include <cmath>
#include <fstream>
#include "raga_fp.h"
#include "utils.h"
///!!!
#include <iostream>
#include <gsl/gsl_integration.h>

namespace smile{

CSolverFP::CSolverFP(CMassModel* _model, size_t dim, double _tau, double _relaxrate,
    const std::vector<double> *masses, const std::vector<double> *energies) : 
  model(_model), tau(_tau), relaxrate(_relaxrate)
{
    xnode.assign(dim+1, 0);
    xcenter.assign(dim+2, 0);
    func.assign(dim, 0);
    deriv.assign(dim, 0);
    dens.assign(dim, 0);
    energy.assign(dim+1, 0);
    phasevol.assign(dim+1, 0);
    // init grid
    if(model->mass(0)>0)
        scaling=new CLogMeshScaling(10.0, 0*model->getPotentialAtCenter());
    else
        scaling=new CUniformMeshScaling();
    energy[0] = model->mass(0)>0 ? model->getPotentialAtCenter()*100 : model->pot(0);
    double xmin=scaling->X(energy[0]);
    double xmax=0;
    //double lowestgridcell=(xmax-xmin)/dim;
    if(model->mass(0)==0) {
        double a1=0.2, a2=0.4;
        double a=a1, b=3-2*a1-a2, c=a1+a2-2;
        for(size_t i=0; i<=dim; i++) {
            double x=i*1.0/dim;
            xnode[i]=xmin + (xmax-xmin)*(a*x+b*x*x+c*x*x*x);
        }
        xnode[dim]=xmax;
    } else {
        for(size_t i=0; i<=dim; i++)
            xnode[i]=xmin + (xmax-xmin)*i/dim;
    }
    for(size_t i=0; i<dim; i++)
    {
        xcenter[i+1]=(xnode[i+1]+xnode[i])/2;
        energy[i+1]=scaling->E(xcenter[i+1]);
        dens[i]=model->ded(energy[i+1]) / scaling->dXdE(energy[i+1]);
    }
    xcenter[0]=2*xnode[0]-xcenter[1];
    xcenter[dim+1]=2*xnode[dim]-xcenter[dim];
    // compute initial phase volume associated with energy[i]; 
    // adiabatic evolution will not change phase volume even though values of energy 
    // at cell centers will change
    for(size_t i=0; i<=dim; i++)
        phasevol[i] = model->phasevol(energy[i]);
    // init distribution function
    if(masses!=NULL && energies!=NULL && masses->size()==energies->size()) {
        for(size_t p=0; p<masses->size(); p++) {
            double x=scaling->X(energies->at(p));
            int index=-1;
            for(size_t i=0; i<dim && index<0; i++)
                if(x<xnode[i+1]) index=i;
            if(index>=0) 
                func[index]+=masses->at(p)/(xnode[index+1]-xnode[index])/dens[index];   ///? works with scaled energy too?
            else
                index=-1;
        }
    }
    else {
        for(size_t i=0; i<dim; i++)
            func[i]=model->distrf(energy[i+1]);
    }
    // init coefs
    v_diag=gsl_vector_alloc(dim);
    v_above=gsl_vector_alloc(dim-1);
    v_below=gsl_vector_alloc(dim-1);
    v_rhs=gsl_vector_alloc(dim);
    v_sol=gsl_vector_alloc(dim);    
    reinitDC();
}

void CSolverFP::step()
{
    for(size_t i=0; i<func.size(); i++)
        gsl_vector_set(v_rhs, i, func[i]);
    gsl_linalg_solve_tridiag(v_diag, v_above, v_below, v_rhs, v_sol);
    for(size_t i=0; i<func.size(); i++) {
        double newVal=gsl_vector_get(v_sol, i);
        deriv[i]=(newVal-func[i])/tau;
        func[i]=newVal;
    }
}

double CSolverFP::norm() const
{
    double sum=0;
    for(size_t i=0; i<func.size(); i++)
        sum+=func[i]*dens[i]*(xnode[i+1]-xnode[i]);
    return sum;
}

double CSolverFP::etot() const
{
    double sum=0;
    for(size_t i=0; i<func.size(); i++)
        sum+=func[i]*dens[i]*energy[i+1]*(xnode[i+1]-xnode[i]);
    return sum+model->mass(0)*model->getPotentialAtCenter();
}

double CSolverFP::ekin() const
{
    double sum=0;
    for(size_t i=0; i<func.size(); i++)
        sum+=func[i]*phasevol[i+1] * scaling->dXdE(energy[i+1]) * (xnode[i+1]-xnode[i]);
    return sum*1.5;
}

void blah(const CMassModel* model);
bool CSolverFP::reinitDF(bool updatePotential)
{
    size_t dim=func.size();
    /*std::vector<double> eeinit(dim);
    for(size_t i=0; i<func.size(); i++)
        eeinit[i]=func[i]*dens[i]*energy[i+1]*(xnode[i+1]-xnode[i]);*/
    //for(size_t kk=0; kk<1; kk++) {
    std::vector<double> energy1(energy.begin()+1, energy.end());
    if(!model->initDistrFuncEnergy(energy1, func))
        return false;
    if(updatePotential) {
        model->computeNewPotential();
        // adiabatically change energies at cell centers preserving the phase volume
        computeNewEnergies(*model, phasevol, &energy);
        // reinit computation grid and slightly adjust values of distr.func. at cell centers
        // while preserving the total mass in each grid cell
        std::vector<double> oldNcell(dim);
        for(size_t i=0; i<dim; i++) 
            oldNcell[i]=func[i]*dens[i]*(xnode[i+1]-xnode[i]);
        // reinit grid
        xnode[0]=scaling->X(energy[0]);
        for(size_t i=0; i<dim; i++)
        {
            xcenter[i+1]=scaling->X(energy[i+1]);
            dens[i]=model->ded(energy[i+1]) / scaling->dXdE(energy[i+1]);
            if(i>0) xnode[i]=(xcenter[i+1]+xcenter[i])/2;  // adjust cell boundaries as well
        }
        for(size_t i=0; i<dim; i++)
            func[i]=oldNcell[i]/dens[i]/(xnode[i+1]-xnode[i]);
        xcenter[0]=2*xnode[0]-xcenter[1];
        xcenter[dim+1]=2*xnode[dim]-xcenter[dim];
    }
    
    // compute diffusion coefficients for the updated distribution function (and possibly updated potential)
    model->initSplineDifCoefE();
    reinitDC();
    
    /*blah(model);
    double res=0;
    for(size_t i=1; i<dim; i++) {
        double Flux=cdiff[i]/(xcenter[i+1]-xcenter[i])*(Wplus[i]*func[i]-Wminus[i]*func[i-1]);
        res+=Flux*scaling->E(xnode[i]);
    }
    std::cout << "Discretized flux:="<<res<<"\n";*/
    /*std::ofstream s("tmp");
    s<<std::setprecision(12);
    for(size_t i=0; i<func.size(); i++)
        s << energy1[i] << "\t" << energy[i+1] << "\t" << eeinit[i] << "\t" << func[i]*dens[i]*energy[i+1]*(xnode[i+1]-xnode[i]) << "\n";*/
    return true;
}

void CSolverFP::reinitDC()
{
    size_t dim=func.size();
    //std::vector<double> cdrift(dim+1,0), cdiff(dim+1,0), Wplus(dim+1,0), Wminus(dim+1,0);
    cdrift.assign(dim+1,0); cdiff.assign(dim+1,0); Wplus.assign(dim+1,0); Wminus.assign(dim+1,0);
    for(size_t i=1; i<dim; i++)   // loop over interior grid nodes
    {
        double DEE;
        double enode=scaling->E(xnode[i]);
        model->DifCoefEL(enode, &DEE, NULL, NULL,NULL,false);
        cdiff[i] = model->ded(enode)*DEE/2*relaxrate * scaling->dXdE(enode);
        cdrift[i] = 16*M_PI*M_PI*model->masse(enode)*relaxrate;
        double w=cdrift[i]/cdiff[i]*(xcenter[i+1]-xcenter[i]);
        double W= w>0.1 ? w/2/sinh(w/2) : 1-1./24*w*w*(1-7./240*w*w);
        Wplus[i]=W*exp(w/2);
        Wminus[i]=W*exp(-w/2);
    }
    for(size_t i=0; i<dim; i++)
    {
        double cfl=tau/(xnode[i+1]-xnode[i])/dens[i];
        if(i>0) gsl_vector_set(v_below, i-1, 
            -cfl*cdiff[i]/(xcenter[i+1]-xcenter[i])*Wminus[i]);
        if(i<dim-1) gsl_vector_set(v_above, i, 
            -cfl*cdiff[i+1]/(xcenter[i+2]-xcenter[i+1])*Wplus[i+1]);
        gsl_vector_set(v_diag, i, 
            1 + cfl*(cdiff[i]/(xcenter[i+1]-xcenter[i])*Wplus[i] + cdiff[i+1]/(xcenter[i+2]-xcenter[i+1])*Wminus[i+1]));
    }
}

double CSolverFP::getmaxdiag() const
{
    double max=0;
    for(size_t i=0; i<func.size(); i++)
        max=std::max<double>(max, gsl_vector_get(v_diag, i));
    return max;
}

CSolverFP::~CSolverFP()
{
    delete scaling;
    gsl_vector_free(v_diag);
    gsl_vector_free(v_above);
    gsl_vector_free(v_below);
    gsl_vector_free(v_rhs);
    gsl_vector_free(v_sol);
}

void CSolverFP::exportGrid(const char* fileName) const
{
    std::ofstream out(fileName, std::ios_base::out);
    for(size_t i=0; i<func.size(); i++) {
        double Enode=scaling->E(xnode[i]);
        double Ecell=energy[i+1];  // same as scaling->E(xcenter[i+1])
        double Flux=i>0 ? cdiff[i]/(xcenter[i+1]-xcenter[i])*(Wplus[i]*func[i]-Wminus[i]*func[i-1]) : 0;
        double r=model->rmax(Ecell);
        out << Ecell << "\t" << func[i] << "\t" << func[i]*model->ded(Ecell) << "\t" << deriv[i] << "\t"
            << Enode << "\t" << Flux << "\t" << scaling->dXdE(Enode) << "\t"
            << r << "\t" << model->dens(r) << "\n";
    }
}

struct CHelperParam1 {
    const CMassModel* model;
    double value;
};

double find_phasevol(double E, void* param)
{
    return ((CHelperParam1*)param)->model->phasevol(E) - ((CHelperParam1*)param)->value;
}

void computeNewEnergies(const CMassModel& model, std::vector<double> &phasevol, std::vector<double> *energy)
{
    // find values of E at which the phase volume in the new potential is the same as old
    gsl_function Fnc;
    CHelperParam1 param;
    param.model=&model;
    Fnc.function=&find_phasevol;
    Fnc.params=&param;
    double eold;
    if(model.mass(0)>0) {
        param.value=phasevol.front();
        eold=model.getPotentialAtCenter();
        while(Fnc.function(eold, &param)>=0) eold*=4;  // ensure to bracket the first point from below
    } else
        eold=model.getPotentialAtCenter();  // this energy contains zero phase volume so will definitely bracket the root finding
    double eupp=model.getPotentialAtCenter();
    param.value=phasevol.back();
    while(Fnc.function(eupp, &param)<=0) eupp/=4;  // bracket from above
    energy->resize(phasevol.size());
    for(size_t i=0; i<phasevol.size(); i++) {
        param.value=phasevol[i];
        double enew=findRoot(&Fnc, eold, eupp);
        if(enew!=0) eold=enew;  // shouldn't have zero here.. but who knows
        energy->at(i)=enew;
    }
}

void exportModel(const CMassModel& model, const char* fileName)
{
    std::ofstream out(fileName, std::ios_base::out);
    double zeroMass=model.mass(0), totalMass=model.getTotalMass()-zeroMass;
    double gridrmin=0.125;
    while(gridrmin>1e-6 && model.mass(gridrmin)-zeroMass>totalMass*1e-8) gridrmin/=2;
    double gridrmax=8;
    while(gridrmax<1e6 && model.mass(gridrmax)-zeroMass<totalMass*(1-1e-3)) gridrmax*=2;
    const size_t numNodes=100;
    std::vector<double> gridRadius;
    createNonuniformGrid(numNodes, gridrmin, gridrmax, false, &gridRadius);
    out << "#r\tM(r)\tPhi=E\trho\tf(E)\tg(E)\tM(E)\tperiod(E)\tr_circ(E)\tL_circ(E)\tsigma\tsigma_los\tdens_proj\tD_EE\tD_E\n";
    out << 0 << "\t" << model.mass(0) << "\t" << smile::convertToString(model.pot(0),10) << "\n";
    for(size_t i=0; i<numNodes; i++)
    {
        double rad = gridRadius[i];
        double pot = model.pot(rad);
        double DEE, DE;
        model.DifCoefEL(pot, &DEE, &DE, NULL, NULL, false);
        out << rad << "\t" << smile::convertToString(model.mass(rad),12) << "\t" 
            << smile::convertToString(pot,10) << "\t" << smile::convertToString(model.dens(rad),10) << "\t" 
            << model.distrf(pot) << "\t" //<< model.drhodphi(pot) << "\t" << model.d2rhodphi2(pot) << "\t"
            << model.ded(pot) << "\t" << model.masse(pot) << "\t"
            << model.trad(pot) << "\t" << model.rcirc(pot) << "\t" << model.lcirc(pot) << "\t"
            << model.veldisp(rad) << "\t" << model.losvd(rad) << "\t" << model.densproj(rad) << "\t"
            << DEE << "\t" << DE << "\t";
        out << "\n";
    }
}

}; // namespace