#pragma once
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_math.h>
#include "massmodel.h"

namespace smile{

class CMeshScaling {
public:
    virtual double X(double E) const=0;
    virtual double E(double X) const=0;
    virtual double dXdE(double E) const=0;
    virtual ~CMeshScaling() {};
};

class CUniformMeshScaling: public CMeshScaling {
public:
    virtual double X(double E) const {return E;};
    virtual double E(double X) const {return X;};
    virtual double dXdE(double /*E*/) const {return 1;};
};

class CInvMeshScaling: public CMeshScaling {
public:
    virtual double X(double E) const {return gsl_isinf(E) ? -1 : E/(1-E);};
    virtual double E(double X) const {return X/(1+X);};
    virtual double dXdE(double E) const {return 1/((1-E)*(1-E));};
};

class CLogMeshScaling: public CMeshScaling {
public:
    explicit CLogMeshScaling(double _b, double _E0) : b(_b), E0(_E0), k(1/b-E0) {};
    virtual double X(double E) const {return E<E0 ? E0-k*log((1-b*E)/(1-b*E0)) : E;};
    virtual double E(double X) const {return X<E0 ? 1/b-k*exp((E0-X)/k) : X;};
    virtual double dXdE(double E) const {return (1-b*E0)/(1-b*E);};
private:
    double b, E0, k;
};

class CSolverFP {
public:
    CSolverFP(CMassModel* model, size_t dim, double tau, double relaxrate,
        const std::vector<double> *masses=NULL, const std::vector<double> *energies=NULL);
    ~CSolverFP();
    void step();
    //double fnc(size_t i) const { return func[i]; }
    //double der(size_t i) const { return deriv[i]; }
    double norm() const;
    double etot() const;
    double ekin() const;
    //double grid(size_t i) const { return energy[i+1]; }
    double getmaxdiag() const;
    bool reinitDF(bool updatePotential=true);
    void exportGrid(const char* fileName) const;
//private:
    CMassModel* model;
    double tau;
    double relaxrate;
    const CMeshScaling* scaling;
    std::vector<double> xnode, xcenter, func, deriv, dens, energy, phasevol;
    std::vector<double> cdrift, cdiff, Wplus, Wminus;
    gsl_vector *v_diag, *v_above, *v_below, *v_rhs, *v_sol;
    void reinitDC();
};

/** computes E(p) for an array of values of phase volume p, which is assumed to be sorted in order of increase */
void computeNewEnergies(const CMassModel& model, std::vector<double> &phasevol, std::vector<double> *energy);

void exportModel(const CMassModel& model, const char* fileName);

}; // namespace