#include "raga_tasks.h"
#include <cassert>
#include <fstream>
#include <algorithm>
#include <gsl/gsl_spline.h>
#include <gsl/gsl_math.h>
#include "io_nemo.h"
#include "configfile.h"
#include "massmodel.h"
#include "createpotential.h"
#include "iopotential.h"
#include "utils.h"
#include "orbitint.h"
#include "raga_fp.h"

// might need to uncomment the following line for NEMO
//extern "C" { char* usage=""; }

namespace smile {

template<typename NumT> double totalEnergySph(const CMassModel* model, const CPosVelPoint<NumT> &point)
{
    return model->pot(sqrt(pow_2(point.Pos[0])+pow_2(point.Pos[1])+pow_2(point.Pos[2])))
                    + 0.5*(pow_2(point.Vel[0])+pow_2(point.Vel[1])+pow_2(point.Vel[2]));
}

/*------ runtime functions ------*/

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeTrajEpisode::Timestep(const double told, const double t, CPosVelPoint<double>* /*vars*/)
{
    {
        int i1=static_cast<int>(told/timeStep); 
        int i2=static_cast<int>(t/timeStep);
        // store trajectory at regular intervals of time    
        for (int i=i1; i<=i2; i++)
        {
            double tt=timeStep*i;
            if (((tt>told) && (tt<=t)) || ((i==0) && (told==0)))  ///!!! added 0th point explicitly
            {
                CPosVelPoint<float> pt;
                for(unsigned int c=0; c<N_DIM; c++)
                {
                    pt.Pos[c]=static_cast<float>(orbit->getPos(c, tt));
                    pt.Vel[c]=static_cast<float>(orbit->getVel(c, tt));
                }
                traj.add(pt, static_cast<float>(timeStep));
            }
        }
    }
    return SR_TERMINATE;
}

//----------------------------------------------------------------------------//
COrbitRuntimeVelPerturber::COrbitRuntimeVelPerturber(const COrbit* _orbit, const CMassModel* _model, double _mult) : 
  CBasicOrbitRuntimeFnc(_orbit), model(_model!=NULL ? new CMassModel(*_model) : NULL), mult(_mult), 
  intTime(0), deltaE(0), deltaE2(0), deltaEcumul(0), deltaL(0), deltaL2(0), deltaLcumul(0) 
{ 
    for(size_t i=0; i<15; i++) 
        avgCoefs[i]=0; 
    if(model==NULL || mult==0) return;
    // find energy below which the perturbations are so strong that orbit-averaged 
    // energy and momentum diffusion coefficients will be used instead of local velocity diffusion coefs
    Eorbitavg=model->getMinEnergyOrbitAvg(mult);
};

COrbitRuntimeVelPerturber::~COrbitRuntimeVelPerturber()
{
    delete model;
}

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeVelPerturber::Timestep(const double told, const double t, CPosVelPoint<double> *vars)
{
    intTime+=t-told;
    if(model==NULL || mult==0) 
        return SR_TERMINATE;
    double timeinterval=t-told, timepassed=0, timestep;
    double rad=sqrt(pow_2(vars->Pos[0])+pow_2(vars->Pos[1])+pow_2(vars->Pos[2]));
    double pot=model->pot(rad);
    size_t numsteps=0;
    do{
        double dvpar, dv2par, dv2per;
        double vel=sqrt(pow_2(vars->Vel[0])+pow_2(vars->Vel[1])+pow_2(vars->Vel[2]));
        if(vel==0 || pot + vel*vel/2>0)
            return SR_TERMINATE;  // don't do anything for a non-moving particle (drag coef diverges), or for an unbound particle (let it go..)
        double mom=sqrt(   // current value of angular momentum (not used for computing instantaneous dif.coefs.)
            pow_2(vars->Pos[1]*vars->Vel[2]-vars->Pos[2]*vars->Vel[1]) +
            pow_2(vars->Pos[2]*vars->Vel[0]-vars->Pos[0]*vars->Vel[2]) +
            pow_2(vars->Pos[0]*vars->Vel[1]-vars->Pos[1]*vars->Vel[0]) );
        if(pot + vel*vel/2 < Eorbitavg) {  /// don't bother computing local coefs; just take orbit-averaged values and set new velocity
            double DEE, DE, deltaE;
            model->DifCoefEL(pot + vel*vel/2, &DEE, &DE);
            if(!gsl_finite(DEE+DE)) {
                my_error(FUNCNAME, 
                    "Error in averaged dif.coefs at E="+convertToString(pot+vel*vel/2)+", DEE="+
                    convertToString(DEE)+", DE="+convertToString(DEE));
                return SR_TERMINATE;
            }
            DEE*=mult;
            DE *=mult;
            timestep=timeinterval-timepassed;
            int ntrials=0;
            do{
                double rand1, rand2;
                getNormalRandomNumbers(&rand1, &rand2);
                deltaE = sqrt(DEE*timestep)*rand1 + DE*timestep;
                ntrials++;
            } while((!gsl_finite(deltaE) || deltaE<=-vel*vel/2) && ntrials<16);
            if(ntrials<16) {  // set new velocity with random direction and updated magnitude
                double vmag=sqrt(2*deltaE+vel*vel);
                getRandomUnitVector(&(vars->Vel[0]), &vars->Vel[1], &vars->Vel[2]);
                vars->Vel[0]*=vmag;
                vars->Vel[1]*=vmag;
                vars->Vel[2]*=vmag;
                //std::cerr << "+";
            }
            // add current values of drift and diffusion coefficients, for computing time-averaged values
            avgCoefs[0] += timestep*DEE;  // D_EE
            avgCoefs[6] += timestep*DE;   // D_E
            // skip accumulation of averaged values of D_EL, D_LL, D_L, as they are more costly to compute
            //avgCoefs[14] += timestep*mom;  // avg.ang.mom.
        } else { 
            // compute local velocity diffusion coefs
            model->DifCoefVel(rad, vel, &dvpar, &dv2par, &dv2per);
            dvpar *=mult;
            dv2par*=mult;
            dv2per*=mult;
            if(!gsl_finite(dvpar+dv2par+dv2per)) {
                my_error(FUNCNAME,
                    "Error in dif.coef. at r="+convertToString(rad)+",v="+convertToString(vel)+
                    ",E="+convertToString(model->pot(rad)+vel*vel/2)+": dvpar="+convertToString(dvpar)+
                    ",dv2par="+convertToString(dv2par)+",dv2per="+convertToString(dv2per));
                return SR_TERMINATE;  // some weird error, dump it
            }
            double maxdeltav=0.4*std::min<double>(vel, sqrt(std::max<double>(-2*pot-vel*vel,-0.001*pot)) );
            double maxtimestep=std::min<double>(pow_2(maxdeltav)/(dv2par+dv2per), maxdeltav/fabs(dvpar/vel));
            timestep = std::min<double>(maxtimestep, timeinterval-timepassed);   // may be necessary to perform several steps
            // add current values of drift and diffusion coefficients, for computing time-averaged values
            avgCoefs[0] += timestep*vel*vel*dv2par;  // D_EE
            avgCoefs[2] += timestep*(pow_2(mom/vel)*(dv2par-dv2per/2) + rad*rad/2*dv2per);  // D_LL
            avgCoefs[4] += timestep*mom*dv2par;  // D_EL
            avgCoefs[6] += timestep*(dvpar + dv2par/2 + dv2per/2);  // D_E
            avgCoefs[8] += timestep*(mom>0 ? mom/pow_2(vel)*dvpar + rad*rad/4/mom*dv2per : 0); // D_L
            //avgCoefs[14] += timestep*mom;  // avg.ang.mom.
            // sample actual value of velocity perturbation
            double newv[3];
            int ntrials=0;
            do {
                double rand1, rand2;
                getNormalRandomNumbers(&rand1, &rand2);
                double deltavpar = rand1*sqrt(dv2par*timestep) + dvpar/vel*timestep;
                double deltavper = rand2*sqrt(dv2per*timestep);
                if(!gsl_finite(deltavpar+deltavper)) {
                    my_error(FUNCNAME,
                        "Error in dif.coef. at r="+convertToString(rad)+",v="+convertToString(vel)+
                        ",E="+convertToString(model->pot(rad)+vel*vel/2)+": dvpar="+convertToString(dvpar)+
                        ",dv2par="+convertToString(dv2par)+",dv2per="+convertToString(dv2per)+
                        " results in deltavpar="+convertToString(deltavpar)+",deltavper="+convertToString(deltavper));
                    return SR_TERMINATE;  // some weird error, dump it
                }
                // assign parallel and perpendicular velocity components
                double rx, ry, rz;  // random vector, assuming 3d motion (diffusion coefs are not valid for 2d anyway)
                getRandomUnitVector(&rx, &ry, &rz);
                double rdotv=rx*vars->Vel[0]+ry*vars->Vel[1]+rz*vars->Vel[2];  // projection of random vector on the velocity
                rx -= rdotv*vars->Vel[0]/(vel*vel);
                ry -= rdotv*vars->Vel[1]/(vel*vel);
                rz -= rdotv*vars->Vel[2]/(vel*vel); // subtract parallel component
                double rmag=sqrt(pow_2(rx)+pow_2(ry)+pow_2(rz));
                if(rmag==0) rmag=1;  // avoid singularity
                // add perturbations to the velocity
                newv[0] = vars->Vel[0] + vars->Vel[0]/vel*deltavpar + rx/rmag*deltavper;
                newv[1] = vars->Vel[1] + vars->Vel[1]/vel*deltavpar + ry/rmag*deltavper;
                newv[2] = vars->Vel[2] + vars->Vel[2]/vel*deltavpar + rz/rmag*deltavper;
                ntrials++;
            }   // check that new energy is still below zero
            while( ntrials<16 && pot + (pow_2(newv[0])+pow_2(newv[1])+pow_2(newv[2]))/2 >= 0 );
            if(ntrials<16 && gsl_finite(newv[0]+newv[1]+newv[2])) {  // update velocity if it will not unbind the particle
                vars->Vel[0]=newv[0];
                vars->Vel[1]=newv[1];
                vars->Vel[2]=newv[2];
            }
        }
        // record actual changes in energy and momentum resulting from perturbations
        double mom_new=sqrt(
            pow_2(vars->Pos[1]*vars->Vel[2]-vars->Pos[2]*vars->Vel[1]) +
            pow_2(vars->Pos[2]*vars->Vel[0]-vars->Pos[0]*vars->Vel[2]) +
            pow_2(vars->Pos[0]*vars->Vel[1]-vars->Pos[1]*vars->Vel[0]) );
        //double deltaEstep= pow_2(deltavper)/2 + pow_2(deltavpar)/2 + vel*deltavpar;
        double deltaEstep=(pow_2(vars->Vel[0])+pow_2(vars->Vel[1])+pow_2(vars->Vel[2]) - vel*vel)*0.5;
        double deltaLstep= mom_new-mom;
        deltaEcumul+=deltaEstep; // current accumulated difference
        deltaE += deltaEcumul*timestep;
        deltaE2+= pow_2(deltaEcumul)*timestep;
        deltaLcumul+=deltaLstep; // current accumulated difference
        deltaL += deltaLcumul*timestep;
        deltaL2+= pow_2(deltaLcumul)*timestep;
        timepassed+=timestep;
        numsteps++;
        if(!gsl_finite(deltaE)) 
            my_error(FUNCNAME,
            "Energy is "+convertToString(deltaE)+" at time t="+convertToString(t));
    } while(timepassed<timeinterval && numsteps<100);   // last condition to avoid deadlock for v->0?
    return SR_REINIT;
}

void COrbitRuntimeVelPerturber::Finish()
{
    if(model==NULL || mult==0 || intTime==0) return;
    // compute actual mean and mean-square changes in energy and ang.mom. during the entire integration
    avgCoefs[10] = deltaE/intTime;
    avgCoefs[11] = sqrt(deltaE2/intTime);
    avgCoefs[12] = deltaL/intTime;
    avgCoefs[13] = sqrt(deltaL2/intTime);
    avgCoefs[14] = deltaEcumul;
}

//----------------------------------------------------------------------------//
COrbitRuntimeLosscone::COrbitRuntimeLosscone(const COrbit* _orbit, double _captureRadius):
  CBasicOrbitRuntimeFnc(_orbit), captureRadius(_captureRadius),
  tcapt(0), Ecapt(0), L2(0), prevL2(0), deltaL2(0)
{
    // init sign
    CPosVelPoint<double> ic = orbit->getInitCond();
    drdtsign= ic.Pos[0]*ic.Vel[0]+ic.Pos[1]*ic.Vel[1]+ic.Pos[2]*ic.Vel[2] >0 ? 1 : -1;
}

/// \cond INTERNAL_DOCS
struct CPericenterParam {
    const COrbit* orbit;
};

double finddrdt(double t, void* params)
{
    const COrbit* o=static_cast<CPericenterParam*>(params)->orbit;
    return o->getPos(0,t)*o->getVel(0,t) + o->getPos(1,t)*o->getVel(1,t) + o->getPos(2,t)*o->getVel(2,t);
}
/// \endcond 

CBasicOrbitRuntimeFnc::STEPRESULT COrbitRuntimeLosscone::Timestep(const double told, const double t, CPosVelPoint<double>* vars)
{
    if(tcapt>0)  // already swallowed
        return SR_TERMINATE;
    if(t<=told) return SR_TERMINATE;  // shouldn't happen
    unsigned int N_dim=orbit->getConfig()->N_dim;  // shortcut
    int prevsign=drdtsign;
    drdtsign= vars->Pos[0]*vars->Vel[0] + vars->Pos[1]*vars->Vel[1] + vars->Pos[2]*vars->Vel[2] >0 ? 1 : -1;
    if(drdtsign*prevsign<0 && drdtsign>0)
    {
        // direction reversed -- passed pericenter; find exact passagee time
        gsl_function F;
        CPericenterParam params;
        params.orbit=orbit;
        F.function=&finddrdt;
        F.params=&params;
        double tperi=findRoot(&F, told, t, -1);
        if(tperi==-1)   // root finder failed - could happen if adding velocity perturbation has changed 
            tperi=told; // the sign of angular momentum at the beginning of this timestep to already positive
        double yperi[N_DIM*2];
        for(unsigned int i=0; i<N_DIM; i++) {
            yperi[i]=orbit->getPos(i, tperi);
            yperi[i+N_DIM]=orbit->getVel(i, tperi);
        }
        double rperi=sqrt(pow_2(yperi[0])+pow_2(yperi[1])+pow_2(yperi[2]));
        double Lx=(N_dim==3)? yperi[1]*yperi[5]-yperi[2]*yperi[4] : 0;
        double Ly=(N_dim==3)? yperi[2]*yperi[3]-yperi[0]*yperi[5] : 0;
        double Lz=(N_dim==3)? yperi[0]*yperi[4]-yperi[1]*yperi[3] : yperi[0]*yperi[3]-yperi[1]*yperi[2];
        double L2=pow_2(Lx)+pow_2(Ly)+pow_2(Lz);
        if(rperi<=captureRadius)  {  // just had been captured by black hole
            tcapt=tperi;
            deltaL2=prevL2>0 ? prevL2-L2 : -1;
            Ecapt=totalEnergy(orbit->getPotential(), *vars);
            return SR_TERMINATE;
        }
        prevL2=L2;
        maxradius=0;  // reset value after each pericenter passage
    }
    double radius=sqrt(pow_2(vars->Pos[0])+pow_2(vars->Pos[1])+pow_2(vars->Pos[2]));
    // record the position at some point before the next pericenter passage (which may turn out to be the last one..)
    // this may be later used to determine what was the type of orbit just before capture (e.g. tube/saucer/pyramid/...)
    if(radius>maxradius) {
        maxradius=radius;
        capturePoint=*vars;  // store the apocenter point...
    }
    else if(radius>sqrt(captureRadius*maxradius))    // ...or some point in between apocenter and pericenter, but not too close to capture radius (to avoid roundoff errors)
        capturePoint=*vars;
    return SR_CONTINUE;
}

/*------ RAGA tasks -------*/

CRagaTaskTrajectory::CRagaTaskTrajectory(CPointMassSet<double>* _particles, const CPotential* const * _ptrPotential, 
    const std::string& _outputFileName, double _timestepOutput) : 
  CBasicRagaTask(_particles), ptrPotential(_ptrPotential), 
  outputFileName(_outputFileName), timestepOutput(_timestepOutput), prevOutputTime(0)
{
    writeSnapshot(particles, 0);  // write output snapshot at time 0
}

CRagaTaskTrajectory::~CRagaTaskTrajectory() 
{
    io_nemo(const_cast<char*>(outputFileName.c_str()), const_cast<char*>("close"));
}
    
template<typename NumT> bool CRagaTaskTrajectory::writeSnapshot(const CPointMassSet<NumT> *points, double dtime)
{
    size_t count=points->size();
    vectorf pos(count*3), vel(count*3), mass(count), poten(count);
    for(size_t i=0; i<count; i++) {
        pos[i*3+0]=static_cast<float>(points->at(i).first.Pos[0]);
        pos[i*3+1]=static_cast<float>(points->at(i).first.Pos[1]);
        pos[i*3+2]=static_cast<float>(points->at(i).first.Pos[2]);
        vel[i*3+0]=static_cast<float>(points->at(i).first.Vel[0]);
        vel[i*3+1]=static_cast<float>(points->at(i).first.Vel[1]);
        vel[i*3+2]=static_cast<float>(points->at(i).first.Vel[2]);
        mass[i]   =static_cast<float>(points->at(i).second);
        poten[i]  =static_cast<float>((*ptrPotential)->Phi(pos[i*3+0], pos[i*3+1], pos[i*3+2], dtime));
    }
    size_t* pcount=&count;
    float time=static_cast<float>(dtime);
    float* ptime=&time;
    float* ppos=&(pos.front());
    float* pvel=&(vel.front());
    float* pmass=&(mass.front());
    float* ppot=&(poten.front());
    return io_nemo(const_cast<char*>(outputFileName.c_str()), const_cast<char*>("float,n,t,m,x,v,p,save"), 
        &pcount, &ptime, &pmass, &ppos, &pvel, &ppot) != 0;
}

bool CRagaTaskTrajectory::episodeDone(double episodeStart, double episodeLength)
{
    if(episodeStart+episodeLength>=prevOutputTime+timestepOutput) {
        writeSnapshot(particles, episodeStart+episodeLength); ///!!! need to write all intermediate snapshots as well
        prevOutputTime=episodeStart+episodeLength;
    }
    return false;
}

//----------------------------------------------------------------------------//
CRagaTaskPotential::CRagaTaskPotential(
    CPointMassSet<double>* const _particles, 
    const CPotential** const _ptrPotential, 
    CMassModel** const _ptrMassModel, 
    const double* _ptrEpisodeLength, 
    size_t _numSamplesPerEpisode, 
    const CConfigPotential* _ptrConfigPotential, 
    const std::string& _outputPotentialFileName, 
    double _timestepOutput, 
    bool _updatePotential
) : 
  CBasicRagaTask(_particles), 
  ptrPotential(_ptrPotential), 
  ptrMassModel(_ptrMassModel), 
  ptrEpisodeLength(_ptrEpisodeLength), 
  numSamplesPerEpisode(_numSamplesPerEpisode), 
  ptrConfigPotential(_ptrConfigPotential), 
  outputPotentialFileName(_outputPotentialFileName),
  updatePotential(_updatePotential),
  timestepOutput(_timestepOutput),
  prevOutputTime(0) 
{
    exportPotential(0);
};

bool CRagaTaskPotential::exportPotential(double time)
{
    if(outputPotentialFileName.empty()) return false;
    if(*ptrMassModel != NULL) {  // export spherical mass model statistics
        exportModel(**ptrMassModel, (outputPotentialFileName+convertToString(time)+".tab").c_str());
    }
    const CPotential* thePotential=(*ptrPotential)->PotentialType()==CPotential::PT_COMPOSITE ? 
        static_cast<const CPotentialComposite*>(*ptrPotential)->component(0) : *ptrPotential;  // obtain the stellar component of potential (without BH)
    return writePotential(outputPotentialFileName+convertToString(time)+
        (thePotential->PotentialType()==CPotential::PT_BSE ? ".coef_bse" : 
         thePotential->PotentialType()==CPotential::PT_SPLINE ? ".coef_spl" : ".coef_pot"), thePotential);
    return true;
}
    
bool CRagaTaskPotential::episodeDone(double episodeStart, double episodeLength)
{
    // obtain all sampling points stored for each orbit
    assert(infoOrbit.size()==particles->size());
    CPointMassSet<float> points;
    points.data.reserve(infoOrbit.size()*numSamplesPerEpisode);
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_TRAJ_SAMPLE) {
            const CPointMassSet<float>::Type& data=static_cast<const CTrajEpisodeInformation*>(infoOrbit[ip])->data.data;
            for(size_t i=1; i<data.size(); i++) { ///!!! starting from 1th point, since 0th is the initial condition
                points.add(data[i].first, static_cast<float>(mass*data[i].second/episodeLength));
            }
        }
    }
    // procedure for updating the potential is specific to each PotentialType
    const CPotential* newPotential=NULL;
    const CPotential* oldPotential=(*ptrPotential)->PotentialType()==CPotential::PT_COMPOSITE ? 
        static_cast<const CPotentialComposite*>(*ptrPotential)->component(0) : *ptrPotential;  // obtain the stellar component of potential (without BH)
    if(updatePotential) {  // recompute potential at the end of episode
        if(oldPotential->PotentialType()==CPotential::PT_BSE) {
            newPotential = new CPotentialBSE(
                static_cast<const CPotentialBSE*>(oldPotential)->getAlpha(),   // keep the same shape parameter as before (don't use configPotential.Alpha as it may be 0-autodetect)
                ptrConfigPotential->Ncoefs_radial, ptrConfigPotential->Ncoefs_angular,
                points, ptrConfigPotential->SymmetryType, true);
        } else if(oldPotential->PotentialType()==CPotential::PT_SPLINE) {
            std::vector< vectord > oldCoefs;
            vectord gridRadii;
            static_cast<const CPotentialSpline*>(oldPotential)->getCoefs(&gridRadii, &oldCoefs, true);  // use the same radial grid as before
            newPotential = new CPotentialSpline(
                ptrConfigPotential->Ncoefs_radial, ptrConfigPotential->Ncoefs_angular,
                points, ptrConfigPotential->SymmetryType, 0, &gridRadii);
        } else if(oldPotential->PotentialType()==CPotential::PT_SPHERICAL) {
            vectord gridRadii, oldMass;
            static_cast<const CPotentialSpherical*>(oldPotential)->getModel()->getMassProfile(&gridRadii, &oldMass);  // use the same radial grid
            newPotential = new CPotentialSpherical(points, ptrConfigPotential->splineSmoothFactor, &gridRadii);
            CMassModel::MODELSTATUS status=static_cast<const CPotentialSpherical*>(newPotential)->getModel()->getStatus();
            switch(status)
            {
            case CMassModel::MS_DF_NEGATIVE:
            case CMassModel::MS_DF_UNDEFINED:
                my_message(__FUNCTION__, "Warning, isotropic DF is poorly defined");  // it is not used anyway
                break;
            case CMassModel::MS_OK:
                break;
            case CMassModel::MS_MASS_NOT_ASC:
                my_message(__FUNCTION__, "Mass not increasing with radius; retained old potential");
                delete newPotential;
                return true;
            default:  // shouldn't occur...
                my_error(__FUNCTION__, "Error in creating new CPotentialSpherical");
                delete newPotential;
                return true;
            }
        } else {  // shouldn't happen - we have checked potential type at start
            my_error(__FUNCTION__, "Unknown potential type is set to be updated");
            return true;  // signal of error
        }
    } else {  // updatePotential==false means that we don't change the stellar potential, but still should reinit the total potential (with black hole)
        newPotential=oldPotential->clone();
    }
    if(ptrConfigPotential->Mbh>0) {  // create a composite potential with black hole(s)
        std::vector<const CPotential*> components(1, newPotential);
        newPotential = new CPotentialComposite(components, ptrConfigPotential->Mbh);
    }
        double totalEcorr=0, totalMass=0;
        for(size_t ip=0; ip<infoOrbit.size(); ip++) {
            double pmass=particles->at(ip).second;
            if(pmass==0 || infoOrbit[ip]==NULL) 
                continue;
            const CPosVelPoint<float>& pOld = static_cast<const CTrajEpisodeInformation*>(infoOrbit[ip])->data.data[0].first;  // 0th point is the initial condition
            const CPosVelPoint<double>&pNew = particles->at(ip).first;
            double oldPotOldPos=(*ptrPotential)->Phi(pOld.Pos[0], pOld.Pos[1], pOld.Pos[2], 0);
            double oldPotNewPos=(*ptrPotential)->Phi(pNew.Pos[0], pNew.Pos[1], pNew.Pos[2], episodeLength);
            double newPotOldPos=newPotential->Phi(pOld.Pos[0], pOld.Pos[1], pOld.Pos[2], -episodeLength);
            double newPotNewPos=newPotential->Phi(pNew.Pos[0], pNew.Pos[1], pNew.Pos[2], 0);
            totalEcorr+=pmass*(newPotOldPos-oldPotNewPos)*0.5;
            double deltaEpart =   // this is how much we need to correct energy of this particle
                0.5*(newPotOldPos - oldPotOldPos)
               -0.5*(newPotNewPos - oldPotNewPos);
            double Vnew2 = pow_2(pNew.Vel[0])+pow_2(pNew.Vel[1])+pow_2(pNew.Vel[2]);
            double Vupd2 = Vnew2 + deltaEpart*2;
            if(Vupd2<Vnew2*0.25) {   // correction is too large
                //my_error("At time "+convertToString(episodeStart+episodeLength)+" couldn't correct #"+convertToString(ip)+"'s Energy "+
                //         convertToString(totalEnergy(*ptrPotential, particles->at(ip).first))+" by "+convertToString(deltaEpart));
                Vupd2=Vnew2*0.25;  // reduce correction
                totalEcorr+=pmass*(Vupd2/2-Vnew2/2-deltaEpart); // and append undercorrected energy to the total correction budget
            }
            totalMass +=pmass;
            double Vmul=sqrt(Vupd2/Vnew2);
            for(size_t c=0; c<3; c++) particles->at(ip).first.Vel[c]*=Vmul;
        }
#ifdef DEBUGPRINT
        my_message(__FUNCTION__, 
            "Need additional correction -E="+convertToString(totalEcorr));
#endif
        // another round of corrections, to cancel any residual imbalance of energy
        for(size_t ip=0; ip<infoOrbit.size(); ip++) {
            double deltaEpart = -totalEcorr/totalMass;
            double Vold2 = pow_2(particles->at(ip).first.Vel[0])+pow_2(particles->at(ip).first.Vel[1])+pow_2(particles->at(ip).first.Vel[2]);
            double Vnew2 = Vold2 + deltaEpart*2;
            if(Vnew2<Vold2*0.25) {   // correction is too large
                Vnew2=Vold2*0.25;
            }
            double Vmul=sqrt(Vnew2/Vold2);
            for(size_t c=0; c<3; c++) particles->at(ip).first.Vel[c]*=Vmul;
        }

        delete (*ptrPotential);
        *ptrPotential=newPotential;

    // compute new distribution function in energy used to calculate diffusion coefficients
    if(*ptrMassModel != NULL) {   // massmodel is used
        if(updatePotential) {     // potential has changed after episode
            CMassModel* newModel = createMassModel(*ptrPotential);  // create a new model with the new potential
            if(newModel!=NULL) {
                delete *ptrMassModel; // delete spherical model associated with the old potential
                *ptrMassModel=newModel;
            }
            else
                my_error(__FUNCTION__, "Cannot update spherical mass model with the new potential; retained the old one");
        }
        vectord pointMasses(points.size()), pointEnergies(points.size());
        for(size_t i=0; i<points.size(); i++) {
            pointMasses[i]=points[i].second;
            pointEnergies[i]=totalEnergySph(*ptrMassModel, points[i].first);
        }
        vectord energyBins, distrFnc;
        if(!fitMassModel(pointEnergies, pointMasses, &energyBins, &distrFnc, ptrConfigPotential->splineSmoothFactor, 
            //configRaga.gridSize, configRaga.numSamplesPerStep*5, 
            0, 0, CMassModel::FT_DF_ENERGY, *ptrMassModel)) {
            my_error(__FUNCTION__, "Cannot compute distribution function");
        } else {
            if(!(*ptrMassModel)->initDistrFuncEnergy(energyBins, distrFnc)) {
                my_error(__FUNCTION__, "Could not re-initialize distribution function");
                std::ofstream strm(("baddf"+convertToString(episodeStart+episodeLength)).c_str());
                for(size_t b=0; b<energyBins.size(); b++)
                    strm << energyBins[b] << "\t" << distrFnc[b] << "\n";
            }
        }
        // recompute diffusion coefficients for the new potential and distribution function
        (*ptrMassModel)->initSplineDifCoefE();
        (*ptrMassModel)->initSplineDifCoefV();
    }
    if(episodeStart+episodeLength>=prevOutputTime+timestepOutput) {
        exportPotential(episodeStart+episodeLength);
        prevOutputTime=episodeStart+episodeLength;
    }
    return false;
}

//----------------------------------------------------------------------------//
CRagaTaskRelaxation::CRagaTaskRelaxation(
    CPointMassSet<double>* const _particles, 
    const CMassModel* const* _ptrMassModel, 
    double* _ptrEpisodeLength, 
    double _relaxRate
) :
  CBasicRagaTask(_particles), 
  ptrMassModel(_ptrMassModel), 
  ptrEpisodeLength(_ptrEpisodeLength), 
  relaxRate(_relaxRate) 
{
    double fiducialradius=(*ptrMassModel)->rad((*ptrMassModel)->getTotalMass()/infoOrbit.size());
    double veldisp=(*ptrMassModel)->veldisp(fiducialradius);
    avgtrelc=0.34*pow(veldisp,3)/(*ptrMassModel)->dens(fiducialradius) / relaxRate;
    if(gsl_isnan(avgtrelc) || gsl_isinf(avgtrelc)) avgtrelc=0;
}

bool CRagaTaskRelaxation::episodeDone(double /*episodeStart*/, double /*episodeLength*/)
{
    // task: correct particle energies for fluctuations due to relaxation
    assert(infoOrbit.size()==particles->size());
    vectord accumProportionParticle(infoOrbit.size());
    double accumProportion=0, deltaE=0;
    // compute energy change of each particle and its average energy diffusion coefficient
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_VEL_PERTURBER) {
            const float* vpdata = static_cast<const CVelPerturberInformation*>(infoOrbit[ip])->getData();
            // accumulated energy change by the end of interval resulting from relaxation only 
            // (not accounting for systematic change, e.g. due to time-dependent potential)
            deltaE += vpdata[14]*mass;
            // average value of energy diffusion coefficient
            accumProportionParticle[ip]=sqrt(vpdata[0]);
            accumProportion+=accumProportionParticle[ip]*mass;
        }
    }
    if(accumProportion==0) accumProportion=1;  // shouldn't probably happen, but will prevent division by 0

    // apply correction for energy fluctuations by changing particle velocity magnitudes
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double deltaEpart=accumProportionParticle[ip]/accumProportion * deltaE;
        double Vold2 = pow_2(particles->at(ip).first.Vel[0])+pow_2(particles->at(ip).first.Vel[1])+pow_2(particles->at(ip).first.Vel[2]);
        double Vnew2 = Vold2 - deltaEpart*2;
        if(Vnew2<Vold2*0.25) {   // correction is too large
            //my_error("At time "+convertToString(episodeStart+episodeLength)+" couldn't correct #"+convertToString(ip)+"'s energy "+
            //     convertToString(totalEnergySph(*ptrMassModel, particles->at(ip).first))+" by "+convertToString(-deltaEpart));
            Vnew2=Vold2*0.25;
        }
        double Vmul=sqrt(Vnew2/Vold2);
        for(size_t c=0; c<3; c++) particles->at(ip).first.Vel[c]*=Vmul;
    }
    
    // check if episode length needs to be shortened (close to core collapse)

        double fiducialradius=(*ptrMassModel)->rad((*ptrMassModel)->getTotalMass()/infoOrbit.size()*50);
        double veldisp=(*ptrMassModel)->veldisp(fiducialradius);
        double trelc=0.34*pow(veldisp,3)/(*ptrMassModel)->dens(fiducialradius) / relaxRate;
        //double rcore=3*veldisp/sqrt(4*M_PI*model->dens(fiducialradius));
        if(!gsl_isinf(trelc) && !gsl_isnan(trelc)) 
            avgtrelc=0.1*trelc+0.9*avgtrelc;
#ifdef DEBUGPRINT
    my_message(__FUNCTION__, 
        "Lowest energy for orbit-avg="+convertToString((*ptrMassModel)->getMinEnergyOrbitAvg(relaxRate))+
        ", potential at center="+convertToString((*ptrMassModel)->getPotentialAtCenter()));
#endif
    return false;
}

//----------------------------------------------------------------------------//
CRagaTaskLosscone::CRagaTaskLosscone(
    CPointMassSet<double>* const _particles, 
    double* _ptrMbh,
    double _captureRadius, 
    const std::string& _outputLossconeFileName
) :
  CBasicRagaTask(_particles), 
  ptrMbh(_ptrMbh), 
  captureRadius(_captureRadius), 
  outputLossconeFileName(_outputLossconeFileName) 
{
    particleMassInit.resize(particles->size());
    for(size_t ip=0; ip<particles->size(); ip++)
        particleMassInit[ip]=particles->at(ip).second;
    particlePericenterMinDist.assign(particles->size(), -1);  // default value means no information
    
    std::ofstream strmCaptD((outputLossconeFileName+".captdata").c_str());
    strmCaptD << "tcapt\tL2\tdeltaL2\tE\tmass\tindex\n";
    std::ofstream strmCaptO((outputLossconeFileName+".captorb" ).c_str());
    strmCaptO << "x\ty\tz\tvx\tvy\tvz\tweight\n";
}
    
bool CRagaTaskLosscone::episodeDone(double episodeStart, double episodeLength)
{
    // task: track loss-cone captures and optionally record pericenter distances
    assert(infoOrbit.size()==particles->size());
    double capturedMass=0;
    std::vector< std::pair<double, size_t> > captures;
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        double mass=particles->at(ip).second;
        if(mass>0 && infoOrbit[ip]!=NULL && infoOrbit[ip]->infoType()==CBasicInformation::IT_PERICENTER) {
            const CLossconeInformation* info = static_cast<const CLossconeInformation*>(infoOrbit[ip]);
            if(info->L2>0) {
                capturedMass += mass;
                captures.push_back(std::pair<double, size_t>(info->tcapt, ip));
            }
        }
    }
    if(captures.size()>0) {
        std::sort(captures.begin(), captures.end(), comparepair<double, size_t>);
        std::ofstream strmCaptD((outputLossconeFileName+".captdata").c_str(), std::ios_base::app);
        std::ofstream strmCaptO((outputLossconeFileName+".captorb" ).c_str(), std::ios_base::app);
        strmCaptO << std::setprecision(12);
        for(size_t c=0; c<captures.size(); c++) {
            size_t ip=captures[c].second;  // index of captured particle
            const CLossconeInformation* info=static_cast<const CLossconeInformation*>(infoOrbit[ip]);
            strmCaptD << (episodeStart+info->tcapt) << "\t" << info->L2 << "\t" 
                << info->deltaL2 << "\t" << info->Ecapt << "\t" << particles->at(ip).second << "\t" << ip << "\n";
            strmCaptO 
                << info->capturePoint.Pos[0] << "\t" << info->capturePoint.Pos[1] << "\t" << info->capturePoint.Pos[2] << "\t"
                << info->capturePoint.Vel[0] << "\t" << info->capturePoint.Vel[1] << "\t" << info->capturePoint.Vel[2] << "\t"
                << particles->at(ip).second << "\n";
            particles->at(ip).second=0;  // eliminate this particle
        }
        my_message(__FUNCTION__,
            "By time "+convertToString(episodeStart+episodeLength)+" captured "+convertToString(captures.size())+
            " particles, total mass="+convertToString(capturedMass));
        if(ptrMbh!=NULL)
            *ptrMbh += capturedMass;
    }
    // update the list of min.reached radii
    std::ofstream strmList((outputLossconeFileName+".listperi").c_str());
    for(size_t ip=0; ip<infoOrbit.size(); ip++) {
        if(particleMassInit[ip]>0 && particlePericenterMinDist[ip]>0)
            strmList << ip << "\t" << particleMassInit[ip] << "\t" << particlePericenterMinDist[ip] << "\n";
    }

    return false;
}

};  // namespace