/** \file    raga_tasks.h
    \brief   Various tasks that can be performed by Raga code
    \author  EV
    \date    2013-2014

This module implements various tasks that can be performed during 
orbit integration in the Raga code.


Each task is derived from CBasicRagaTask abstract class, 
creates a corresponding runtime function for each orbit during integration, 
which keeps track of the recorded data in an information class.

*/

#pragma once
#include "potential.h"

namespace smile{

    /** Trajectory stored at regular intervals during one episode. 
        This information object is used for writing down the output file with trajectories, 
        and for sampling the trajectories for updating the potential and distribution function.
    */
    class CTrajEpisodeInformation: public CBasicInformation
    {
    public:
        CTrajEpisodeInformation(const CPointMassSet<float>& src) : 
          data(src) {};
        virtual INFOTYPE infoType() const { return IT_TRAJ_SAMPLE; };
        virtual std::string toString() const { return ""; }
        CPointMassSet<float> data;
    };

    /** Runtime function that stores trajectory for each orbit */
    class COrbitRuntimeTrajEpisode: public CBasicOrbitRuntimeFnc
    {
    public:
        COrbitRuntimeTrajEpisode(const COrbit* _orbit, double _timeStep) : 
          CBasicOrbitRuntimeFnc(_orbit), timeStep(_timeStep), lastTime(0)
        {};
        virtual FNCTYPE FncType() const {return FT_TRAJ_SAMPLE; };
        virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
        virtual CBasicInformation* getData() const { return new CTrajEpisodeInformation(traj); };
    private:
        const double timeStep;        ///< time interval for recording the trajectory (must be less than duration of integration divided by numSamplingPoints)
        CPointMassSet<float> traj;    ///< recorded trajectory
        double lastTime;
    };

    /// information object for velocity perturber class
    /// (keep track of accumulated energy and angular momentum change and diffusion coefficients)
    class CVelPerturberInformation: public CBasicInformation
    {
    public:
        CVelPerturberInformation(const double _data[])
        { for(size_t i=0; i<15; i++)
            data[i]=static_cast<float>(_data[i]);
        };
        virtual INFOTYPE infoType() const { return IT_VEL_PERTURBER; };
        virtual std::string toString() const { return "";};
        /// get data
        const float* getData() const
        { return data; }
    private:
        float data[15];
    };

    /** class for applying random perturbations on velocity during orbit integration,
        mimicking the effect of two-body relaxation.
        The diffusion coefficients come in two varieties: local (position-dependent) coefficients for velocity,
        and orbit-averaged coefficients for energy and angular momentum.
        They are computed from a spherical mass model constructed as an approximation to the true (non-spherical)
        potential used in orbit integration itself. There is a free parameter that determines the scaling of
        these coefficients: its numerical value corresponds to \f$ (\ln\Lambda)/N \f$, where N is the number of
        stars in the model galaxy, and log Lambda is the Coulomb logarithm (~log N).
        */
    class COrbitRuntimeVelPerturber : public CBasicOrbitRuntimeFnc
    {
    public:
        COrbitRuntimeVelPerturber(const COrbit* _orbit, const CMassModel* _model, double _mult);
        virtual ~COrbitRuntimeVelPerturber();
        virtual FNCTYPE FncType() const { return FT_VEL_PERTURBER; };
        /// the timestep function applies a random perturbation to velocity in \param[in,out] vars,
        /// and records accumulated changes in E and L resulting from the perturbation
        virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
        virtual void Finish();   ///< fit pericenter distribution and compute fitting coefficients
        virtual CBasicInformation* getData() const
        { return new CVelPerturberInformation(avgCoefs); }
    private:
        const CMassModel* model;   ///< spherical mass model used to compute diffusion coefficients
        const double mult;         ///< multiplier for the diffusion coefficients (=\log\Lambda/N)
        double Eorbitavg;          ///< energy below which we use orbit-averaged dif.coefs instead of local ones
        double intTime;            ///< total integration time
        double deltaE, deltaE2, deltaEcumul;
        double deltaL, deltaL2, deltaLcumul;
        double avgCoefs[15];       ///< compute time-averaged coefficients DEE, DLL, DEL, DE, DL and their counterparts in spherical model
    };
    
    /** Information object for the COrbitRuntimeLosscone class */
    class CLossconeInformation: public CBasicInformation
    {
    public:
        CLossconeInformation(double _L2, double _deltaL2, double _tcapt, double _Ecapt, const CPosVelPoint<double> &_capturePoint) : 
          L2(_L2), deltaL2(_deltaL2), tcapt(_tcapt), Ecapt(_Ecapt), capturePoint(_capturePoint) {};
        virtual INFOTYPE infoType() const { return IT_PERICENTER; };
        virtual std::string toString() const { return std::string(); }
        double L2, deltaL2, tcapt, Ecapt;
        size_t ind;
        CPosVelPoint<double> capturePoint;
    };

    /** Keep track of close approach to the supermassive black hole */
    class COrbitRuntimeLosscone: public CBasicOrbitRuntimeFnc
    {
    public:
        COrbitRuntimeLosscone(const COrbit* _orbit, double _captureRadius);
        virtual FNCTYPE FncType() const {return FT_PERICENTER; };
        virtual STEPRESULT Timestep(const double told, const double t, CPosVelPoint<double> *vars);
        virtual void Finish() {};
        virtual CBasicInformation* getData() const 
        { return new CLossconeInformation(L2, deltaL2, tcapt, Ecapt, capturePoint); };
    private:
        int drdtsign;
        const double captureRadius;
        double tcapt, Ecapt, L2, prevL2, deltaL2, maxradius;
        CPosVelPoint<double> capturePoint;
    };

    /** Base class for tasks that are performed on each orbit (as RuntimeFunctions) 
        and before/after each episode.
        These tasks are descendants of RuntimeFncCreator class, which means that they are used 
        to create associated runtime functions for each orbit.
        They receive the pointer to the entire orbit library, so that at the end of each episode 
        they may walk through each COrbitDesc and collect necessary data from its information objects.
        The index of the relevant information object is passed as a parameter in the constructor
        (it is also the index of the given task in the array of RuntimeFncCreators).
    */
    class CBasicRagaTask: public CBasicOrbitRuntimeFncCreator
    {
    public:
        explicit CBasicRagaTask(CPointMassSet<double> * const _particles): 
          particles(_particles) 
        { infoOrbit.assign(particles->size(), NULL); };
        /// perform tasks at the end of each episode; return true if the entire simulation needs to be terminated
        virtual bool episodeDone(double episodeStart, double episodeLength)=0;
        void setInfoOrbit(size_t ind, const CBasicInformation* info) 
        { if(ind<infoOrbit.size()) { delete infoOrbit[ind]; infoOrbit[ind]=info; } }
    protected:
        CPointMassSet<double>* particles;
        std::vector<const CBasicInformation*> infoOrbit;
    };
    
    class CRagaTaskTrajectory: public CBasicRagaTask
    {
    public:
        CRagaTaskTrajectory(
            CPointMassSet<double>* const _particles, 
            const CPotential* const * _ptrPotential, 
            const std::string& _outputFileName, 
            double _timestepOutput
        );
        ~CRagaTaskTrajectory();
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeTrajEpisode(orbit, timestepOutput); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CPotential* const* ptrPotential;
        const std::string outputFileName;
        double timestepOutput;
        double prevOutputTime;
        template <typename NumT> bool writeSnapshot(const CPointMassSet<NumT> *points, double time);
    };

    class CRagaTaskPotential: public CBasicRagaTask
    {
    public:
        CRagaTaskPotential(
            CPointMassSet<double>* const _particles, 
            const CPotential** const _ptrPotential, 
            CMassModel** const _ptrMassModel, 
            const double* _ptrEpisodeLength, 
            size_t _numSamplesPerEpisode, 
            const CConfigPotential* _ptrConfigPotential, 
            const std::string& _outputPotentialFileName, 
            double _timestepOutput, 
            bool _updatePotential
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeTrajEpisode(orbit, (*ptrEpisodeLength)/numSamplesPerEpisode); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CPotential** ptrPotential;
        CMassModel** ptrMassModel;
        const double* ptrEpisodeLength;
        const size_t numSamplesPerEpisode;
        const CConfigPotential* ptrConfigPotential;
        const std::string outputPotentialFileName;
        bool updatePotential;
        double timestepOutput;
        double prevOutputTime;
        bool exportPotential(double time);
    };

    class CRagaTaskRelaxation: public CBasicRagaTask
    {
    public:
        CRagaTaskRelaxation(
            CPointMassSet<double>* const _particles, 
            const CMassModel* const* _ptrMassModel, 
            double* _ptrEpisodeLength, 
            double _relaxRate
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeVelPerturber(orbit, *ptrMassModel, relaxRate); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        const CMassModel* const* ptrMassModel;
        double* ptrEpisodeLength;
        const double relaxRate;
        double avgtrelc;
    };
        
    class CRagaTaskLosscone: public CBasicRagaTask
    {
    public:
        CRagaTaskLosscone(
            CPointMassSet<double>* const _particles, 
            double* _ptrMbh,
            double _captureRadius, 
            const std::string& _outputLossconeFileName
        );
        virtual CBasicOrbitRuntimeFnc* createRuntimeFnc(const COrbit* orbit) const 
        { return new COrbitRuntimeLosscone(orbit, captureRadius); };
        bool episodeDone(double episodeStart, double episodeLength);
    private:
        double* ptrMbh;  // pointer to configPotential.Mbh, to be updated when particles are captured
        const double captureRadius;
        const std::string outputLossconeFileName;
        double prevTimeSim;
        vectord particleMassInit; 
        vectord particlePericenterMinDist;
    };

    /// utility function to compute total energy of a particle in a spherical mass model
    template<typename NumT> double totalEnergySph(const CMassModel* model, const CPosVelPoint<NumT> &point);

}; // namespace